Package: featdelta
Type: Package
Title: Incremental Feature Engineering with Database Persistence
Version: 0.1.0
Authors@R: person(given = "Rudolfs", family = "Kregers", role = c("aut", "cre"),
    email = "rudolfs.kregers@gmail.com")
Author: Rudolfs Kregers [aut, cre]
Maintainer: Rudolfs Kregers <rudolfs.kregers@gmail.com>
Description: Define feature logic, compute only new or unprocessed
    rows, and persist the resulting flat feature table in a database.
    The package provides an explicit incremental pipeline for fetching
    source rows, computing feature definitions, and writing computed
    features to a database table.
License: GPL-3
BugReports: https://github.com/LordRudolf/featdelta/issues
Encoding: UTF-8
Imports: DBI, rlang
Suggests: RSQLite, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-08 18:21:14 UTC; Rudolfs
Repository: CRAN
Date/Publication: 2026-05-13 08:00:02 UTC
Built: R 4.6.0; ; 2026-05-13 08:35:58 UTC; unix

export(fd_block)
export(fd_compute)
export(fd_connect)
export(fd_define)
export(fd_fetch)
export(fd_run)
export(fd_upsert)
S3method(print,featdelta_block)
S3method(print,featdelta_con)
S3method(print,featdelta_defs)
S3method(print,fd_run_report)
S3method(summary,featdelta_defs)

# featdelta

`featdelta` is an R package for incremental feature engineering with database
persistence. It is designed for workflows where raw observations live in a
database, feature logic is easier to write and test in R, and the computed
features should be stored back in a database table for modelling, reporting,
monitoring, or downstream reuse.

Instead of rebuilding the same feature table from scratch whenever new rows
arrive, `featdelta` lets you define feature expressions in R, fetch only rows
that have not yet been processed, compute the features locally, and upsert the
results into a persistent feature table.

## Installation

```r
install.packages("featdelta")
```

You can install the development version from GitHub with:

```r
# install.packages("remotes")
remotes::install_github("LordRudolf/featdelta")
```

## Core idea

The standard pipeline is:

1. define reusable feature logic with `fd_define()`;
2. fetch raw database rows that are missing from the feature table;
3. compute the requested features in R;
4. create, extend, insert into, or update the database feature table.

The main orchestration function is `fd_run()`, which combines the fetch,
compute, and upsert steps.

## Small example

This example uses an in-memory SQLite database and `mtcars`, but the same
pattern applies to database tables selected with ordinary SQL.

```r
library(DBI)
library(RSQLite)
library(featdelta)

cars <- mtcars
cars$id <- seq_len(nrow(cars))

con <- dbConnect(SQLite(), ":memory:")
dbWriteTable(con, "raw_cars", cars[1:20, ], overwrite = TRUE)

defs <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp
)

run_day_one <- fd_run(
  con = con,
  sql = "SELECT * FROM raw_cars ORDER BY id",
  defs = defs,
  key = "id",
  feat_table_name = "car_features"
)

dbGetQuery(con, "SELECT * FROM car_features ORDER BY id")
```

When more raw rows arrive, call the same pipeline again. By default,
`fd_run(fetch_mode = "new_only")` processes only keys that are not already in the
feature table.

```r
dbAppendTable(con, "raw_cars", cars[21:30, ])

run_day_two <- fd_run(
  con = con,
  sql = "SELECT * FROM raw_cars ORDER BY id",
  defs = defs,
  key = "id",
  feat_table_name = "car_features"
)

dbGetQuery(con, "SELECT * FROM car_features ORDER BY id")
```

If feature definitions change and existing rows should be recomputed, use
`fetch_mode = "all"` to refresh the rows returned by the SQL query.

```r
defs_v2 <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp,
  mpg_per_cyl = mpg / cyl
)

fd_run(
  con = con,
  sql = "SELECT * FROM raw_cars ORDER BY id",
  defs = defs_v2,
  key = "id",
  feat_table_name = "car_features",
  fetch_mode = "all"
)
```

## Multi-column feature blocks

For feature logic that naturally produces several columns at once, use
`fd_block()`.

```r
defs <- fd_define(
  engine_ratios = fd_block({
    data.frame(
      hp_per_cyl = hp / cyl,
      disp_per_cyl = disp / cyl,
      wt_per_hp = wt / hp
    )
  })
)
```

## Learn more

The package includes vignettes covering the getting-started workflow, feature
definition patterns, database pipeline details, production patterns, and
scheduled runs.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(featdelta)

raw_cars <- mtcars
raw_cars$car_id <- seq_len(nrow(raw_cars))

raw_cars <- raw_cars[, c("car_id", "mpg", "cyl", "disp", "hp", "wt", "am")]

head(raw_cars)

## -----------------------------------------------------------------------------
defs_basic <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp
)

defs_basic

## -----------------------------------------------------------------------------
features_basic <- fd_compute(
  data = raw_cars,
  defs = defs_basic,
  key = "car_id"
)

head(features_basic)

## -----------------------------------------------------------------------------
defs_ordered <- fd_define(
  hp_per_cyl = hp / cyl,
  strong_engine = hp_per_cyl > 30,
  engine_label = ifelse(strong_engine, "strong", "regular")
)

features_ordered <- fd_compute(
  data = raw_cars,
  defs = defs_ordered,
  key = "car_id"
)

head(features_ordered)

## -----------------------------------------------------------------------------
log_hp_expr <- expression(log(hp))
heavy_car_expr <- expression(wt > 3.5)

defs_programmatic <- fd_define(
  log_hp = log_hp_expr,
  heavy_car = heavy_car_expr
)

features_programmatic <- fd_compute(
  data = raw_cars,
  defs = defs_programmatic,
  key = "car_id"
)

head(features_programmatic)

## -----------------------------------------------------------------------------
defs_block <- fd_define(
  engine_ratios = fd_block({
    data.frame(
      hp_per_cyl = hp / cyl,
      disp_per_cyl = disp / cyl,
      wt_per_hp = wt / hp
    )
  })
)

features_block <- fd_compute(
  data = raw_cars,
  defs = defs_block,
  key = "car_id"
)

head(features_block)

## -----------------------------------------------------------------------------
defs_script_block <- fd_define(
  engine_script = fd_block({
    hp_per_cyl <- hp / cyl
    disp_per_cyl <- disp / cyl

    ratio_average <- (hp_per_cyl + disp_per_cyl) / 2
    high_ratio <- ratio_average > stats::median(ratio_average, na.rm = TRUE)

    data.frame(
      hp_per_cyl = hp_per_cyl,
      disp_per_cyl = disp_per_cyl,
      engine_ratio_average = ratio_average,
      high_engine_ratio = high_ratio
    )
  })
)

features_script_block <- fd_compute(
  data = raw_cars,
  defs = defs_script_block,
  key = "car_id"
)

head(features_script_block)

## -----------------------------------------------------------------------------
make_engine_features <- function(data) {
  hp_per_cyl <- data$hp / data$cyl
  disp_per_cyl <- data$disp / data$cyl

  data.frame(
    hp_per_cyl = hp_per_cyl,
    disp_per_cyl = disp_per_cyl,
    engine_index = hp_per_cyl + disp_per_cyl
  )
}

defs_function_block <- fd_define(
  engine_features = fd_block(make_engine_features)
)

features_function_block <- fd_compute(
  data = raw_cars,
  defs = defs_function_block,
  key = "car_id"
)

head(features_function_block)

## -----------------------------------------------------------------------------
make_scaled_features <- function(data) {
  vars <- c("hp", "disp", "wt")
  out <- list()

  for (var in vars) {
    center <- mean(data[[var]], na.rm = TRUE)
    spread <- stats::sd(data[[var]], na.rm = TRUE)

    out[[paste0(var, "_scaled")]] <- (data[[var]] - center) / spread
  }

  as.data.frame(out)
}

defs_loop_block <- fd_define(
  scaled_inputs = fd_block(make_scaled_features)
)

features_loop_block <- fd_compute(
  data = raw_cars,
  defs = defs_loop_block,
  key = "car_id"
)

head(features_loop_block)

## -----------------------------------------------------------------------------
defs_combined <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  engine_features = fd_block(make_engine_features),
  scaled_inputs = fd_block(make_scaled_features),
  engine_per_weight = engine_index / wt
)

features_combined <- fd_compute(
  data = raw_cars,
  defs = defs_combined,
  key = "car_id"
)

head(features_combined)

## -----------------------------------------------------------------------------
defs_expected <- fd_define(
  optional_engine_flags = fd_block(
    {
      data.frame(
        high_hp = hp > 150
      )
    },
    expected_names = c("high_hp", "high_disp")
  )
)

features_expected <- fd_compute(
  data = raw_cars,
  defs = defs_expected,
  key = "car_id"
)

head(features_expected)

## ----eval = FALSE-------------------------------------------------------------
# # Local computation while developing feature logic
# fd_compute(raw_data, defs, key = "id")
# 
# # Full database pipeline once the definitions are ready
# fd_run(con, sql, defs, key = "id", feat_table_name = "feature_table")

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(DBI)
library(RSQLite)
library(featdelta)

con <- dbConnect(SQLite(), ":memory:")

readings <- data.frame(
  reading_id = 1:10,
  device_id = c("A", "A", "B", "C", "A", "B", "C", "A", "B", "C"),
  temperature = c(68, 71, 75, 70, 83, 78, 72, 88, 81, 74),
  vibration = c(0.12, 0.15, 0.31, 0.18, 0.44, 0.39, 0.22, 0.55, 0.41, 0.24),
  pressure = c(31, 33, 36, 32, 39, 37, 34, 42, 38, 35),
  runtime_hours = c(120, 125, 210, 88, 130, 216, 93, 136, 223, 97),
  stringsAsFactors = FALSE
)

day_one <- 1:6
day_two <- 7:10

dbWriteTable(
  con,
  "raw_readings",
  readings[day_one, ],
  overwrite = TRUE
)

source_sql <- "
  SELECT
    reading_id,
    device_id,
    temperature,
    vibration,
    pressure,
    runtime_hours
  FROM raw_readings
  ORDER BY reading_id
"

key <- "reading_id"

dbGetQuery(con, source_sql)

## -----------------------------------------------------------------------------
defs <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  pressure_per_hour = pressure / runtime_hours,
  maintenance_flag = temp_above_80 | vibration_score > 80
)

raw_preview <- dbGetQuery(con, source_sql)

features_preview <- fd_compute(
  data = raw_preview,
  defs = defs,
  key = key
)

features_preview

## -----------------------------------------------------------------------------
run_initial <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "reading_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM reading_features ORDER BY reading_id")

## -----------------------------------------------------------------------------
dev_report <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "reading_features_dev",
  fetch_limit = 3,
  return_data = "both",
  preview_n = 2,
  verbose = FALSE
)

dev_report$preview$raw
dev_report$preview$features
dev_report$data$features

## -----------------------------------------------------------------------------
run_initial$success
run_initial$stage
run_initial$fetch$n_rows
run_initial$compute$feature_names
run_initial$upsert$counts

## -----------------------------------------------------------------------------
bad_defs <- fd_define(
  temp_above_80 = temperature > 80,
  broken_feature = missing_sensor_column / 10
)

failed_report <- fd_run(
  con = con,
  sql = source_sql,
  defs = bad_defs,
  key = key,
  feat_table_name = "reading_features_broken",
  fail_fast = FALSE,
  return_data = "both",
  verbose = FALSE
)

failed_report$success
failed_report$stage
failed_report$error
failed_report$compute$report

## ----eval = FALSE-------------------------------------------------------------
# fd_run(
#   con = con,
#   sql = source_sql,
#   defs = bad_defs,
#   key = key,
#   feat_table_name = "reading_features",
#   fail_fast = TRUE
# )

## -----------------------------------------------------------------------------
dbAppendTable(con, "raw_readings", readings[day_two, ])

run_incremental <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "reading_features",
  verbose = FALSE
)

# The default fetch mode is "new_only".
run_incremental$fetch$mode

# Only the four new day-two readings were fetched.
run_incremental$fetch$n_rows

# Four rows were inserted, and no existing feature rows were updated.
run_incremental$upsert$counts

## -----------------------------------------------------------------------------
defs_v2 <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  pressure_per_hour = pressure / runtime_hours,
  maintenance_flag = temp_above_80 | vibration_score > 80,
  high_pressure = pressure >= 38
)

run_refresh <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "reading_features",
  fetch_mode = "all",
  verbose = FALSE
)

# Refresh mode fetches all rows returned by the source query.
run_refresh$fetch$mode
run_refresh$fetch$n_rows

# All rows already existed in the feature table, so they were updated.
# No new rows were inserted.
run_refresh$upsert$counts

## -----------------------------------------------------------------------------
defs_v3 <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  maintenance_flag = temp_above_80 | vibration_score > 80,
  high_pressure = pressure >= 38
)

run_removed_column <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v3,
  key = key,
  feat_table_name = "reading_features",
  fetch_mode = "all",
  verbose = FALSE
)

# All ten source rows already existed in the feature table, so they were
# counted as updates rather than inserts.
run_removed_column$upsert$counts

# The database table still contains the old `pressure_per_hour` column.
dbGetQuery(con, "pragma table_info(reading_features)")

## -----------------------------------------------------------------------------
defs_v4 <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  maintenance_flag = temp_above_80 | vibration_score > 80,
  high_pressure = pressure >= 38,
  thermal_load = temperature * runtime_hours
)

alter_error <- tryCatch(
  fd_run(
    con = con,
    sql = source_sql,
    defs = defs_v4,
    key = key,
    feat_table_name = "reading_features",
    fetch_mode = "all",
    alter_table = FALSE,
    verbose = FALSE
  ),
  error = function(e) conditionMessage(e)
)

alter_error

## -----------------------------------------------------------------------------
insert_only_error <- tryCatch(
  fd_run(
    con = con,
    sql = source_sql,
    defs = defs_v3,
    key = key,
    feat_table_name = "reading_features",
    fetch_mode = "all",
    update_table = FALSE,
    verbose = FALSE
  ),
  error = function(e) conditionMessage(e)
)

insert_only_error

## ----eval = FALSE-------------------------------------------------------------
# fd_run(
#   con = con,
#   sql = source_sql,
#   defs = defs,
#   key = key,
#   feat_table_name = "reading_features",
#   chunk_size = 10000
# )

## ----cleanup, include = FALSE-------------------------------------------------
dbDisconnect(con)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# # SQLite
# con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
# 
# # PostgreSQL
# con <- DBI::dbConnect(
#   RPostgres::Postgres(),
#   host = "localhost",
#   dbname = "analytics",
#   user = "analyst",
#   password = "secret"
# )
# 
# # MySQL or MariaDB
# con <- DBI::dbConnect(
#   RMariaDB::MariaDB(),
#   host = "localhost",
#   dbname = "analytics",
#   user = "analyst",
#   password = "secret"
# )

## -----------------------------------------------------------------------------
library(DBI)
library(RSQLite)
library(featdelta)

con <- dbConnect(SQLite(), ":memory:")

orders <- data.frame(
  order_id = 1:7,
  customer_id = c(101, 102, 103, 101, 104, 105, 102),
  gross_amount = c(120, 250, 80, 310, 45, 520, 160),
  discount_amount = c(0, 25, 5, 30, 0, 60, 10),
  shipping_fee = c(8, 0, 6, 0, 5, 0, 7),
  order_to_ship_days = c(1, 3, 2, 5, 1, 4, 2),
  stringsAsFactors = FALSE
)

day_one <- 1:4
day_two <- 5:7

dbWriteTable(
  con,
  "raw_orders",
  orders[day_one, ],
  overwrite = TRUE
)

dbGetQuery(con, "SELECT * FROM raw_orders ORDER BY order_id")

## -----------------------------------------------------------------------------
source_sql <- "
  SELECT
    order_id,
    customer_id,
    gross_amount,
    discount_amount,
    shipping_fee,
    order_to_ship_days
  FROM raw_orders
  ORDER BY order_id
"

key <- "order_id"

dbGetQuery(con, source_sql)

## -----------------------------------------------------------------------------
defs <- fd_define(
  net_revenue = gross_amount - discount_amount + shipping_fee,
  discount_rate = discount_amount / gross_amount,
  free_shipping = shipping_fee == 0,
  slow_fulfillment = order_to_ship_days > 3
)

defs

## -----------------------------------------------------------------------------
run_day_one <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "order_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")

## -----------------------------------------------------------------------------
dbAppendTable(con, "raw_orders", orders[day_two, ])

dbGetQuery(con, "SELECT * FROM raw_orders ORDER BY order_id")

## -----------------------------------------------------------------------------
dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")

## -----------------------------------------------------------------------------
run_day_two <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "order_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")

## -----------------------------------------------------------------------------
new_order <- data.frame(
  order_id = 8,
  customer_id = 106,
  gross_amount = 275,
  discount_amount = 20,
  shipping_fee = 0,
  order_to_ship_days = 6
)

dbAppendTable(con, "raw_orders", new_order)

new_rows <- fd_fetch(
  con = con,
  sql = source_sql,
  key = key,
  feat_table_name = "order_features"
)

new_rows

## -----------------------------------------------------------------------------
attr(new_rows, "fd_fetch")

## -----------------------------------------------------------------------------
new_features <- fd_compute(
  data = new_rows,
  defs = defs,
  key = key
)

upsert_report <- fd_upsert(
  con = con,
  features_df = new_features,
  feat_table_name = "order_features",
  key = key,
  verbose = FALSE
)

upsert_report

dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")

## -----------------------------------------------------------------------------
defs_v2 <- fd_define(
  net_revenue = gross_amount - discount_amount + shipping_fee,
  discount_rate = discount_amount / gross_amount,
  free_shipping = shipping_fee == 0,
  slow_fulfillment = order_to_ship_days > 3,
  high_value_order = gross_amount >= 250
)

refresh_report <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "order_features",
  fetch_mode = "all",
  verbose = FALSE
)

dbListFields(con, "order_features")

dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")

## ----eval = FALSE-------------------------------------------------------------
# # Error if the table does not already exist
# fd_run(
#   con = con,
#   sql = source_sql,
#   defs = defs,
#   key = key,
#   feat_table_name = "order_features",
#   create_table = FALSE
# )
# 
# # Error if new feature columns are missing from the existing table
# fd_run(
#   con = con,
#   sql = source_sql,
#   defs = defs_v2,
#   key = key,
#   feat_table_name = "order_features",
#   alter_table = FALSE
# )

## ----eval = FALSE-------------------------------------------------------------
# fd_run(
#   con = con,
#   sql = source_sql,
#   defs = defs,
#   key = key,
#   feat_table_name = "order_features",
#   fetch_mode = "all",
#   update_table = FALSE
# )

## ----eval = FALSE-------------------------------------------------------------
# fd_run(
#   con = con,
#   sql = source_sql,
#   defs = defs,
#   key = key,
#   feat_table_name = "order_features",
#   chunk_size = 5000
# )

## ----eval = FALSE-------------------------------------------------------------
# fd_run(
#   con = con,
#   sql = "SELECT * FROM raw_schema.orders",
#   defs = defs,
#   key = "order_id",
#   feat_table_name = "feature_schema.order_features",
#   dialect = "postgres"
# )
# 
# fd_run(
#   con = con,
#   sql = "SELECT * FROM raw_schema.orders",
#   defs = defs,
#   key = "order_id",
#   feat_table_name = "feature_schema.order_features",
#   dialect = "mysql"
# )

## -----------------------------------------------------------------------------
names(refresh_report)

refresh_report$fetch

refresh_report$compute$feature_names

refresh_report$upsert$counts

refresh_report$upsert$columns_added

## ----cleanup, include = FALSE-------------------------------------------------
dbDisconnect(con)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(DBI)
library(RSQLite)
library(featdelta)

cars <- mtcars
cars$id <- seq_len(nrow(cars))

# Name of the unique row identifier used throughout the pipeline.
key <- "id"

day_one <- 1:20
day_two <- 21:30

## -----------------------------------------------------------------------------
con <- dbConnect(SQLite(), ":memory:")
dbWriteTable(con, "raw_cars", cars[day_one, ], overwrite = TRUE)

dbGetQuery(con, "SELECT id, mpg, cyl, hp, wt, am FROM raw_cars ORDER BY id")

## -----------------------------------------------------------------------------
source_sql <- "SELECT * FROM raw_cars WHERE id > 15 ORDER BY id"

dbGetQuery(con, source_sql)

## -----------------------------------------------------------------------------
defs <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp
)

defs

## -----------------------------------------------------------------------------
raw_preview <- dbGetQuery(con, source_sql)

fd_compute(
  data = raw_preview,
  defs = defs,
  key = key
)

## -----------------------------------------------------------------------------
run_day_one <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "car_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM car_features ORDER BY id")

## -----------------------------------------------------------------------------
dbAppendTable(con, "raw_cars", cars[day_two, ])

dbGetQuery(
  con,
  "SELECT id, mpg, cyl, hp, wt, am FROM raw_cars WHERE id > 15 ORDER BY id DESC"
)

## -----------------------------------------------------------------------------
dbGetQuery(
  con,
  "SELECT * FROM car_features ORDER BY id DESC"
)

## -----------------------------------------------------------------------------
run_day_two <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "car_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM car_features ORDER BY id DESC")

## -----------------------------------------------------------------------------
defs_v2 <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp,
  power_to_weight = hp / wt
)

run_refresh <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "car_features",
  fetch_mode = "all",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM car_features ORDER BY id DESC")

## ----eval = FALSE-------------------------------------------------------------
# defs <- fd_define(
#   feature_a = some_r_expression,
#   feature_b = another_r_expression
# )
# 
# fd_run(
#   con = con,
#   sql = "SELECT * FROM raw_table",
#   defs = defs,
#   key = "id",
#   feat_table_name = "feature_table"
# )

## -----------------------------------------------------------------------------
run_report <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "car_features",
  fetch_mode = "all",
  return_data = "features",
  preview_n = 3,
  verbose = FALSE
)

run_report

## ----cleanup, include = FALSE-------------------------------------------------
dbDisconnect(con)

render_sql <- function(template, values) {
  out <- template

  for (nm in names(values)) {
    val <- values[[nm]]

    if (is.null(val) || length(val) != 1L || is.na(val)) {
      stop(
        sprintf("render_sql(): placeholder {%s} has invalid replacement value.", nm),
        call. = FALSE
      )
    }

    out <- gsub(paste0("{", nm, "}"), as.character(val), out, fixed = TRUE)
  }

  out
}


#' SQL templates used internally by featdelta
#'
#' A structured list of SQL templates keyed by feature (`fd_upsert`) and backend.
#' Templates are rendered by substituting placeholders with pre-quoted identifiers
#' and pre-built clauses.
#'
#' This object is internal and not part of the public API.
#'
#' ## Placeholder conventions
#' Templates expect the caller to provide the following placeholders as strings:
#' - `{target}`: quoted target table identifier
#' - `{stage}`: quoted staging table identifier
#' - `{cols_all}`: comma-separated quoted column identifiers
#' - `{key}`: quoted key column identifier
#' - `{set_clause}`: update `SET` clause, e.g. `"f1" = EXCLUDED."f1", "f2" = EXCLUDED."f2"`
#' - `{t_key}` / `{s_key}`: qualified key references for joins, e.g. `t."id"` and `s."id"`
#' - `{limit}`: integer literal for sampling conflicts
#'
#' @noRd
fd_sql_templates <- list(

  # ---- upsert: merge / write paths ----
  upsert_merge_upsert = list(
    postgres = paste(
      "INSERT INTO {target} ({cols_all})",
      "SELECT {cols_all} FROM {stage}",
      "ON CONFLICT ({key}) DO UPDATE SET {set_clause}"
    ),

    sqlite = paste(
      "INSERT INTO {target} ({cols_all})",
      "SELECT {cols_all} FROM {stage}",
      "WHERE TRUE",
      "ON CONFLICT({key}) DO UPDATE SET {set_clause}"
    ),

    mysql = paste(
      "INSERT INTO {target} ({cols_all})",
      "SELECT {cols_all}",
      "FROM (SELECT {cols_all} FROM {stage}) AS fd_src",
      "ON DUPLICATE KEY UPDATE {set_clause}"
    )
  ),

  upsert_merge_insert_only = list(
    postgres = paste(
      "INSERT INTO {target} ({cols_insert})",
      "SELECT {cols_select}",
      "FROM {stage} s",
      "LEFT JOIN {target} t ON {t_key} = {s_key}",
      "WHERE {t_key} IS NULL"
    ),

    sqlite = paste(
      "INSERT INTO {target} ({cols_insert})",
      "SELECT {cols_select}",
      "FROM {stage} s",
      "LEFT JOIN {target} t ON {t_key} = {s_key}",
      "WHERE {t_key} IS NULL"
    ),

    mysql = paste(
      "INSERT INTO {target} ({cols_insert})",
      "SELECT {cols_select}",
      "FROM {stage} s",
      "LEFT JOIN {target} t ON {t_key} = {s_key}",
      "WHERE {t_key} IS NULL"
    )
  ),

  # ---- upsert: counts (existence-based, pre-merge) ----
  upsert_count_would_insert = list(
    postgres = paste(
      "SELECT COUNT(*) AS n",
      "FROM {stage} s",
      "LEFT JOIN {target} t ON {t_key} = {s_key}",
      "WHERE {t_key} IS NULL"
    ),
    sqlite = paste(
      "SELECT COUNT(*) AS n",
      "FROM {stage} s",
      "LEFT JOIN {target} t ON {t_key} = {s_key}",
      "WHERE {t_key} IS NULL"
    ),
    mysql = paste(
      "SELECT COUNT(*) AS n",
      "FROM {stage} s",
      "LEFT JOIN {target} t ON {t_key} = {s_key}",
      "WHERE {t_key} IS NULL"
    )
  ),

  upsert_count_would_update = list(
    postgres = paste(
      "SELECT COUNT(*) AS n",
      "FROM {stage} s",
      "JOIN {target} t ON {t_key} = {s_key}"
    ),
    sqlite = paste(
      "SELECT COUNT(*) AS n",
      "FROM {stage} s",
      "JOIN {target} t ON {t_key} = {s_key}"
    ),
    mysql = paste(
      "SELECT COUNT(*) AS n",
      "FROM {stage} s",
      "JOIN {target} t ON {t_key} = {s_key}"
    )
  ),

  # ---- upsert: conflict sampling for insert-only mode ----
  upsert_find_conflicts = list(
    postgres = paste(
      "SELECT {s_key} AS key_value",
      "FROM {stage} s",
      "JOIN {target} t ON {t_key} = {s_key}",
      "LIMIT {limit}"
    ),
    sqlite = paste(
      "SELECT {s_key} AS key_value",
      "FROM {stage} s",
      "JOIN {target} t ON {t_key} = {s_key}",
      "LIMIT {limit}"
    ),
    mysql = paste(
      "SELECT {s_key} AS key_value",
      "FROM {stage} s",
      "JOIN {target} t ON {t_key} = {s_key}",
      "LIMIT {limit}"
    )
  ),

  # ---- DDL helpers (used for messaging / schema evolution) ----
  drop_table = list(
    postgres = "DROP TABLE {target}",
    sqlite   = "DROP TABLE {target}",
    mysql    = "DROP TABLE {target}"
  ),

  alter_add_column = list(
    postgres = "ALTER TABLE {target} ADD COLUMN {col} {type}",
    sqlite   = "ALTER TABLE {target} ADD COLUMN {col} {type}",
    mysql    = "ALTER TABLE {target} ADD COLUMN {col} {type}"
  )
)

fd_test_raw_data <- function() {
  data.frame(
    id = 1:5,
    x = c(10, 20, 30, 40, 50),
    y = c(2, 4, 5, 8, 10),
    group = c("a", "a", "b", "b", "c"),
    stringsAsFactors = FALSE
  )
}

fd_test_features_data <- function(ids = integer()) {
  data.frame(
    id = ids,
    existing_feature = ids * 100,
    stringsAsFactors = FALSE
  )
}

fd_test_sqlite_con <- function() {
  testthat::skip_if_not_installed("RSQLite")
  DBI::dbConnect(RSQLite::SQLite(), ":memory:")
}

fd_test_disconnect <- function(con) {
  if (inherits(con, "DBIConnection") && DBI::dbIsValid(con)) {
    DBI::dbDisconnect(con)
  }
  invisible(TRUE)
}

fd_test_seed_sqlite <- function(con,
                                raw = fd_test_raw_data(),
                                feature_ids = integer()) {
  features <- fd_test_features_data(feature_ids)

  DBI::dbWriteTable(con, "raw_tbl", raw, overwrite = TRUE)
  DBI::dbWriteTable(con, "features_tbl", features, overwrite = TRUE)

  invisible(list(
    raw = raw,
    features = features
  ))
}

test_that("shared raw-data fixture is keyed and deterministic", {
  raw <- fd_test_raw_data()

  expect_s3_class(raw, "data.frame")
  expect_named(raw, c("id", "x", "y", "group"))
  expect_identical(raw$id, 1:5)
  expect_equal(anyDuplicated(raw$id), 0L)
  expect_false(anyNA(raw$id))
})

test_that("shared feature-data fixture is keyed by requested ids", {
  features <- fd_test_features_data(c(2L, 4L))

  expect_s3_class(features, "data.frame")
  expect_named(features, c("id", "existing_feature"))
  expect_identical(features$id, c(2L, 4L))
  expect_identical(features$existing_feature, c(200, 400))
})

test_that("SQLite fixture seeds raw and feature tables", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  seeded <- fd_test_seed_sqlite(con, feature_ids = c(1L, 3L))

  expect_true(DBI::dbExistsTable(con, "raw_tbl"))
  expect_true(DBI::dbExistsTable(con, "features_tbl"))
  expect_equal(DBI::dbReadTable(con, "raw_tbl"), seeded$raw)
  expect_equal(DBI::dbReadTable(con, "features_tbl"), seeded$features)
})

test_that("fd_block expression blocks compute through fd_define and fd_compute", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    ratios = fd_block({
      data.frame(
        ratio = x / y,
        sum_xy = x + y
      )
    }),
    after_block = ratio + sum_xy
  )

  out <- fd_compute(raw, defs, key = "id")
  report <- attr(out, "fd_report")

  expect_named(out, c("id", "ratio", "sum_xy", "after_block"))
  expect_equal(out$ratio, raw$x / raw$y)
  expect_equal(out$sum_xy, raw$x + raw$y)
  expect_equal(out$after_block, (raw$x / raw$y) + raw$x + raw$y)

  expect_equal(report$step_name, c("ratios", "after_block"))
  expect_equal(report$type, c("block", "column"))
  expect_equal(report$mode, c("expression", "expression"))
  expect_true(all(report$ok))
  expect_equal(report$output_names, c("ratio, sum_xy", "after_block"))
})

test_that("fd_block function blocks receive raw and previously produced columns", {
  raw <- fd_test_raw_data()
  make_more <- function(data) {
    data.frame(
      ratio_plus_one = data$ratio + 1,
      group_copy = data$group
    )
  }

  defs <- fd_define(
    ratio = x / y,
    more = fd_block(make_more)
  )

  out <- fd_compute(raw, defs, key = "id")

  expect_named(out, c("id", "ratio", "ratio_plus_one", "group_copy"))
  expect_equal(out$ratio, raw$x / raw$y)
  expect_equal(out$ratio_plus_one, (raw$x / raw$y) + 1)
  expect_identical(out$group_copy, raw$group)
})

test_that("fd_block expected names complete missing outputs with NA and define order", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    expected = fd_block(
      {
        data.frame(sum_xy = x + y)
      },
      expected_names = c("ratio", "sum_xy")
    )
  )

  out <- fd_compute(raw, defs, key = "id")
  report <- attr(out, "fd_report")

  expect_named(out, c("id", "ratio", "sum_xy"))
  expect_true(all(is.na(out$ratio)))
  expect_equal(out$sum_xy, raw$x + raw$y)
  expect_true(report$ok)
  expect_equal(report$output_names, "ratio, sum_xy")
})

test_that("fd_block rejects unexpected names when expected names are declared", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    expected = fd_block(
      {
        data.frame(sum_xy = x + y, unexpected = x - y)
      },
      expected_names = c("ratio", "sum_xy")
    )
  )

  err <- expect_error(
    fd_compute(raw, defs, key = "id"),
    class = "featdelta_compute_error"
  )
  report <- attr(err, "fd_report")

  expect_equal(report$step_name, "expected")
  expect_false(report$ok)
  expect_match(report$error, "unexpected column name")
})

test_that("fd_block rejects block outputs that collide with key or earlier features", {
  raw <- fd_test_raw_data()

  key_collision <- fd_define(
    bad = fd_block({
      data.frame(id = x + y)
    })
  )
  err_key <- expect_error(
    fd_compute(raw, key_collision, key = "id"),
    class = "featdelta_compute_error"
  )
  expect_match(attr(err_key, "fd_report")$error, "colliding with `key`")

  earlier_collision <- fd_define(
    ratio = x / y,
    bad = fd_block({
      data.frame(ratio = x + y)
    })
  )
  err_earlier <- expect_error(
    fd_compute(raw, earlier_collision, key = "id"),
    class = "featdelta_compute_error"
  )
  expect_match(attr(err_earlier, "fd_report")$error[[2]], "already produced earlier")
})

test_that("fd_block reports malformed block return values", {
  raw <- fd_test_raw_data()

  expect_error(
    fd_compute(
      raw,
      fd_define(bad = fd_block({ x + y })),
      key = "id"
    ),
    "one or more definition steps failed",
    class = "featdelta_compute_error"
  )

  expect_error(
    fd_compute(
      raw,
      fd_define(bad = fd_block({ data.frame(value = x[1:2]) })),
      key = "id"
    ),
    "one or more definition steps failed",
    class = "featdelta_compute_error"
  )

  expect_error(
    fd_compute(
      raw,
      fd_define(bad = fd_block({
        out <- data.frame(a = x, b = y)
        names(out) <- c("dup", "dup")
        out
      })),
      key = "id"
    ),
    "one or more definition steps failed",
    class = "featdelta_compute_error"
  )
})

test_that("fd_block non-strict failures can fill declared outputs with NA", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    bad = fd_block(
      {
        stop("broken block")
      },
      expected_names = c("first", "second")
    ),
    after = x + y
  )

  expect_warning(
    out <- fd_compute(raw, defs, key = "id", compute_strict = FALSE),
    "step 'bad' failed"
  )
  report <- attr(out, "fd_report")

  expect_named(out, c("id", "first", "second", "after"))
  expect_true(all(is.na(out$first)))
  expect_true(all(is.na(out$second)))
  expect_equal(out$after, raw$x + raw$y)
  expect_equal(report$ok, c(FALSE, TRUE))
  expect_equal(report$output_names, c("first, second", "after"))
})

test_that("fd_block validates constructor-only arguments", {
  expect_error(fd_block(), "`x` must be supplied")
  expect_error(fd_block({ data.frame(a = x) }, envir = list()), "`envir`")
  expect_error(
    fd_block({ data.frame(a = x) }, expected_names = character()),
    "must not be empty"
  )
  expect_error(
    fd_block({ data.frame(a = x) }, expected_names = c("a", NA_character_)),
    "non-empty, non-NA"
  )
  expect_error(
    fd_block({ data.frame(a = x) }, expected_names = c("a", "a")),
    "Duplicate names"
  )
})

test_that("print.featdelta_block returns input invisibly and shows block metadata", {
  block <- fd_block(
    {
      data.frame(ratio = x / y)
    },
    expected_names = "ratio"
  )

  printed <- capture.output(ret <- print(block))

  expect_s3_class(block, "featdelta_block")
  expect_identical(ret, block)
  expect_true(any(grepl("<featdelta_block>", printed, fixed = TRUE)))
  expect_true(any(grepl("Mode: expression", printed, fixed = TRUE)))
  expect_true(any(grepl("Expected names: ratio", printed, fixed = TRUE)))
  expect_true(any(grepl("Definition:", printed, fixed = TRUE)))
})

test_that("fd_fetch returns rows whose keys are absent from the feature table", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_test_seed_sqlite(con, feature_ids = c(1L, 3L))

  out <- fd_fetch(
    con = con,
    sql = "SELECT * FROM raw_tbl ORDER BY id",
    key = "id",
    feat_table_name = "features_tbl",
    use_max_key = FALSE
  )

  expect_s3_class(out, "data.frame")
  expect_identical(out$id, c(2L, 4L, 5L))
  expect_equal(out$x, c(20, 40, 50))
  expect_equal(out$y, c(4, 8, 10))
  expect_equal(out$group, c("a", "b", "c"))
})

test_that("fd_fetch attaches fetch metadata", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_test_seed_sqlite(con, feature_ids = c(1L, 3L))

  out <- fd_fetch(
    con = con,
    sql = "  SELECT * FROM raw_tbl ORDER BY id;  ",
    key = "id",
    feat_table_name = "features_tbl",
    use_max_key = FALSE
  )
  info <- attr(out, "fd_fetch")

  expect_type(info, "list")
  expect_identical(info$key, "id")
  expect_identical(info$feat_table_name, "features_tbl")
  expect_false(info$use_max_key)
  expect_true(is.na(info$max_key))
  expect_identical(info$sql, "SELECT * FROM raw_tbl ORDER BY id")
  expect_match(info$executed_sql, "LEFT JOIN `features_tbl` AS f", fixed = TRUE)
  expect_match(info$executed_sql, "WHERE f.`id` IS NULL", fixed = TRUE)
  expect_identical(info$n_rows, 3L)
})

test_that("fd_fetch use_max_key fetches only keys above the current maximum", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_test_seed_sqlite(con, feature_ids = c(1L, 3L))

  out <- fd_fetch(
    con = con,
    sql = "SELECT * FROM raw_tbl ORDER BY id",
    key = "id",
    feat_table_name = "features_tbl",
    use_max_key = TRUE
  )
  info <- attr(out, "fd_fetch")

  expect_identical(out$id, c(4L, 5L))
  expect_identical(info$max_key, 3L)
  expect_match(info$executed_sql, "WHERE r.`id` > 3", fixed = TRUE)
  expect_identical(info$n_rows, 2L)
})

test_that("fd_fetch use_max_key returns all rows when the feature table is empty", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_test_seed_sqlite(con, feature_ids = integer())

  out <- fd_fetch(
    con = con,
    sql = "SELECT * FROM raw_tbl ORDER BY id",
    key = "id",
    feat_table_name = "features_tbl",
    use_max_key = TRUE
  )
  info <- attr(out, "fd_fetch")

  expect_identical(out$id, 1:5)
  expect_true(is.na(info$max_key))
  expect_match(
    info$executed_sql,
    "SELECT * FROM (SELECT * FROM raw_tbl ORDER BY id) AS r",
    fixed = TRUE
  )
  expect_identical(info$n_rows, 5L)
})

test_that("fd_fetch supports WITH queries that can be used as derived tables", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_test_seed_sqlite(con, feature_ids = c(2L, 4L))

  out <- fd_fetch(
    con = con,
    sql = paste(
      "WITH raw_subset AS (",
      "SELECT id, x, y FROM raw_tbl WHERE x >= 20",
      ")",
      "SELECT * FROM raw_subset ORDER BY id"
    ),
    key = "id",
    feat_table_name = "features_tbl"
  )

  expect_named(out, c("id", "x", "y"))
  expect_identical(out$id, c(3L, 5L))
})

test_that("fd_fetch errors when the feature table is missing", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  DBI::dbWriteTable(con, "raw_tbl", fd_test_raw_data(), overwrite = TRUE)

  expect_error(
    fd_fetch(
      con = con,
      sql = "SELECT * FROM raw_tbl",
      key = "id",
      feat_table_name = "features_tbl"
    ),
    "`feat_table_name` does not exist"
  )
})

test_that("fd_fetch errors when the feature table lacks the key column", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  DBI::dbWriteTable(con, "raw_tbl", fd_test_raw_data(), overwrite = TRUE)
  DBI::dbWriteTable(
    con,
    "features_tbl",
    data.frame(other_id = c(1L, 3L), existing_feature = c(100, 300)),
    overwrite = TRUE
  )

  expect_error(
    fd_fetch(
      con = con,
      sql = "SELECT * FROM raw_tbl",
      key = "id",
      feat_table_name = "features_tbl"
    ),
    "does not contain the `key` column"
  )
})

test_that("fd_fetch errors when SQL does not return the key column", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_test_seed_sqlite(con, feature_ids = c(1L, 3L))

  expect_error(
    fd_fetch(
      con = con,
      sql = "SELECT x, y FROM raw_tbl",
      key = "id",
      feat_table_name = "features_tbl"
    ),
    "`sql` does not return a column named 'id'"
  )
})

test_that("fd_fetch validates common arguments before querying", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_test_seed_sqlite(con, feature_ids = c(1L))

  expect_error(
    fd_fetch(con, "DELETE FROM raw_tbl", key = "id", feat_table_name = "features_tbl"),
    "SELECT query"
  )
  expect_error(
    fd_fetch(con, "SELECT * FROM raw_tbl", key = "", feat_table_name = "features_tbl"),
    "key.*non-empty"
  )
  expect_error(
    fd_fetch(con, "SELECT * FROM raw_tbl", key = "id", feat_table_name = ""),
    "`feat_table_name`"
  )
  expect_error(
    fd_fetch(con, "SELECT * FROM raw_tbl", key = "id", feat_table_name = "features_tbl", use_max_key = NA),
    "use_max_key must be TRUE/FALSE"
  )
  expect_error(
    fd_fetch(con, "SELECT * FROM raw_tbl", key = "id", feat_table_name = "features_tbl", verbose = "yes"),
    "verbose must be TRUE/FALSE"
  )
})

test_that("fd_fetch emits executed SQL when verbose", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_test_seed_sqlite(con, feature_ids = c(1L, 3L))

  expect_message(
    out <- fd_fetch(
      con = con,
      sql = "SELECT * FROM raw_tbl ORDER BY id",
      key = "id",
      feat_table_name = "features_tbl",
      verbose = TRUE
    ),
    "fd_fetch SQL:"
  )

  expect_identical(out$id, c(2L, 4L, 5L))
})

fd_upsert_read_table <- function(con, table = "features_tbl") {
  DBI::dbGetQuery(con, paste0("SELECT * FROM ", table, " ORDER BY id"))
}

fd_upsert_create_target <- function(con, features) {
  fd_upsert(
    con = con,
    features_df = features,
    feat_table_name = "features_tbl",
    key = "id",
    create_table = TRUE,
    verbose = FALSE
  )
  invisible(TRUE)
}

test_that("fd_upsert creates a missing table with create_table auto and inserts rows", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  features <- data.frame(
    id = 1:3,
    score = c(10, 20, 30),
    flag = c(TRUE, FALSE, TRUE),
    label = c("a", "b", "c"),
    stringsAsFactors = FALSE
  )

  report <- fd_upsert(
    con = con,
    features_df = features,
    feat_table_name = "features_tbl",
    key = "id",
    create_table = "auto",
    verbose = FALSE
  )

  expect_s3_class(report, "fd_upsert_report")
  expect_true(report$table_created)
  expect_identical(report$columns_added, character())
  expect_identical(report$n_rows, 3L)
  expect_identical(report$n_chunks, 1L)
  expect_identical(report$counts$would_insert, 3L)
  expect_identical(report$counts$would_update, 0L)
  expect_equal(report$chunk_details$n, 3L)
  expect_equal(report$chunk_details$would_insert, 3L)

  out <- fd_upsert_read_table(con)
  expect_equal(out$id, features$id)
  expect_equal(out$score, features$score)
  expect_equal(out$flag, as.integer(features$flag))
  expect_equal(out$label, features$label)
})

test_that("fd_upsert inserts new keys and updates existing keys", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  initial <- data.frame(
    id = 1:3,
    score = c(10, 20, 30),
    label = c("a", "b", "c"),
    stringsAsFactors = FALSE
  )
  fd_upsert_create_target(con, initial)

  incoming <- data.frame(
    id = c(2L, 3L, 4L),
    score = c(21, 31, 40),
    label = c("b2", "c2", "d"),
    stringsAsFactors = FALSE
  )
  report <- fd_upsert(
    con = con,
    features_df = incoming,
    feat_table_name = "features_tbl",
    key = "id",
    update_table = TRUE,
    verbose = FALSE
  )

  expect_false(report$table_created)
  expect_identical(report$counts$would_insert, 1L)
  expect_identical(report$counts$would_update, 2L)

  out <- fd_upsert_read_table(con)
  expect_equal(out$id, 1:4)
  expect_equal(out$score, c(10, 21, 31, 40))
  expect_equal(out$label, c("a", "b2", "c2", "d"))
})

test_that("fd_upsert insert-only mode inserts non-conflicting rows", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  fd_upsert_create_target(con, data.frame(id = 1:2, score = c(10, 20)))

  incoming <- data.frame(id = 3:4, score = c(30, 40))
  report <- fd_upsert(
    con = con,
    features_df = incoming,
    feat_table_name = "features_tbl",
    key = "id",
    update_table = FALSE,
    verbose = FALSE
  )

  expect_identical(report$counts$would_insert, 2L)
  expect_identical(report$counts$would_update, 0L)
  expect_equal(fd_upsert_read_table(con)$id, 1:4)
})

test_that("fd_upsert insert-only mode errors and leaves existing rows unchanged on conflicts", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  initial <- data.frame(id = 1:3, score = c(10, 20, 30))
  fd_upsert_create_target(con, initial)

  expect_error(
    fd_upsert(
      con = con,
      features_df = data.frame(id = c(3L, 4L), score = c(31, 40)),
      feat_table_name = "features_tbl",
      key = "id",
      update_table = FALSE,
      verbose = FALSE
    ),
    "Conflicts detected"
  )

  expect_equal(fd_upsert_read_table(con), initial)
})

test_that("fd_upsert adds missing feature columns when alter_table is TRUE", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  fd_upsert_create_target(con, data.frame(id = 1:2, score = c(10, 20)))

  incoming <- data.frame(
    id = c(2L, 3L),
    score = c(21, 30),
    label = c("b2", "c"),
    stringsAsFactors = FALSE
  )
  report <- fd_upsert(
    con = con,
    features_df = incoming,
    feat_table_name = "features_tbl",
    key = "id",
    alter_table = TRUE,
    verbose = FALSE
  )

  expect_identical(report$columns_added, "label")
  expect_identical(report$counts$would_insert, 1L)
  expect_identical(report$counts$would_update, 1L)
  expect_equal(DBI::dbListFields(con, "features_tbl"), c("id", "score", "label"))

  out <- fd_upsert_read_table(con)
  expect_equal(out$id, 1:3)
  expect_equal(out$score, c(10, 21, 30))
  expect_equal(out$label, c(NA, "b2", "c"))
})

test_that("fd_upsert reports extra target columns without mutating them", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  initial <- data.frame(
    id = 1:2,
    score = c(10, 20),
    retired = c("old-a", "old-b"),
    stringsAsFactors = FALSE
  )
  fd_upsert_create_target(con, initial)

  report <- fd_upsert(
    con = con,
    features_df = data.frame(id = c(2L, 3L), score = c(21, 30)),
    feat_table_name = "features_tbl",
    key = "id",
    verbose = FALSE
  )

  expect_identical(report$extra_columns, "retired")
  expect_identical(report$columns_added, character())
  expect_identical(report$counts$would_insert, 1L)
  expect_identical(report$counts$would_update, 1L)

  out <- fd_upsert_read_table(con)
  expect_equal(out$id, 1:3)
  expect_equal(out$score, c(10, 21, 30))
  expect_equal(out$retired, c("old-a", "old-b", NA))
})

test_that("fd_upsert errors clearly when upsert target lacks key constraint", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  DBI::dbWriteTable(
    con,
    "features_tbl",
    data.frame(id = 1:2, score = c(10, 20)),
    overwrite = TRUE
  )

  expect_error(
    fd_upsert(
      con = con,
      features_df = data.frame(id = c(2L, 3L), score = c(21, 30)),
      feat_table_name = "features_tbl",
      key = "id",
      update_table = TRUE,
      verbose = FALSE
    ),
    "PRIMARY KEY or UNIQUE constraint"
  )
})

test_that("fd_upsert errors on missing feature columns when alter_table is FALSE", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  initial <- data.frame(id = 1:2, score = c(10, 20))
  DBI::dbWriteTable(con, "features_tbl", initial, overwrite = TRUE)

  expect_error(
    fd_upsert(
      con = con,
      features_df = data.frame(id = 3L, score = 30, label = "c"),
      feat_table_name = "features_tbl",
      key = "id",
      alter_table = FALSE,
      verbose = FALSE
    ),
    "missing columns"
  )

  expect_equal(DBI::dbListFields(con, "features_tbl"), names(initial))
  expect_equal(fd_upsert_read_table(con), initial)
})

test_that("fd_upsert chunks writes and reports pre-merge insert/update counts", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  fd_upsert_create_target(con, data.frame(id = c(1L, 3L), score = c(10, 30)))

  incoming <- data.frame(
    id = c(1L, 2L, 3L, 4L, 5L),
    score = c(11, 20, 31, 40, 50)
  )
  report <- fd_upsert(
    con = con,
    features_df = incoming,
    feat_table_name = "features_tbl",
    key = "id",
    chunk_size = 2L,
    verbose = FALSE
  )

  expect_identical(report$n_chunks, 3L)
  expect_identical(report$counts$would_insert, 3L)
  expect_identical(report$counts$would_update, 2L)
  expect_equal(
    report$chunk_details,
    data.frame(
      chunk = 1:3,
      n = c(2L, 2L, 1L),
      would_insert = c(1L, 1L, 1L),
      would_update = c(1L, 1L, 0L)
    )
  )
  expect_equal(fd_upsert_read_table(con)$score, c(11, 20, 31, 40, 50))
})

test_that("fd_upsert handles empty feature frames with a structured report", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  features <- data.frame(id = integer(), score = numeric())

  report <- fd_upsert(
    con = con,
    features_df = features,
    feat_table_name = "features_tbl",
    key = "id",
    create_table = "auto",
    verbose = FALSE
  )

  expect_s3_class(report, "fd_upsert_report")
  expect_true(report$table_created)
  expect_identical(report$n_rows, 0L)
  expect_identical(report$n_chunks, 0L)
  expect_identical(report$counts$would_insert, 0L)
  expect_identical(report$counts$would_update, 0L)
  expect_equal(DBI::dbListFields(con, "features_tbl"), c("id", "score"))
  expect_equal(nrow(DBI::dbReadTable(con, "features_tbl")), 0L)
})

test_that("fd_upsert can run for side effects only", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  features <- data.frame(id = 1:2, score = c(10, 20))
  ret <- fd_upsert(
    con = con,
    features_df = features,
    feat_table_name = "features_tbl",
    key = "id",
    create_table = TRUE,
    verbose = FALSE,
    return_report = FALSE
  )

  expect_true(ret)
  expect_equal(fd_upsert_read_table(con), features)
})

test_that("fd_upsert refuses to overwrite an existing table with create_table TRUE", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  initial <- data.frame(id = 1L, score = 10)
  DBI::dbWriteTable(con, "features_tbl", initial, overwrite = TRUE)

  expect_error(
    fd_upsert(
      con = con,
      features_df = data.frame(id = 2L, score = 20),
      feat_table_name = "features_tbl",
      key = "id",
      create_table = TRUE,
      verbose = FALSE
    ),
    "create_table=TRUE but table already exists"
  )
  expect_equal(fd_upsert_read_table(con), initial)
})

test_that("fd_upsert errors when table is missing and create_table is FALSE", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  expect_error(
    fd_upsert(
      con = con,
      features_df = data.frame(id = 1L, score = 10),
      feat_table_name = "features_tbl",
      key = "id",
      create_table = FALSE,
      verbose = FALSE
    ),
    "Table does not exist"
  )
  expect_false(DBI::dbExistsTable(con, "features_tbl"))
})

test_that("fd_upsert validates feature keys and scalar controls", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  expect_error(
    fd_upsert(con, data.frame(other_id = 1L), "features_tbl", key = "id", verbose = FALSE),
    "`key` must be a column"
  )
  expect_error(
    fd_upsert(con, data.frame(id = c(1L, 1L)), "features_tbl", key = "id", verbose = FALSE),
    "duplicates"
  )
  expect_error(
    fd_upsert(con, data.frame(id = c(1L, NA_integer_)), "features_tbl", key = "id", verbose = FALSE),
    "must not contain NAs"
  )
  expect_error(
    fd_upsert(con, data.frame(id = 1L), "features_tbl", key = "id", create_table = "yes", verbose = FALSE),
    "create_table must be TRUE/FALSE, or 'auto'"
  )
  expect_error(
    fd_upsert(con, data.frame(id = 1L), "features_tbl", key = "id", chunk_size = 0, verbose = FALSE),
    "`chunk_size` must be NULL or a positive number"
  )
  expect_error(
    fd_upsert(con, data.frame(id = 1L), "features_tbl", key = "id", update_table = NA, verbose = FALSE),
    "update_table must be TRUE/FALSE"
  )
})

test_that("fd_upsert normalizes factor columns before writing", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  features <- data.frame(
    id = 1:2,
    label = factor(c("a", "b"))
  )
  fd_upsert(
    con = con,
    features_df = features,
    feat_table_name = "features_tbl",
    key = "id",
    create_table = TRUE,
    verbose = FALSE
  )

  out <- fd_upsert_read_table(con)
  expect_type(out$label, "character")
  expect_equal(out$label, c("a", "b"))
})

test_that("fd_upsert rolls back schema changes and data writes on failure", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  initial <- data.frame(id = 1:2, score = c(10, 20))
  DBI::dbWriteTable(con, "features_tbl", initial, overwrite = TRUE)

  expect_error(
    fd_upsert(
      con = con,
      features_df = data.frame(
        id = c(2L, 3L),
        score = c(21, 30),
        label = c("b2", "c"),
        stringsAsFactors = FALSE
      ),
      feat_table_name = "features_tbl",
      key = "id",
      alter_table = TRUE,
      update_table = FALSE,
      verbose = FALSE
    ),
    "Conflicts detected"
  )

  expect_equal(DBI::dbListFields(con, "features_tbl"), names(initial))
  expect_equal(fd_upsert_read_table(con), initial)
})

test_that("fd_upsert emits useful progress messages when verbose", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  messages <- capture.output(
    report <- fd_upsert(
      con = con,
      features_df = data.frame(id = 1:2, score = c(10, 20)),
      feat_table_name = "features_tbl",
      key = "id",
      create_table = TRUE,
      verbose = TRUE
    ),
    type = "message"
  )

  expect_true(any(grepl("Creating table: features_tbl", messages, fixed = TRUE)))
  expect_true(any(grepl("Writing features to features_tbl", messages, fixed = TRUE)))
  expect_true(any(grepl("Done. would_insert=2, would_update=0", messages, fixed = TRUE)))
  expect_s3_class(report, "fd_upsert_report")
  expect_true(DBI::dbExistsTable(con, "features_tbl"))
})

fd_run_ns <- function(name) {
  getFromNamespace(name, "featdelta")
}

fd_run_seed_raw <- function(con, raw = fd_test_raw_data()) {
  DBI::dbWriteTable(con, "raw_tbl", raw, overwrite = TRUE)
  invisible(raw)
}

fd_run_defs <- function() {
  fd_define(
    ratio = x / y,
    total = x + y
  )
}

fd_run_expected_features <- function(raw) {
  data.frame(
    id = raw$id,
    ratio = raw$x / raw$y,
    total = raw$x + raw$y
  )
}

fd_run_read_features <- function(con) {
  DBI::dbGetQuery(con, "SELECT * FROM features_tbl ORDER BY id")
}

fd_run_create_features <- function(con, features) {
  fd_upsert(
    con = con,
    features_df = features,
    feat_table_name = "features_tbl",
    key = "id",
    create_table = TRUE,
    verbose = FALSE
  )
  invisible(TRUE)
}

test_that("fd_run performs an initial new_only run when the feature table is missing", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)
  expected <- fd_run_expected_features(raw)

  report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = fd_run_defs(),
    key = "id",
    feat_table_name = "features_tbl",
    fetch_mode = "new_only",
    preview_n = 2L,
    verbose = FALSE
  )

  expect_s3_class(report, "fd_run_report")
  expect_true(report$success)
  expect_identical(report$stage, "complete")
  expect_identical(report$key, "id")
  expect_identical(report$feat_table_name, "features_tbl")
  expect_identical(report$dialect, "sqlite")
  expect_identical(report$fetch$source, "initial_all")
  expect_false(report$fetch$table_exists)
  expect_identical(report$fetch$n_rows, 5L)
  expect_identical(report$compute$n_rows, 5L)
  expect_identical(report$compute$n_features, 2L)
  expect_equal(report$compute$feature_names, c("ratio", "total"))
  expect_true(all(report$compute$report$ok))
  expect_true(report$upsert$table_created)
  expect_identical(report$upsert$counts$would_insert, 5L)
  expect_identical(report$upsert$counts$would_update, 0L)
  expect_null(report$data)
  expect_equal(report$preview$raw, raw[1:2, c("id", "x", "y")])
  expect_s3_class(report$preview$features, "featdelta_features")
  expect_equal(report$preview$features, expected[1:2, ], ignore_attr = TRUE)
  expect_equal(fd_run_read_features(con), expected)
})

test_that("fd_run new_only fetches only rows absent from an existing feature table", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)
  expected <- fd_run_expected_features(raw)
  fd_run_create_features(con, expected[expected$id %in% c(1L, 3L), ])

  report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = fd_run_defs(),
    key = "id",
    feat_table_name = "features_tbl",
    fetch_mode = "new_only",
    verbose = FALSE
  )

  expect_true(report$success)
  expect_identical(report$fetch$source, "fd_fetch")
  expect_true(report$fetch$table_exists)
  expect_false(report$fetch$limit_applied)
  expect_identical(report$fetch$n_rows, 3L)
  expect_equal(report$fetch$fd_fetch$n_rows, 3L)
  expect_identical(report$compute$n_rows, 3L)
  expect_identical(report$upsert$counts$would_insert, 3L)
  expect_identical(report$upsert$counts$would_update, 0L)
  expect_equal(fd_run_read_features(con), expected)
})

test_that("fd_run new_only handles existing feature table with no new raw rows", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)
  expected <- fd_run_expected_features(raw)
  fd_run_create_features(con, expected)

  report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = fd_run_defs(),
    key = "id",
    feat_table_name = "features_tbl",
    fetch_mode = "new_only",
    verbose = FALSE
  )

  expect_true(report$success)
  expect_identical(report$fetch$source, "fd_fetch")
  expect_identical(report$fetch$n_rows, 0L)
  expect_identical(report$compute$n_rows, 0L)
  expect_identical(report$upsert$n_rows, 0L)
  expect_identical(report$upsert$counts$would_insert, 0L)
  expect_identical(report$upsert$counts$would_update, 0L)
  expect_equal(fd_run_read_features(con), expected)
})

test_that("fd_run fetch_mode all refreshes existing rows and inserts new rows", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)
  expected <- fd_run_expected_features(raw)
  stale <- expected[expected$id %in% c(1L, 3L), ]
  stale$ratio <- -1
  stale$total <- -1
  fd_run_create_features(con, stale)

  report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = fd_run_defs(),
    key = "id",
    feat_table_name = "features_tbl",
    fetch_mode = "all",
    verbose = FALSE
  )

  expect_true(report$success)
  expect_identical(report$fetch$source, "all")
  expect_identical(report$fetch$n_rows, 5L)
  expect_identical(report$upsert$counts$would_insert, 3L)
  expect_identical(report$upsert$counts$would_update, 2L)
  expect_equal(fd_run_read_features(con), expected)
})

test_that("fd_run documents feature-definition backfill behavior through fetch_mode", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)

  initial_defs <- fd_define(ratio = x / y)
  initial_features <- data.frame(id = raw$id, ratio = raw$x / raw$y)
  fd_run_create_features(con, initial_features)

  expanded_defs <- fd_define(
    ratio = x / y,
    total = x + y
  )

  new_only_report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = expanded_defs,
    key = "id",
    feat_table_name = "features_tbl",
    fetch_mode = "new_only",
    verbose = FALSE
  )

  new_only_out <- fd_run_read_features(con)
  expect_true(new_only_report$success)
  expect_identical(new_only_report$fetch$n_rows, 0L)
  expect_identical(new_only_report$upsert$columns_added, "total")
  expect_true(all(is.na(new_only_out$total)))

  all_report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = expanded_defs,
    key = "id",
    feat_table_name = "features_tbl",
    fetch_mode = "all",
    verbose = FALSE
  )

  all_out <- fd_run_read_features(con)
  expect_true(all_report$success)
  expect_identical(all_report$fetch$n_rows, 5L)
  expect_identical(all_report$upsert$counts$would_insert, 0L)
  expect_identical(all_report$upsert$counts$would_update, 5L)
  expect_equal(all_out$total, raw$x + raw$y)
})

test_that("fd_run applies fetch_limit after fetching and records return data", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)
  expected <- fd_run_expected_features(raw[1:2, ])

  report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = fd_run_defs(),
    key = "id",
    feat_table_name = "features_tbl",
    fetch_limit = 2L,
    return_data = "both",
    preview_n = 1L,
    verbose = FALSE
  )

  expect_true(report$success)
  expect_true(report$fetch$limit_applied)
  expect_identical(report$fetch$n_rows_before_limit, 5L)
  expect_identical(report$fetch$n_rows, 2L)
  expect_named(report$data, c("raw", "features"))
  expect_equal(report$data$raw, raw[1:2, c("id", "x", "y")])
  expect_s3_class(report$data$features, "featdelta_features")
  expect_equal(report$data$features, expected, ignore_attr = TRUE)
  expect_equal(report$preview$raw, raw[1, c("id", "x", "y"), drop = FALSE])
  expect_s3_class(report$preview$features, "featdelta_features")
  expect_equal(report$preview$features, expected[1, , drop = FALSE], ignore_attr = TRUE)
  expect_equal(fd_run_read_features(con), expected)
})

test_that("fd_run supports compute_envir overrides", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)

  defs_env <- new.env(parent = baseenv())
  defs_env$offset <- 10
  compute_env <- new.env(parent = baseenv())
  compute_env$offset <- 100
  defs <- fd_define(shifted = x + offset, envir = defs_env)

  report <- fd_run(
    con = con,
    sql = "SELECT id, x FROM raw_tbl ORDER BY id",
    defs = defs,
    key = "id",
    feat_table_name = "features_tbl",
    compute_envir = compute_env,
    return_data = "features",
    verbose = FALSE
  )

  expect_true(report$success)
  expect_equal(report$data$features$shifted, raw$x + 100)
  expect_equal(fd_run_read_features(con)$shifted, raw$x + 100)
})

test_that("fd_run captures compute failures when fail_fast is FALSE", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)
  defs <- fd_define(
    ratio = x / y,
    bad = missing_column + 1
  )

  report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = defs,
    key = "id",
    feat_table_name = "features_tbl",
    fail_fast = FALSE,
    return_data = "both",
    preview_n = 2L,
    verbose = FALSE
  )

  expect_s3_class(report, "fd_run_report")
  expect_false(report$success)
  expect_identical(report$stage, "compute")
  expect_identical(report$error$stage, "compute")
  expect_match(report$error$message, "one or more definition steps failed")
  expect_identical(report$compute$n_rows, 5L)
  expect_identical(report$compute$n_features, 1L)
  expect_equal(report$compute$report$ok, c(TRUE, FALSE))
  expect_named(report$data, c("raw", "features"))
  expect_equal(report$data$raw, raw[, c("id", "x", "y")])
  expect_equal(report$data$features$ratio, raw$x / raw$y)
  expect_null(report$upsert)
  expect_false(DBI::dbExistsTable(con, "features_tbl"))
})

test_that("fd_run raises stage errors when fail_fast is TRUE", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_run_seed_raw(con)
  defs <- fd_define(bad = missing_column + 1)

  expect_error(
    fd_run(
      con = con,
      sql = "SELECT id, x FROM raw_tbl ORDER BY id",
      defs = defs,
      key = "id",
      feat_table_name = "features_tbl",
      fail_fast = TRUE,
      verbose = FALSE
    ),
    "one or more definition steps failed",
    class = "featdelta_compute_error"
  )
})

test_that("fd_run captures upsert failures when fail_fast is FALSE", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)
  expected <- fd_run_expected_features(raw)
  fd_run_create_features(con, expected[expected$id %in% c(1L, 3L), ])

  report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = fd_run_defs(),
    key = "id",
    feat_table_name = "features_tbl",
    fetch_mode = "all",
    update_table = FALSE,
    fail_fast = FALSE,
    return_data = "features",
    verbose = FALSE
  )

  expect_false(report$success)
  expect_identical(report$stage, "upsert")
  expect_identical(report$error$stage, "upsert")
  expect_match(report$error$message, "Conflicts detected")
  expect_identical(report$compute$n_rows, 5L)
  expect_identical(report$compute$n_features, 2L)
  expect_named(report$data, "features")
  expect_s3_class(report$data$features, "featdelta_features")
  expect_equal(report$data$features, expected, ignore_attr = TRUE)
  expect_null(report$upsert)
  expect_equal(
    fd_run_read_features(con),
    expected[expected$id %in% c(1L, 3L), ],
    ignore_attr = TRUE
  )
})

test_that("fd_run validates controls and rejects unsupported dialects", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_run_seed_raw(con)
  defs <- fd_run_defs()

  expect_error(
    fd_run(con, "SELECT * FROM raw_tbl", defs, key = "id", feat_table_name = "features_tbl", fetch_limit = 0),
    "`fetch_limit` must be NULL or a positive number"
  )
  expect_error(
    fd_run(con, "SELECT * FROM raw_tbl", defs, key = "id", feat_table_name = "features_tbl", chunk_size = -1),
    "`chunk_size` must be NULL or a positive number"
  )
  expect_error(
    fd_run(con, "SELECT * FROM raw_tbl", defs, key = "id", feat_table_name = "features_tbl", preview_n = -1),
    "`preview_n` must be a non-negative number"
  )
  expect_error(
    fd_run(con, "SELECT * FROM raw_tbl", defs, key = "id", feat_table_name = "features_tbl", fetch_mode = "bad"),
    "'arg' should be one of"
  )
  expect_error(
    fd_run(con, "SELECT * FROM raw_tbl", defs, key = "id", feat_table_name = "features_tbl", return_data = "bad"),
    "'arg' should be one of"
  )
  expect_error(
    fd_run(con, "SELECT * FROM raw_tbl", defs, key = "id", feat_table_name = "features_tbl", dialect = "oracle"),
    "Supported: postgres, sqlite, mysql"
  )
})

test_that("fd_run_fetch_stage records source and limit metadata", {
  resolve_ctx <- fd_run_ns("resolve_ctx")
  fetch_stage <- fd_run_ns("fd_run_fetch_stage")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  raw <- fd_run_seed_raw(con)
  expected <- fd_run_expected_features(raw)
  fd_run_create_features(con, expected[expected$id %in% c(1L, 2L), ])
  ctx <- resolve_ctx(con, feat_table_name = "features_tbl", key = "id")

  stage <- fetch_stage(
    ctx = ctx,
    sql_clean = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    fetch_mode = "new_only",
    table_exists = TRUE,
    use_max_key = FALSE,
    fetch_limit = 2L,
    verbose = FALSE
  )

  expect_true(stage$ok)
  expect_equal(stage$data$id, c(3L, 4L))
  expect_identical(stage$info$source, "fd_fetch")
  expect_true(stage$info$limit_applied)
  expect_identical(stage$info$n_rows_before_limit, 3L)
  expect_identical(stage$info$n_rows, 2L)
  expect_type(stage$info$fd_fetch, "list")
})

test_that("fd_run report helpers select data and previews", {
  select_data <- fd_run_ns("fd_run_select_return_data")
  preview <- fd_run_ns("fd_run_preview")
  validate_positive <- fd_run_ns("fd_run_validate_optional_positive_int")
  validate_nonnegative <- fd_run_ns("fd_run_validate_nonnegative_int")

  raw <- fd_test_raw_data()
  features <- data.frame(id = raw$id, ratio = raw$x / raw$y)

  expect_null(select_data(raw, features, "none"))
  expect_named(select_data(raw, features, "raw"), "raw")
  expect_named(select_data(raw, features, "features"), "features")
  expect_named(select_data(raw, features, "both"), c("raw", "features"))
  expect_equal(preview(raw, 2L), raw[1:2, ])
  expect_null(preview(raw, 0L))
  expect_null(preview(NULL, 2L))
  expect_null(validate_positive(NULL, "x"))
  expect_identical(validate_positive(2.9, "x"), 2L)
  expect_error(validate_positive(0, "x"), "`x` must be NULL or a positive number")
  expect_identical(validate_nonnegative(0, "x"), 0L)
  expect_error(validate_nonnegative(-1, "x"), "`x` must be a non-negative number")
})

test_that("fd_run_stage_error and fd_run_build_report produce structured failures", {
  stage_error <- fd_run_ns("fd_run_stage_error")
  build_report <- fd_run_ns("fd_run_build_report")
  resolve_ctx <- fd_run_ns("resolve_ctx")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  ctx <- resolve_ctx(con, feat_table_name = "features_tbl", key = "id")
  err <- stage_error("fetch", simpleError("broken fetch"))

  expect_false(err$ok)
  expect_identical(err$stage, "fetch")
  expect_identical(err$message, "broken fetch")
  expect_true("simpleError" %in% err$class)

  report <- build_report(
    success = FALSE,
    stage = "fetch",
    started_at = Sys.time(),
    ctx = ctx,
    sql_clean = "SELECT * FROM raw_tbl",
    fetch_mode = "new_only",
    table_exists = FALSE,
    use_max_key = FALSE,
    fetch_limit = NULL,
    error = err,
    preview_n = 2L
  )

  expect_s3_class(report, "fd_run_report")
  expect_false(report$success)
  expect_identical(report$stage, "fetch")
  expect_identical(report$error$stage, "fetch")
  expect_identical(report$error$message, "broken fetch")
  expect_null(report$preview$raw)
  expect_null(report$preview$features)
})

test_that("print.fd_run_report returns input invisibly and shows summary fields", {
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)
  fd_run_seed_raw(con)

  report <- fd_run(
    con = con,
    sql = "SELECT id, x, y FROM raw_tbl ORDER BY id",
    defs = fd_run_defs(),
    key = "id",
    feat_table_name = "features_tbl",
    verbose = FALSE
  )

  printed <- capture.output(ret <- print(report))

  expect_identical(ret, report)
  expect_true(any(grepl("<fd_run_report>", printed, fixed = TRUE)))
  expect_true(any(grepl("success:   TRUE", printed, fixed = TRUE)))
  expect_true(any(grepl("stage:     complete", printed, fixed = TRUE)))
  expect_true(any(grepl("table:     features_tbl", printed, fixed = TRUE)))
  expect_true(any(grepl("dialect:   sqlite", printed, fixed = TRUE)))
  expect_true(any(grepl("upsert:    would_insert=5, would_update=0", printed, fixed = TRUE)))
})

test_that("fd_compute evaluates ordinary column definitions in order", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    ratio = x / y,
    doubled = ratio * 2,
    shifted = doubled + 1
  )

  out <- fd_compute(raw, defs, key = "id")

  expect_s3_class(out, "featdelta_features")
  expect_s3_class(out, "data.frame")
  expect_named(out, c("id", "ratio", "doubled", "shifted"))
  expect_identical(out$id, raw$id)
  expect_equal(out$ratio, raw$x / raw$y)
  expect_equal(out$doubled, (raw$x / raw$y) * 2)
  expect_equal(out$shifted, ((raw$x / raw$y) * 2) + 1)
  expect_identical(attr(out, "fd_key"), "id")
})

test_that("fd_compute recycles scalar column results to all input rows", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    one = 1,
    label = "constant"
  )

  out <- fd_compute(raw, defs, key = "id")

  expect_named(out, c("id", "one", "label"))
  expect_identical(out$one, rep(1, nrow(raw)))
  expect_identical(out$label, rep("constant", nrow(raw)))
})

test_that("fd_compute uses compute_envir as an evaluation override", {
  raw <- fd_test_raw_data()
  defs_env <- new.env(parent = baseenv())
  defs_env$offset <- 10

  compute_env <- new.env(parent = baseenv())
  compute_env$offset <- 100

  defs <- fd_define(
    shifted = x + offset,
    envir = defs_env
  )

  out_stored <- fd_compute(raw, defs, key = "id")
  out_override <- fd_compute(raw, defs, key = "id", compute_envir = compute_env)

  expect_equal(out_stored$shifted, raw$x + 10)
  expect_equal(out_override$shifted, raw$x + 100)
})

test_that("fd_compute returns a report when requested", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    ratio = x / y,
    doubled = ratio * 2
  )

  res <- fd_compute(raw, defs, key = "id", return_report = TRUE)

  expect_s3_class(res, "featdelta_compute_result")
  expect_named(res, c("data", "report"))
  expect_s3_class(res$data, "featdelta_features")
  expect_s3_class(res$report, "data.frame")
  expect_named(
    res$report,
    c(
      "step_name", "type", "mode", "ok", "error", "error_raw",
      "error_class", "hint", "time_sec", "text", "output_names"
    )
  )
  expect_equal(res$report$step_name, c("ratio", "doubled"))
  expect_equal(res$report$type, c("column", "column"))
  expect_true(all(res$report$ok))
  expect_equal(res$report$output_names, c("ratio", "doubled"))
  expect_identical(attr(res$data, "fd_report"), res$report)
})

test_that("fd_compute attaches a report when returning only data", {
  raw <- fd_test_raw_data()
  defs <- fd_define(ratio = x / y)

  out <- fd_compute(raw, defs, key = "id")
  report <- attr(out, "fd_report")

  expect_s3_class(report, "data.frame")
  expect_equal(report$step_name, "ratio")
  expect_true(report$ok)
  expect_equal(report$output_names, "ratio")
})

test_that("fd_compute handles zero-row input with scalar definitions", {
  raw <- fd_test_raw_data()[0, , drop = FALSE]
  defs <- fd_define(
    one = 1,
    ratio = x / y
  )

  out <- fd_compute(raw, defs, key = "id")

  expect_s3_class(out, "featdelta_features")
  expect_equal(nrow(out), 0L)
  expect_named(out, c("id", "one", "ratio"))
  expect_identical(out$id, integer())
  expect_identical(out$one, numeric())
  expect_identical(out$ratio, numeric())
})

test_that("fd_compute validates data and key inputs", {
  raw <- fd_test_raw_data()
  defs <- fd_define(ratio = x / y)

  expect_error(fd_compute(list(id = 1), defs, key = "id"), "`data` must be")
  expect_error(fd_compute(raw, defs, key = "missing_id"), "`key` must be a column")
  expect_error(fd_compute(transform(raw, id = c(1, 1, 3, 4, 5)), defs, key = "id"), "duplicates")
  expect_error(fd_compute(transform(raw, id = c(1, NA, 3, 4, 5)), defs, key = "id"), "must not contain NAs")
  expect_error(fd_compute(raw, list(), key = "id"), "`defs` must be")
  expect_error(fd_compute(raw, defs, key = ""), "`key` must be")
  expect_error(fd_compute(raw, defs, key = "id", compute_envir = list()), "`compute_envir`")
})

test_that("fd_compute strict mode raises a compute error with report and partial data", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    ratio = x / y,
    bad = missing_column + 1
  )

  err <- expect_error(
    fd_compute(raw, defs, key = "id", compute_strict = TRUE),
    class = "featdelta_compute_error"
  )

  report <- attr(err, "fd_report")
  partial <- attr(err, "fd_partial")

  expect_s3_class(report, "data.frame")
  expect_equal(report$step_name, c("ratio", "bad"))
  expect_equal(report$ok, c(TRUE, FALSE))
  expect_equal(report$error_class[[2]], "missing_object")
  expect_match(report$hint[[2]], "raw input column")

  expect_s3_class(partial, "featdelta_features")
  expect_named(partial, c("id", "ratio"))
  expect_equal(partial$ratio, raw$x / raw$y)
})

test_that("fd_compute non-strict mode records failed column steps as NA outputs", {
  raw <- fd_test_raw_data()
  defs <- fd_define(
    ratio = x / y,
    bad = missing_column + 1,
    after = ratio + 1
  )

  expect_warning(
    out <- fd_compute(raw, defs, key = "id", compute_strict = FALSE),
    "step 'bad' failed"
  )

  report <- attr(out, "fd_report")

  expect_named(out, c("id", "ratio", "bad", "after"))
  expect_equal(out$ratio, raw$x / raw$y)
  expect_true(all(is.na(out$bad)))
  expect_equal(out$after, (raw$x / raw$y) + 1)
  expect_equal(report$step_name, c("ratio", "bad", "after"))
  expect_equal(report$ok, c(TRUE, FALSE, TRUE))
  expect_equal(report$error_class[[2]], "missing_object")
})

test_that("fd_compute rejects unsupported and mis-sized column results", {
  raw <- fd_test_raw_data()

  expect_error(
    fd_compute(raw, fd_define(bad = list(1, 2, 3, 4, 5)), key = "id"),
    class = "featdelta_compute_error"
  )

  expect_error(
    fd_compute(raw, fd_define(bad = c(1, 2)), key = "id"),
    class = "featdelta_compute_error"
  )
})

fd_ns <- function(name) {
  getFromNamespace(name, "featdelta")
}

test_that("fd_compute_validate_defs_structure accepts valid column definitions", {
  validate_defs <- fd_ns("fd_compute_validate_defs_structure")

  defs <- fd_define(
    ratio = x / y,
    shifted = ratio + 1
  )

  expect_true(validate_defs(defs, key = "id"))
})

test_that("fd_compute_validate_defs_structure rejects malformed defs containers", {
  validate_defs <- fd_ns("fd_compute_validate_defs_structure")
  defs <- fd_define(ratio = x / y)

  empty_defs <- defs
  empty_defs$steps <- list()
  expect_error(validate_defs(empty_defs, key = "id"), "non-empty list")

  unnamed_defs <- defs
  names(unnamed_defs$steps) <- NULL
  expect_error(validate_defs(unnamed_defs, key = "id"), "named list")

  duplicate_defs <- defs
  duplicate_defs$steps <- c(defs$steps, defs$steps)
  names(duplicate_defs$steps) <- c("ratio", "ratio")
  expect_error(validate_defs(duplicate_defs, key = "id"), "duplicate step names")

  non_list_step <- defs
  non_list_step$steps$ratio <- "not a step"
  expect_error(validate_defs(non_list_step, key = "id"), "must be stored as a list")
})

test_that("fd_compute_validate_defs_structure rejects malformed column steps", {
  validate_defs <- fd_ns("fd_compute_validate_defs_structure")
  defs <- fd_define(ratio = x / y)

  bad_type <- defs
  bad_type$steps$ratio$type <- NA_character_
  expect_error(validate_defs(bad_type, key = "id"), "malformed `type`")

  bad_name <- defs
  bad_name$steps$ratio$step_name <- "other"
  expect_error(validate_defs(bad_name, key = "id"), "inconsistent `step_name`")

  key_collision <- defs
  key_collision$steps$ratio$output_names <- "id"
  expect_error(validate_defs(key_collision, key = "id"), "collide with `key`")

  bad_env <- defs
  bad_env$steps$ratio$env <- NULL
  expect_error(validate_defs(bad_env, key = "id"), "malformed `env`")

  bad_text <- defs
  bad_text$steps$ratio$text <- NA_character_
  expect_error(validate_defs(bad_text, key = "id"), "malformed `text`")

  bad_quo <- defs
  bad_quo$steps$ratio$quo <- quote(x / y)
  expect_error(validate_defs(bad_quo, key = "id"), "malformed `quo`")

  bad_output <- defs
  bad_output$steps$ratio$output_names <- c("ratio", "extra")
  expect_error(validate_defs(bad_output, key = "id"), "exactly one output name")
})

test_that("fd_compute_validate_defs_structure rejects malformed block steps", {
  validate_defs <- fd_ns("fd_compute_validate_defs_structure")
  defs <- fd_define(block = fd_block({ data.frame(value = x + y) }))

  bad_mode <- defs
  bad_mode$steps$block$mode <- NA_character_
  expect_error(validate_defs(bad_mode, key = "id"), "malformed `mode`")

  bad_fn <- defs
  bad_fn$steps$block$mode <- "function"
  bad_fn$steps$block$fn <- NULL
  expect_error(validate_defs(bad_fn, key = "id"), "malformed `fn`")

  bad_quo <- defs
  bad_quo$steps$block$quo <- quote(data.frame(value = x + y))
  expect_error(validate_defs(bad_quo, key = "id"), "malformed `quo`")

  unsupported_mode <- defs
  unsupported_mode$steps$block$mode <- "other"
  expect_error(validate_defs(unsupported_mode, key = "id"), "unsupported `mode`")

  duplicate_outputs <- defs
  duplicate_outputs$steps$block$output_names <- c("a", "a")
  expect_error(validate_defs(duplicate_outputs, key = "id"), "duplicate `output_names`")
})

test_that("fd_compute_eval_step reports unsupported step types", {
  eval_step <- fd_ns("fd_compute_eval_step")

  res <- eval_step(
    step = list(type = "unknown"),
    step_name = "bad",
    working_data = fd_test_raw_data(),
    mask = fd_test_raw_data(),
    n = 5L,
    key = "id",
    produced_names = character()
  )

  expect_false(res$ok)
  expect_match(res$error, "Unsupported step type")
  expect_equal(res$cols, list())
})

test_that("fd_compute_eval_column_step returns columns or diagnostics", {
  eval_column <- fd_ns("fd_compute_eval_column_step")
  raw <- fd_test_raw_data()
  step <- fd_define(ratio = x / y)$steps$ratio

  ok <- eval_column(
    step = step,
    step_name = "ratio",
    mask = raw,
    n = nrow(raw)
  )

  expect_true(ok$ok)
  expect_named(ok$cols, "ratio")
  expect_equal(ok$cols$ratio, raw$x / raw$y)

  bad_step <- fd_define(bad = missing_column + 1)$steps$bad
  bad <- eval_column(
    step = bad_step,
    step_name = "bad",
    mask = raw,
    n = nrow(raw),
    compute_strict = TRUE
  )

  expect_false(bad$ok)
  expect_named(bad$cols, "bad")
  expect_true(all(is.na(bad$cols$bad)))
  expect_equal(bad$error_class, "missing_object")
  expect_match(bad$hint, "raw input column")
})

test_that("fd_compute_eval_block_step returns block columns or declared fallbacks", {
  eval_block <- fd_ns("fd_compute_eval_block_step")
  raw <- fd_test_raw_data()
  step <- fd_define(
    block = fd_block({
      data.frame(sum_xy = x + y)
    })
  )$steps$block

  ok <- eval_block(
    step = step,
    step_name = "block",
    working_data = raw,
    mask = raw,
    n = nrow(raw),
    key = "id",
    produced_names = character()
  )

  expect_true(ok$ok)
  expect_named(ok$cols, "sum_xy")
  expect_equal(ok$cols$sum_xy, raw$x + raw$y)

  bad_step <- fd_define(
    block = fd_block(
      {
        missing_column + 1
      },
      expected_names = c("a", "b")
    )
  )$steps$block
  expect_warning(
    bad <- eval_block(
      step = bad_step,
      step_name = "block",
      working_data = raw,
      mask = raw,
      n = nrow(raw),
      key = "id",
      produced_names = character(),
      compute_strict = FALSE
    ),
    "step 'block' failed"
  )

  expect_false(bad$ok)
  expect_named(bad$cols, c("a", "b"))
  expect_true(all(is.na(bad$cols$a)))
  expect_true(all(is.na(bad$cols$b)))
  expect_equal(bad$error_class, "missing_object")
})

test_that("fd_compute_validate_column_result validates supported result shapes", {
  validate_column <- fd_ns("fd_compute_validate_column_result")

  scalar <- validate_column(1, n = 3L, name = "one")
  expect_true(scalar$ok)
  expect_identical(scalar$result, c(1, 1, 1))

  vector <- validate_column(c(1, 2, 3), n = 3L, name = "value")
  expect_true(vector$ok)
  expect_identical(vector$result, c(1, 2, 3))

  zero_scalar <- validate_column(1, n = 0L, name = "one")
  expect_true(zero_scalar$ok)
  expect_identical(zero_scalar$result, numeric())

  zero_empty <- validate_column(integer(), n = 0L, name = "empty")
  expect_true(zero_empty$ok)
  expect_identical(zero_empty$result, integer())
})

test_that("fd_compute_validate_column_result rejects invalid outputs", {
  validate_column <- fd_ns("fd_compute_validate_column_result")

  null_res <- validate_column(NULL, n = 3L, name = "bad")
  expect_false(null_res$ok)
  expect_match(null_res$error, "is NULL")

  list_res <- validate_column(list(1, 2, 3), n = 3L, name = "bad")
  expect_false(list_res$ok)
  expect_match(list_res$error, "unsupported type/class")

  short_res <- validate_column(c(1, 2), n = 3L, name = "bad")
  expect_false(short_res$ok)
  expect_match(short_res$error, "expected 3")

  zero_long <- validate_column(c(1, 2), n = 0L, name = "bad")
  expect_false(zero_long$ok)
  expect_match(zero_long$error, "expected 0")
})

test_that("fd_compute_validate_block_result validates expected-name completion", {
  validate_block <- fd_ns("fd_compute_validate_block_result")

  res <- validate_block(
    res = data.frame(b = c(2, 4), stringsAsFactors = FALSE),
    n = 2L,
    step_name = "block",
    key = "id",
    produced_names = "existing",
    expected_names = c("a", "b")
  )

  expect_true(res$ok)
  expect_named(res$result, c("a", "b"))
  expect_true(all(is.na(res$result$a)))
  expect_equal(res$result$b, c(2, 4))
})

test_that("fd_compute_validate_block_result rejects malformed block outputs", {
  validate_block <- fd_ns("fd_compute_validate_block_result")
  valid_args <- list(
    n = 2L,
    step_name = "block",
    key = "id",
    produced_names = character(),
    expected_names = NULL
  )

  expect_false(do.call(validate_block, c(list(res = NULL), valid_args))$ok)
  expect_false(do.call(validate_block, c(list(res = c(1, 2)), valid_args))$ok)
  expect_false(do.call(validate_block, c(list(res = data.frame(a = 1)), valid_args))$ok)

  unnamed <- data.frame(1:2)
  names(unnamed) <- ""
  unnamed_res <- do.call(validate_block, c(list(res = unnamed), valid_args))
  expect_false(unnamed_res$ok)
  expect_match(unnamed_res$error, "unnamed or empty-named")

  duplicate <- data.frame(a = 1:2, b = 3:4)
  names(duplicate) <- c("dup", "dup")
  duplicate_res <- do.call(validate_block, c(list(res = duplicate), valid_args))
  expect_false(duplicate_res$ok)
  expect_match(duplicate_res$error, "duplicate column names")

  key_collision <- do.call(
    validate_block,
    c(list(res = data.frame(id = 1:2)), valid_args)
  )
  expect_false(key_collision$ok)
  expect_match(key_collision$error, "colliding with `key`")

  earlier_collision <- validate_block(
    res = data.frame(existing = 1:2),
    n = 2L,
    step_name = "block",
    key = "id",
    produced_names = "existing"
  )
  expect_false(earlier_collision$ok)
  expect_match(earlier_collision$error, "already produced earlier")

  unexpected <- validate_block(
    res = data.frame(extra = 1:2),
    n = 2L,
    step_name = "block",
    key = "id",
    produced_names = character(),
    expected_names = "allowed"
  )
  expect_false(unexpected$ok)
  expect_match(unexpected$error, "unexpected column name")

  bad_column <- do.call(
    validate_block,
    c(list(res = data.frame(a = I(list(1, 2)))), valid_args)
  )
  expect_false(bad_column$ok)
  expect_match(bad_column$error, "unsupported type/class")
})

test_that("fd_compute_complete_expected_block_outputs fills missing columns in declared order", {
  complete_outputs <- fd_ns("fd_compute_complete_expected_block_outputs")

  out <- complete_outputs(
    res = data.frame(b = c(2, 4), stringsAsFactors = FALSE),
    expected_names = c("a", "b", "c"),
    n = 2L
  )

  expect_named(out, c("a", "b", "c"))
  expect_true(all(is.na(out$a)))
  expect_equal(out$b, c(2, 4))
  expect_true(all(is.na(out$c)))
})

test_that("fd_compute_call_block_function supports zero- and one-argument functions", {
  call_block <- fd_ns("fd_compute_call_block_function")
  raw <- fd_test_raw_data()

  no_arg <- function() {
    data.frame(a = 1:2)
  }
  one_arg <- function(data) {
    data.frame(a = data$x + data$y)
  }

  expect_equal(call_block(no_arg, raw), data.frame(a = 1:2))
  expect_equal(call_block(one_arg, raw), data.frame(a = raw$x + raw$y))
})

test_that("fd_compute_diagnose_error classifies common evaluation failures", {
  diagnose <- fd_ns("fd_compute_diagnose_error")

  missing_object <- diagnose(simpleError("object 'x' not found"))
  expect_identical(missing_object$class, "missing_object")
  expect_match(missing_object$message, "Possible reasons")
  expect_match(missing_object$hint, "raw input column")

  missing_function <- diagnose(simpleError("could not find function \"foo\""))
  expect_identical(missing_function$class, "missing_function")
  expect_match(missing_function$hint, "function is not available")

  missing_argument <- diagnose(simpleError("argument \"x\" is missing"))
  expect_identical(missing_argument$class, "missing_argument")
  expect_match(missing_argument$hint, "requires different arguments")

  other <- diagnose(simpleError("plain error"))
  expect_identical(other$class, "other_error")
  expect_identical(other$message, "plain error")
  expect_true(is.na(other$hint))
})

test_that("fd_compute_format_diagnostic_message handles malformed diagnostics", {
  format_diag <- fd_ns("fd_compute_format_diagnostic_message")

  expect_identical(
    format_diag(list(message = "specific message")),
    "specific message"
  )
  expect_identical(format_diag(NULL), "Unknown compute error.")
  expect_identical(format_diag(list()), "Unknown compute error.")
})

test_that("fd_compute_is_supported_result accepts vector-like columns only", {
  is_supported <- fd_ns("fd_compute_is_supported_result")

  expect_true(is_supported(c(1, 2)))
  expect_true(is_supported(c("a", "b")))
  expect_true(is_supported(as.Date("2026-05-02")))
  expect_true(is_supported(as.POSIXct("2026-05-02", tz = "UTC")))
  expect_true(is_supported(as.difftime(1, units = "days")))
  expect_false(is_supported(list(1, 2)))
  expect_false(is_supported(data.frame(a = 1)))
})

# Context layer
#
# This file defines the featdelta execution context used by all DB-facing
# functions. The context is the package's single source of truth for:
# - the DB connection,
# - selected target/raw table names,
# - the primary key,
# - the detected dialect,
# - stable derived identifiers such as parsed table ids and quoted names.
#
# Rule of thumb:
# - DB-facing functions such as fd_fetch(), fd_upsert(), and fd_run()
#   should call resolve_ctx() at the top and then use ctx fields.
# - Pure in-memory functions such as fd_define() and fd_compute()
#   should not use ctx at all.


#' Apply explicit overrides to an existing featdelta context
#'
#' Internal helper used by `resolve_ctx()` when the user supplies an existing
#' featdelta connection/context object rather than a plain DBI connection.
#'
#' This function updates only raw context fields such as `raw_table`,
#' `feat_table_name`, `key`, `dialect`, and metadata controls.
#' It does not build derived fields by itself; callers should pass the result
#' through `ctx_finalize()`.
#'
#' @param ctx A list-like featdelta context object.
#' @param raw_table Optional character scalar override.
#' @param feat_table_name Optional character scalar override.
#' @param key Optional character scalar override.
#' @param dialect Optional character scalar override.
#' @param meta_enabled Optional logical override.
#' @param meta_schema Optional list override.
#' @param warn Logical. If `TRUE`, emit warnings when an existing stored value
#'   is replaced by an explicit argument.
#'
#' @return The updated context object, not yet finalized.
#'
#' @family featdelta context helpers
#' @noRd
ctx_apply_overrides <- function(
    ctx,
    raw_table = NULL,
    feat_table_name = NULL,
    key = NULL,
    dialect = NULL,
    meta_enabled = NULL,
    meta_schema = NULL,
    warn = TRUE
) {
  overrides <- list(
    raw_table = raw_table,
    feat_table_name = feat_table_name,
    key = key,
    dialect = dialect,
    meta_enabled = if (is.null(meta_enabled)) NULL else isTRUE(meta_enabled),
    meta_schema = meta_schema
  )

  for (nm in names(overrides)) {
    value <- overrides[[nm]]
    if (is.null(value)) next

    if (!identical(ctx[[nm]], value)) {
      if (isTRUE(warn) && inherits(ctx, "featdelta_con")) {
        warning(
          "Overriding fd_con$", nm, " with `", nm, "` argument.",
          call. = FALSE
        )
      }
      ctx[[nm]] <- value
    }
  }

  ctx
}


#' Build stable derived fields for a featdelta context
#'
#' Computes reusable derived values from the raw context fields. This keeps
#' downstream DB-facing code from reparsing table names or requoting
#' identifiers repeatedly.
#'
#' Typical derived fields include:
#' \itemize{
#'   \item parsed raw table identifier (`raw_table_id`)
#'   \item quoted raw table name (`raw_table_q`)
#'   \item parsed feature table identifier (`feat_table_id`)
#'   \item quoted feature table name (`feat_table_q`)
#'   \item quoted key identifier (`key_q`)
#' }
#'
#' Only stable derived fields should be created here. Mutable state such as
#' current table existence or operation-specific settings should be computed
#' by the calling function instead.
#'
#' @param ctx A partially normalized featdelta context.
#'
#' @return The same context object with derived fields added.
#'
#' @family featdelta context helpers
#' @noRd
ctx_build_derived <- function(ctx) {
  ctx$dialect <- ctx$dialect %||% detect_backend(ctx$con)

  ctx$raw_table_id <- NULL
  ctx$raw_table_q  <- NULL
  if (!is.null(ctx$raw_table)) {
    ctx$raw_table_id <- parse_table_id(ctx$dialect, ctx$raw_table)
    ctx$raw_table_q  <- as.character(
      DBI::dbQuoteIdentifier(ctx$con, ctx$raw_table_id)
    )
  }

  ctx$feat_table_id <- NULL
  ctx$feat_table_q  <- NULL
  if (!is.null(ctx$feat_table_name)) {
    ctx$feat_table_id <- parse_table_id(ctx$dialect, ctx$feat_table_name)
    ctx$feat_table_q  <- as.character(
      DBI::dbQuoteIdentifier(ctx$con, ctx$feat_table_id)
    )
  }

  ctx$key_q <- NULL
  if (!is.null(ctx$key)) {
    ctx$key_q <- as.character(DBI::dbQuoteIdentifier(ctx$con, ctx$key))
  }

  ctx
}


#' Validate a featdelta execution context
#'
#' Performs structural and type validation on a context object after raw fields
#' and derived fields have been assembled.
#'
#' Validation covers:
#' \itemize{
#'   \item `ctx$con` is a live `DBIConnection`
#'   \item scalar character fields such as `raw_table`, `feat_table_name`,
#'         `key`, and `dialect`
#'   \item metadata controls
#'   \item internal consistency of derived fields such as `feat_table_id`,
#'         `feat_table_q`, and `key_q`
#' }
#'
#' This function validates the context object itself. It does not check
#' database-side state such as whether a table exists or whether a column is
#' present in a given table.
#'
#' @param ctx A featdelta context object.
#'
#' @return The validated context object.
#'
#' @family featdelta context helpers
#' @noRd
ctx_validate <- function(ctx) {
  stopifnot(is.list(ctx))

  if (!is_dbi_connection(ctx$con)) {
    stop("`ctx$con` must be a DBIConnection.", call. = FALSE)
  }
  if (!DBI::dbIsValid(ctx$con)) {
    stop("`ctx$con` is not a valid (open) DBIConnection.", call. = FALSE)
  }

  check_chr1 <- function(x, nm) {
    if (is.null(x)) return(invisible(NULL))
    if (!is.character(x) || length(x) != 1L || !nzchar(x)) {
      stop("`ctx$", nm, "` must be a non-empty character scalar.", call. = FALSE)
    }
  }

  check_chr1(ctx$raw_table, "raw_table")
  check_chr1(ctx$feat_table_name, "feat_table_name")
  check_chr1(ctx$key, "key")
  check_chr1(ctx$dialect, "dialect")

  if (!is.logical(ctx$meta_enabled) || length(ctx$meta_enabled) != 1L || is.na(ctx$meta_enabled)) {
    stop("`ctx$meta_enabled` must be TRUE/FALSE.", call. = FALSE)
  }

  if (!is.null(ctx$meta_schema) && !is.list(ctx$meta_schema)) {
    stop("`ctx$meta_schema` must be a list (or NULL).", call. = FALSE)
  }

  if (!is.null(ctx$raw_table) && is.null(ctx$raw_table_id)) {
    stop("Internal ctx error: `raw_table_id` was not built.", call. = FALSE)
  }
  if (!is.null(ctx$feat_table_name) && is.null(ctx$feat_table_id)) {
    stop("Internal ctx error: `feat_table_id` was not built.", call. = FALSE)
  }
  if (!is.null(ctx$feat_table_name) &&
      (!is.character(ctx$feat_table_q) || length(ctx$feat_table_q) != 1L)) {
    stop("Internal ctx error: `feat_table_q` must be a single quoted identifier.", call. = FALSE)
  }
  if (!is.null(ctx$key) &&
      (!is.character(ctx$key_q) || length(ctx$key_q) != 1L)) {
    stop("Internal ctx error: `key_q` must be a single quoted identifier.", call. = FALSE)
  }

  ctx
}


#' Finalize a featdelta context
#'
#' @noRd
ctx_finalize <- function(ctx) {
  class(ctx) <- unique(c(class(ctx), "featdelta_ctx", "list"))
  ctx <- ctx_build_derived(ctx)
  ctx_validate(ctx)
}


#' Construct a featdelta context from explicit arguments
#'
#' Low-level internal constructor used when the caller supplies a plain
#' `DBIConnection` and explicit featdelta settings rather than an existing
#' featdelta context object.
#'
#' This constructor records raw context fields and then finalizes the context
#' via `ctx_finalize()`.
#'
#' @param con A live `DBIConnection`.
#' @param raw_table Optional character scalar naming the raw/source table.
#' @param feat_table_name Optional character scalar naming the feature table.
#' @param key Optional character scalar naming the primary key column.
#' @param dialect Optional character scalar. If `NULL`, the dialect is
#'   inferred from `con`.
#' @param meta_schema Optional list storing metadata layout/configuration.
#' @param meta_enabled Logical. Whether featdelta metadata tracking is enabled
#'   for this context.
#' @param ... Reserved for future internal extensions.
#'
#' @return A finalized context object of class `"featdelta_ctx"`.
#'
#' @family featdelta context helpers
#' @noRd
make_ctx_from_args <- function(
    con,
    raw_table = NULL,
    feat_table_name = NULL,
    key = NULL,
    dialect = NULL,
    meta_schema = NULL,
    meta_enabled = FALSE,
    ...
) {
  dots <- list(...)
  if (length(dots)) {
    # reserved for future internal fields
  }

  ctx <- list(
    con = con,
    dialect = dialect,
    raw_table = raw_table,
    feat_table_name = feat_table_name,
    key = key,
    meta_enabled = isTRUE(meta_enabled),
    meta_schema = meta_schema
  )

  ctx_finalize(ctx)
}


#' Resolve a featdelta execution context
#'
#' Central internal entry point for DB-facing featdelta functions.
#' `resolve_ctx()` normalizes either a plain `DBIConnection` or a previously
#' constructed featdelta connection/context object into a validated
#' `featdelta_ctx`.
#'
#' This function is intended to be called at the beginning of exported
#' DB-facing functions such as `fd_fetch()`, `fd_upsert()`, and `fd_run()`.
#' The returned context should then be used as the single source of truth for
#' database identity fields instead of reparsing table names or requoting
#' identifiers in downstream code.
#'
#' Context resolution has two stages:
#' 1. normalize raw inputs and explicit overrides;
#' 2. compute stable derived fields (for example parsed table identifiers and
#'    quoted names) and validate the final object.
#'
#' Pure in-memory functions such as `fd_define()` and `fd_compute()` do not
#' use this function.
#'
#' @param con_or_fd Either a live `DBIConnection` or a featdelta connection
#'   object such as one produced by `fd_connect()`.
#' @param ... Reserved for future internal extensions passed through to
#'   `make_ctx_from_args()`.
#' @param raw_table Optional character scalar overriding the raw/source table
#'   name stored in the context.
#' @param feat_table_name Optional character scalar overriding the feature
#'   table name stored in the context.
#' @param key Optional character scalar overriding the primary key stored in
#'   the context.
#' @param dialect Optional character scalar overriding the detected dialect.
#'   If `NULL`, the dialect is inferred from `con`.
#' @param meta_enabled Optional logical overriding metadata enablement.
#' @param meta_schema Optional list overriding metadata schema settings.
#'
#' @return A validated context object of class `"featdelta_ctx"`. Depending on
#'   the supplied fields, the object may contain:
#'   \itemize{
#'     \item `con`
#'     \item `dialect`
#'     \item `raw_table`, `raw_table_id`, `raw_table_q`
#'     \item `feat_table_name`, `feat_table_id`, `feat_table_q`
#'     \item `key`, `key_q`
#'     \item `meta_enabled`, `meta_schema`
#'   }
#'
#' @details
#' Explicit arguments take precedence over stored values in a
#' `"featdelta_con"` object. When this happens, the context is rebuilt so that
#' all derived fields remain consistent with the overridden raw fields.
#'
#' The context stores only stable derived values. Operation-specific or
#' mutable state such as whether a table currently exists, chunk sizes, or
#' cleaned SQL strings should not be stored in the context.
#'
#' @family featdelta context helpers
#' @noRd
resolve_ctx <- function(
    con_or_fd,
    ...,
    raw_table = NULL,
    feat_table_name = NULL,
    key = NULL,
    dialect = NULL,
    meta_enabled = NULL,
    meta_schema = NULL
) {
  if (is_featdelta_con(con_or_fd)) {
    ctx <- ctx_apply_overrides(
      ctx = con_or_fd,
      raw_table = raw_table,
      feat_table_name = feat_table_name,
      key = key,
      dialect = dialect,
      meta_enabled = meta_enabled,
      meta_schema = meta_schema,
      warn = TRUE
    )
    return(ctx_finalize(ctx))
  }

  if (is_dbi_connection(con_or_fd)) {
    return(make_ctx_from_args(
      con = con_or_fd,
      raw_table = raw_table,
      feat_table_name = feat_table_name,
      key = key,
      dialect = dialect,
      meta_enabled = isTRUE(meta_enabled),
      meta_schema = meta_schema,
      ...
    ))
  }

  stop("Expected a DBIConnection or a featdelta_con.", call. = FALSE)
}

ctx_ns <- function(name) {
  getFromNamespace(name, "featdelta")
}

test_that("null-coalescing and SQL cleanup utilities behave predictably", {
  `%||%` <- ctx_ns("%||%")
  clean_sql <- ctx_ns("clean_sql")

  expect_identical(NULL %||% "fallback", "fallback")
  expect_identical(FALSE %||% TRUE, FALSE)
  expect_identical(character() %||% "fallback", character())

  expect_identical(clean_sql("  SELECT * FROM raw_tbl;  "), "SELECT * FROM raw_tbl")
  expect_identical(clean_sql("WITH x AS (SELECT 1) SELECT * FROM x"), "WITH x AS (SELECT 1) SELECT * FROM x")
})

test_that("parse_table_id validates dialect-specific table names", {
  parse_table_id <- ctx_ns("parse_table_id")

  expect_identical(parse_table_id("sqlite", "features_tbl"), "features_tbl")
  expect_error(
    parse_table_id("sqlite", "main.features_tbl"),
    "SQLite does not support schema-qualified"
  )

  id <- parse_table_id("postgres", "public.features_tbl")
  expect_true(methods::is(id, "Id"))
  expect_equal(id@name, c(schema = "public", table = "features_tbl"))

  mysql_id <- parse_table_id("mysql", "analytics.features_tbl")
  expect_true(methods::is(mysql_id, "Id"))
  expect_equal(mysql_id@name, c(schema = "analytics", table = "features_tbl"))

  expect_error(
    parse_table_id("postgres", "too.many.parts"),
    "must be 'table' or 'schema.table'"
  )
})

test_that("database utility helpers recognize SQLite connections", {
  is_dbi_connection <- ctx_ns("is_dbi_connection")
  is_featdelta_con <- ctx_ns("is_featdelta_con")
  quote_table <- ctx_ns("quote_table")
  detect_backend <- ctx_ns("detect_backend")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  expect_true(is_dbi_connection(con))
  expect_false(is_featdelta_con(con))
  expect_identical(detect_backend(con), "sqlite")
  expect_identical(quote_table(con, "features_tbl", "sqlite"), "`features_tbl`")
  expect_error(
    quote_table(con, "main.features_tbl", "sqlite"),
    "SQLite does not support schema-qualified"
  )
})

test_that("detect_backend rejects non-DBI and closed connections", {
  detect_backend <- ctx_ns("detect_backend")

  expect_error(detect_backend(list()), "must be a DBIConnection")

  con <- fd_test_sqlite_con()
  DBI::dbDisconnect(con)

  expect_error(detect_backend(con), "not a valid")
})

test_that("make_ctx_from_args builds finalized context fields", {
  make_ctx_from_args <- ctx_ns("make_ctx_from_args")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  ctx <- make_ctx_from_args(
    con = con,
    raw_table = "raw_tbl",
    feat_table_name = "features_tbl",
    key = "id",
    meta_enabled = TRUE,
    meta_schema = list(version = 1)
  )

  expect_s3_class(ctx, "featdelta_ctx")
  expect_identical(ctx$con, con)
  expect_identical(ctx$dialect, "sqlite")
  expect_identical(ctx$raw_table, "raw_tbl")
  expect_identical(ctx$feat_table_name, "features_tbl")
  expect_identical(ctx$key, "id")
  expect_true(ctx$meta_enabled)
  expect_equal(ctx$meta_schema, list(version = 1))
  expect_identical(ctx$raw_table_id, "raw_tbl")
  expect_identical(ctx$feat_table_id, "features_tbl")
  expect_identical(ctx$raw_table_q, "`raw_tbl`")
  expect_identical(ctx$feat_table_q, "`features_tbl`")
  expect_identical(ctx$key_q, "`id`")
})

test_that("resolve_ctx accepts plain DBI connections and featdelta contexts", {
  resolve_ctx <- ctx_ns("resolve_ctx")
  is_featdelta_con <- ctx_ns("is_featdelta_con")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  ctx <- resolve_ctx(
    con,
    feat_table_name = "features_tbl",
    key = "id"
  )

  expect_s3_class(ctx, "featdelta_ctx")
  expect_false(is_featdelta_con(ctx))
  expect_identical(ctx$feat_table_name, "features_tbl")
  expect_identical(ctx$key, "id")

  fd_con <- fd_connect(
    RSQLite::SQLite(),
    ":memory:",
    raw_table = "raw_tbl",
    feat_table_name = "features_tbl",
    key = "id"
  )
  on.exit(fd_test_disconnect(fd_con$con), add = TRUE)

  expect_true(is_featdelta_con(fd_con))

  expect_warning(
    overridden <- resolve_ctx(fd_con, feat_table_name = "features_new"),
    "Overriding fd_con\\$feat_table_name"
  )
  expect_s3_class(overridden, "featdelta_ctx")
  expect_s3_class(overridden, "featdelta_con")
  expect_identical(overridden$feat_table_name, "features_new")
  expect_identical(overridden$feat_table_q, "`features_new`")
  expect_identical(overridden$key, "id")
  expect_identical(overridden$key_q, "`id`")

  expect_error(resolve_ctx(list()), "Expected a DBIConnection or a featdelta_con")
})

test_that("ctx_apply_overrides updates raw fields and warns for featdelta_con objects", {
  ctx_apply_overrides <- ctx_ns("ctx_apply_overrides")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  fd_con <- fd_connect(
    RSQLite::SQLite(),
    ":memory:",
    feat_table_name = "features_tbl",
    key = "id"
  )
  on.exit(fd_test_disconnect(fd_con$con), add = TRUE)

  expect_warning(
    out <- ctx_apply_overrides(fd_con, feat_table_name = "features_new"),
    "Overriding fd_con\\$feat_table_name"
  )
  expect_identical(out$feat_table_name, "features_new")
  expect_identical(out$key, "id")

  plain <- list(con = con, feat_table_name = "features_tbl", key = "id")
  expect_silent(
    out_plain <- ctx_apply_overrides(
      plain,
      feat_table_name = "features_new",
      key = "row_id",
      warn = TRUE
    )
  )
  expect_identical(out_plain$feat_table_name, "features_new")
  expect_identical(out_plain$key, "row_id")
})

test_that("ctx_build_derived, ctx_validate, and ctx_finalize enforce context invariants", {
  ctx_build_derived <- ctx_ns("ctx_build_derived")
  ctx_validate <- ctx_ns("ctx_validate")
  ctx_finalize <- ctx_ns("ctx_finalize")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  raw_ctx <- list(
    con = con,
    dialect = NULL,
    raw_table = "raw_tbl",
    feat_table_name = "features_tbl",
    key = "id",
    meta_enabled = FALSE,
    meta_schema = NULL
  )

  derived <- ctx_build_derived(raw_ctx)
  expect_identical(derived$dialect, "sqlite")
  expect_identical(derived$raw_table_q, "`raw_tbl`")
  expect_identical(derived$feat_table_q, "`features_tbl`")
  expect_identical(derived$key_q, "`id`")

  expect_s3_class(ctx_finalize(raw_ctx), "featdelta_ctx")
  expect_type(ctx_validate(derived), "list")

  bad_meta <- derived
  bad_meta$meta_enabled <- NA
  expect_error(ctx_validate(bad_meta), "`ctx\\$meta_enabled`")

  bad_key <- derived
  bad_key$key <- ""
  expect_error(ctx_validate(bad_key), "`ctx\\$key`")

  missing_derived <- derived
  missing_derived$feat_table_id <- NULL
  expect_error(ctx_validate(missing_derived), "`feat_table_id` was not built")
})

test_that("fd_connect creates featdelta_con objects and handles deprecated raw_table_name", {
  testthat::skip_if_not_installed("RSQLite")

  fd_con <- fd_connect(
    RSQLite::SQLite(),
    ":memory:",
    raw_table = "raw_tbl",
    feat_table_name = "features_tbl",
    key = "id",
    meta_enabled = TRUE
  )
  on.exit(fd_test_disconnect(fd_con$con), add = TRUE)

  expect_s3_class(fd_con, "featdelta_con")
  expect_s3_class(fd_con, "featdelta_ctx")
  expect_identical(fd_con$dialect, "sqlite")
  expect_identical(fd_con$raw_table, "raw_tbl")
  expect_identical(fd_con$feat_table_name, "features_tbl")
  expect_identical(fd_con$key, "id")
  expect_true(fd_con$meta_enabled)

  expect_warning(
    aliased <- fd_connect(
      RSQLite::SQLite(),
      ":memory:",
      raw_table_name = "raw_tbl"
    ),
    "deprecated"
  )
  on.exit(fd_test_disconnect(aliased$con), add = TRUE)
  expect_identical(aliased$raw_table, "raw_tbl")

  expect_error(
    fd_connect(
      RSQLite::SQLite(),
      ":memory:",
      raw_table = "raw_tbl",
      raw_table_name = "other_raw"
    ),
    "Use only one"
  )
})

test_that("print.featdelta_con returns input invisibly and shows core context fields", {
  testthat::skip_if_not_installed("RSQLite")

  fd_con <- fd_connect(
    RSQLite::SQLite(),
    ":memory:",
    raw_table = "raw_tbl",
    feat_table_name = "features_tbl",
    key = "id"
  )
  on.exit(fd_test_disconnect(fd_con$con), add = TRUE)

  printed <- capture.output(ret <- print(fd_con))

  expect_identical(ret, fd_con)
  expect_true(any(grepl("<featdelta_con>", printed, fixed = TRUE)))
  expect_true(any(grepl("dialect:   sqlite", printed, fixed = TRUE)))
  expect_true(any(grepl("raw_table: raw_tbl", printed, fixed = TRUE)))
  expect_true(any(grepl("feat_table_name:features_tbl", printed, fixed = TRUE)))
  expect_true(any(grepl("key:       id", printed, fixed = TRUE)))
})

test_that("validate_general_args accepts valid common argument combinations", {
  validate_general_args <- ctx_ns("validate_general_args")

  raw <- fd_test_raw_data()
  features <- fd_test_features_data(c(1L, 2L))
  defs <- fd_define(ratio = x / y)
  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  expect_true(validate_general_args(
    data = raw,
    con = con,
    sql = "SELECT * FROM raw_tbl",
    key = "id",
    defs = defs,
    feat_table_name = "features_tbl",
    features_df = features,
    compute_envir = new.env(parent = emptyenv()),
    logicals = list(verbose = FALSE),
    logicals_with_text = list(create_table = "auto")
  ))

  expect_true(validate_general_args(sql = "WITH x AS (SELECT 1) SELECT * FROM x"))
})

test_that("validate_general_args rejects invalid data, SQL, defs, and scalar controls", {
  validate_general_args <- ctx_ns("validate_general_args")
  raw <- fd_test_raw_data()
  defs <- fd_define(ratio = x / y)

  expect_error(validate_general_args(data = list(id = 1)), "`data` must be")
  expect_error(validate_general_args(defs = list()), "`defs` must be")
  expect_error(validate_general_args(compute_envir = list()), "`compute_envir`")
  expect_error(validate_general_args(sql = ""), "`sql` must be")
  expect_error(validate_general_args(sql = "DELETE FROM raw_tbl"), "SELECT query")
  expect_error(validate_general_args(feat_table_name = ""), "`feat_table_name`")
  expect_error(validate_general_args(key = ""), "`key`")
  expect_error(validate_general_args(features_df = list(id = 1)), "`features_df`")
  expect_error(validate_general_args(data = raw, key = "missing"), "`key` must be a column")
  expect_error(
    validate_general_args(data = transform(raw, id = c(1, 1, 3, 4, 5)), key = "id"),
    "duplicates"
  )
  expect_error(
    validate_general_args(data = transform(raw, id = c(1, NA, 3, 4, 5)), key = "id"),
    "must not contain NAs"
  )
  expect_error(
    validate_general_args(
      features_df = data.frame(id = c(1, 1), value = 1:2),
      key = "id"
    ),
    "duplicates"
  )
  expect_error(
    validate_general_args(logicals = list(verbose = NA)),
    "verbose must be TRUE/FALSE"
  )
  expect_error(
    validate_general_args(logicals_with_text = list(create_table = "yes")),
    "create_table must be TRUE/FALSE, or 'auto'"
  )
  expect_error(
    validate_general_args(logicals_with_text = list(create_table = NA)),
    "create_table must be TRUE/FALSE, or 'auto'"
  )

  expect_true(validate_general_args(defs = defs))
})

#' Fetch source rows that are not yet present in the features table
#'
#' `fd_fetch()` executes a user-supplied SQL `SELECT` query against a database and
#' returns **only** those rows whose primary key (`key`) is **not present** in an
#' existing features table (`feat_table_name`).
#'
#' This function is intentionally not a general-purpose query runner. It always
#' applies a "not yet processed" filter against `feat_table_name` and errors if that
#' cannot be done honestly (e.g., missing tables/columns).
#'
#' ## Important assumption
#' `fd_fetch()` assumes that:
#' - the `sql` query is executed on the **same database** (same `con`) where
#'   `feat_table_name` is stored, and
#' - the returned `key` values are comparable to `feat_table_name.key`.
#'
#' Cross-database fetching (e.g., pulling data from one database and comparing to
#' a features table stored in another database) is not supported, because the
#' "not present in `feat_table_name`" filter must be evaluated by the database engine
#' in a single query context.
#'
#' @param con A live `DBIConnection`, created with [DBI::dbConnect()].
#' @param sql A single SQL string (typically a `SELECT`) that defines the **source
#'   dataset** you want to process into features. In other words, it should return
#'   the rows you would normally feed into your feature creation pipeline.
#'   The query may join multiple tables and apply filters.
#'
#'   The query must be usable as a derived table: `FROM (<sql>) AS r`.
#'   Trailing semicolons are tolerated and removed.
#' @param key Name of the primary key column (character scalar). Must exist in
#'   both the result of `sql` and in `feat_table_name`. This is the identifier used to
#'   decide whether a source row has already been processed into the features table.
#' @param feat_table_name Name of the existing features table in the database
#'   (character scalar). Must exist. This table is used only to identify which
#'   `key` values are already present.
#' @param use_max_key Logical. If `TRUE`, `fd_fetch()` may reduce join work by
#'   first computing `MAX(feat_table_name.key)` and selecting rows with `key > max_key`
#'   (guaranteed not in `feat_table_name`). If `FALSE`, `fd_fetch()` selects all the rows
#'   with any key no in `feat_table_name.key` which may be computationally more expensive.
#' @param verbose Logical. If `TRUE`, prints the executed SQL.
#'
#' @return A `data.frame` containing only rows returned by `sql` whose `key` is not
#' present in `feat_table_name`. The result includes an attribute `attr(x, "fd_fetch")`
#' (a list) with metadata such as `key`, `feat_table_name`, `use_max_key`, `max_key`
#' (if computed), `executed_sql`, and `n_rows`.
#'
#' @examples
#' if (requireNamespace("RSQLite", quietly = TRUE)) {
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' on.exit(DBI::dbDisconnect(con), add = TRUE)
#'
#' DBI::dbExecute(con, "CREATE TABLE raw (id INTEGER, x INTEGER)")
#' DBI::dbExecute(con, "CREATE TABLE r_variables_table (id INTEGER)")
#' DBI::dbExecute(con, "INSERT INTO raw (id, x) VALUES (1,10), (2,20), (3,30), (4,40), (5,50)")
#' DBI::dbExecute(con, "INSERT INTO r_variables_table (id) VALUES (1), (2), (4)")
#'
#' # Returns ids 3 and 5 only (rows not yet present in the features table)
#' new_rows <- fd_fetch(
#'   con = con,
#'   sql = "SELECT * FROM raw",
#'   key = "id",
#'   feat_table_name = "r_variables_table"
#' )
#' }
#'
#' @export
fd_fetch <- function(
    con,
    sql,
    key,
    feat_table_name,
    use_max_key = FALSE,
    verbose = FALSE
) {

  sql_clean <- clean_sql(sql)

  ctx <- resolve_ctx(
    con_or_fd = con,
    feat_table_name = feat_table_name,
    key = key
  )

  validate_general_args(
    con = ctx$con,
    sql = sql_clean,
    key = ctx$key,
    feat_table_name = ctx$feat_table_name,
    logicals = list(
      use_max_key = use_max_key,
      verbose = verbose
    )
  )

  if (!DBI::dbExistsTable(ctx$con, ctx$feat_table_id)) {
    stop("`feat_table_name` does not exist in the database: ", ctx$feat_table_name)
  }

  feat_fields <- DBI::dbListFields(ctx$con, ctx$feat_table_id)
  if (!(ctx$key %in% feat_fields)) {
    stop("`feat_table_name` does not contain the `key` column '", ctx$key, "'.")
  }

  probe_sql <- paste0("SELECT * FROM (", sql_clean, ") AS fd_raw LIMIT 0")
  probe_df <- DBI::dbGetQuery(ctx$con, probe_sql)
  if (!(ctx$key %in% names(probe_df))) {
    stop("`sql` does not return a column named '", ctx$key, "'.")
  }

  max_key <- NA

  if (!use_max_key) {
    executed_sql <- paste0(
      "SELECT r.* FROM (", sql_clean, ") AS r ",
      "LEFT JOIN ", ctx$feat_table_q, " AS f ",
      "ON r.", ctx$key_q, " = f.", ctx$key_q, " ",
      "WHERE f.", ctx$key_q, " IS NULL"
    )
  } else {
    max_sql <- paste0(
      "SELECT MAX(", ctx$key_q, ") AS max_key FROM ", ctx$feat_table_q
    )
    max_df <- DBI::dbGetQuery(ctx$con, max_sql)
    max_key <- max_df$max_key[[1]]

    if (is.null(max_key) || is.na(max_key)) {
      executed_sql <- paste0("SELECT * FROM (", sql_clean, ") AS r")
    } else {
      q_max <- as.character(DBI::dbQuoteLiteral(ctx$con, max_key))
      executed_sql <- paste0(
        "SELECT * FROM (", sql_clean, ") AS r ",
        "WHERE r.", ctx$key_q, " > ", q_max
      )
    }
  }

  if (isTRUE(verbose)) {
    message("fd_fetch SQL:\n", executed_sql)
  }

  out <- DBI::dbGetQuery(ctx$con, executed_sql)

  attr(out, "fd_fetch") <- list(
    key = ctx$key,
    feat_table_name = ctx$feat_table_name,
    use_max_key = use_max_key,
    max_key = max_key,
    sql = sql_clean,
    executed_sql = executed_sql,
    n_rows = nrow(out)
  )

  out
}

#' Run the featdelta incremental feature pipeline
#'
#' Executes the standard featdelta pipeline:
#'
#' `fd_fetch()` -> `fd_compute()` -> `fd_upsert()`
#'
#' `fd_run()` is the orchestration entry point that connects database fetching,
#' in-memory computation, and database upsert into a single incremental run.
#'
#' @param con A live `DBI` connection.
#' @param sql SQL query used to fetch raw rows.
#' @param defs A `featdelta_defs` object created by `fd_define()`.
#' @param key Character scalar naming the key column shared by raw data and
#'   feature table.
#' @param feat_table_name Character scalar naming the target feature table.
#' @param verbose Logical. If `TRUE`, print progress messages.
#' @param fetch_mode Fetch mode. `"new_only"` fetches rows whose key is not yet
#'   present in `feat_table_name`. `"all"` fetches all rows returned by `sql`,
#'   allowing explicit refresh/backfill runs.
#' @param use_max_key Logical. Passed to `fd_fetch()` when `fetch_mode =
#'   "new_only"` and the feature table already exists.
#' @param fetch_limit Optional positive row limit applied after fetching. This
#'   is intended for previews and small dry development runs, not as a SQL
#'   optimizer.
#' @param compute_strict Logical strictness flag passed to `fd_compute()`.
#' @param compute_envir Optional environment override passed to `fd_compute()`.
#' @param create_table Logical or `"auto"`. Passed to `fd_upsert()`.
#' @param alter_table Logical. Passed to `fd_upsert()`.
#' @param update_table Logical. Passed to `fd_upsert()`.
#' @param dialect Optional dialect override. Supported values are `"postgres"`,
#'   `"sqlite"`, and `"mysql"`.
#' @param chunk_size Optional chunk size. Passed to `fd_upsert()`.
#' @param fail_fast Logical. If `TRUE`, stage errors are raised immediately. If
#'   `FALSE`, errors from fetch, compute, or upsert are captured in the returned
#'   `fd_run_report`.
#' @param return_data Controls whether raw and/or computed data should be
#'   included in the returned run report.
#' @param preview_n Non-negative number of rows to include in report previews.
#' @param ... Reserved for future context options passed to `resolve_ctx()`.
#'
#' @return An object of class `"fd_run_report"` containing stage summaries,
#'   timings, row counts, the compute report, and the upsert report. Depending
#'   on `return_data`, it may also contain raw and/or computed feature data.
#'
#' @details
#' `fetch_mode = "new_only"` is the default incremental mode. If the feature
#' table already exists, `fd_fetch()` returns only rows whose key is missing
#' from the feature table. If the feature table does not exist yet, all rows
#' returned by `sql` are fetched because no rows have been processed. This mode
#' is key-based: it does not recompute existing feature-table rows just because
#' feature definitions changed or new feature definitions were added.
#'
#' `fetch_mode = "all"` is an explicit refresh/backfill mode. It recomputes all
#' rows returned by `sql` and passes them to `fd_upsert()`. With the default
#' `update_table = TRUE`, existing keys are updated and new keys are inserted.
#' Use this mode when existing feature values should be refreshed, for example
#' after changing a definition or adding a feature that should be backfilled for
#' already-processed keys.
#'
#' `fd_run()` does not currently coordinate concurrent writers. For the MVP,
#' avoid running multiple `fd_run()`/`fd_upsert()` calls against the same feature
#' table at the same time. Concurrent writes to different feature tables are
#' independent.
#'
#' The returned upsert report can include `extra_columns`: columns that exist in
#' the target feature table but are not produced by the current definitions.
#' These columns are left untouched; the package does not drop, rename, or retire
#' columns automatically.
#'
#' @examples
#' if (requireNamespace("RSQLite", quietly = TRUE)) {
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' on.exit(DBI::dbDisconnect(con), add = TRUE)
#'
#' raw_cars <- mtcars
#' raw_cars$id <- seq_len(nrow(raw_cars))
#' DBI::dbWriteTable(con, "raw_cars_table", raw_cars)
#'
#' defs <- fd_define(
#'   hp_per_cyl = hp / cyl,
#'   engine_ratios = fd_block({
#'     data.frame(
#'       disp_per_cyl = disp / cyl,
#'       wt_per_hp = wt / hp
#'     )
#'   })
#' )
#'
#' res <- fd_run(
#'   con = con,
#'   sql = "SELECT * FROM raw_cars_table",
#'   defs = defs,
#'   key = "id",
#'   feat_table_name = "features",
#'   verbose = FALSE
#' )
#' res$success
#' }
#'
#' @family featdelta pipeline helpers
#' @export
fd_run <- function(
    con,
    sql,
    defs,
    key,
    feat_table_name,
    verbose = FALSE,

    # ---- fetch controls
    fetch_mode = c("new_only", "all"),
    use_max_key = FALSE,
    fetch_limit = NULL,

    # ---- compute controls
    compute_strict = TRUE,
    compute_envir = NULL,

    # ---- upsert controls
    create_table = "auto",
    alter_table  = TRUE,
    update_table = TRUE,
    dialect = NULL,
    chunk_size = NULL,

    # ---- orchestration
    fail_fast = TRUE,
    return_data = c("none", "features", "raw", "both"),
    preview_n = 10L,
    ...
) {
  fetch_mode <- match.arg(fetch_mode)
  return_data <- match.arg(return_data)
  started_at <- Sys.time()

  sql_clean <- clean_sql(sql)

  ctx <- resolve_ctx(
    con,
    feat_table_name = feat_table_name,
    key = key,
    dialect = dialect,
    ...
  )

  ensure_supported_dialect(ctx$dialect)

  validate_general_args(
    key = key,
    feat_table_name = feat_table_name,
    sql = sql_clean,
    defs = defs,
    compute_envir = compute_envir,
    logicals = list(
      verbose = verbose,
      use_max_key = use_max_key,
      compute_strict = compute_strict,
      alter_table = alter_table,
      update_table = update_table,
      fail_fast = fail_fast
    ),
    logicals_with_text = list(
      create_table = create_table
    )
  )

  fetch_limit <- fd_run_validate_optional_positive_int(fetch_limit, "fetch_limit")
  chunk_size <- fd_run_validate_optional_positive_int(chunk_size, "chunk_size")
  preview_n <- fd_run_validate_nonnegative_int(preview_n, "preview_n")

  table_exists <- DBI::dbExistsTable(ctx$con, ctx$feat_table_id)

  fetch_stage <- tryCatch(
    fd_run_fetch_stage(
      ctx = ctx,
      sql_clean = sql_clean,
      fetch_mode = fetch_mode,
      table_exists = table_exists,
      use_max_key = use_max_key,
      fetch_limit = fetch_limit,
      verbose = verbose
    ),
    error = function(e) fd_run_stage_error("fetch", e)
  )

  if (!isTRUE(fetch_stage$ok)) {
    return(fd_run_handle_stage_failure(
      err = fetch_stage,
      fail_fast = fail_fast,
      started_at = started_at,
      ctx = ctx,
      sql_clean = sql_clean,
      fetch_mode = fetch_mode,
      table_exists = table_exists,
      use_max_key = use_max_key,
      fetch_limit = fetch_limit,
      preview_n = preview_n
    ))
  }

  raw_data <- fetch_stage$data

  validate_general_args(
    data = raw_data,
    key = ctx$key
  )

  compute_stage <- tryCatch({
    compute_result <- fd_compute(
      raw_data,
      defs,
      key = ctx$key,
      compute_strict = compute_strict,
      compute_envir = compute_envir,
      verbose = verbose,
      return_report = TRUE
    )

    list(
      ok = TRUE,
      data = compute_result$data,
      report = compute_result$report,
      error = NULL
    )
  }, error = function(e) {
    fd_run_stage_error(
      stage = "compute",
      err = e,
      report = attr(e, "fd_report"),
      data = attr(e, "fd_partial")
    )
  })

  if (!isTRUE(compute_stage$ok)) {
    return(fd_run_handle_stage_failure(
      err = compute_stage,
      fail_fast = fail_fast,
      started_at = started_at,
      ctx = ctx,
      sql_clean = sql_clean,
      fetch_mode = fetch_mode,
      table_exists = table_exists,
      use_max_key = use_max_key,
      fetch_limit = fetch_limit,
      preview_n = preview_n,
      raw_data = raw_data,
      compute_report = compute_stage$report,
      features_df = compute_stage$data,
      return_data = return_data
    ))
  }

  features_df <- compute_stage$data

  upsert_stage <- tryCatch({
    upsert_report <- fd_upsert(
      ctx$con,
      features_df = features_df,
      feat_table_name = ctx$feat_table_name,
      key = ctx$key,
      create_table = create_table,
      alter_table = alter_table,
      update_table = update_table,
      chunk_size = chunk_size,
      dialect = ctx$dialect,
      verbose = verbose,
      return_report = TRUE
    )

    list(ok = TRUE, report = upsert_report, error = NULL)
  }, error = function(e) {
    fd_run_stage_error("upsert", e)
  })

  if (!isTRUE(upsert_stage$ok)) {
    return(fd_run_handle_stage_failure(
      err = upsert_stage,
      fail_fast = fail_fast,
      started_at = started_at,
      ctx = ctx,
      sql_clean = sql_clean,
      fetch_mode = fetch_mode,
      table_exists = table_exists,
      use_max_key = use_max_key,
      fetch_limit = fetch_limit,
      preview_n = preview_n,
      raw_data = raw_data,
      compute_report = compute_stage$report,
      features_df = features_df,
      return_data = return_data
    ))
  }

  fd_run_build_report(
    success = TRUE,
    stage = "complete",
    started_at = started_at,
    ctx = ctx,
    sql_clean = sql_clean,
    fetch_mode = fetch_mode,
    table_exists = table_exists,
    use_max_key = use_max_key,
    fetch_limit = fetch_limit,
    fetch_info = fetch_stage$info,
    raw_data = raw_data,
    compute_report = compute_stage$report,
    features_df = features_df,
    upsert_report = upsert_stage$report,
    return_data = return_data,
    preview_n = preview_n
  )
}


fd_run_fetch_stage <- function(ctx,
                               sql_clean,
                               fetch_mode,
                               table_exists,
                               use_max_key,
                               fetch_limit,
                               verbose) {
  if (isTRUE(verbose)) {
    message("fd_run(): fetch_mode = ", fetch_mode)
  }

  if (identical(fetch_mode, "new_only") && isTRUE(table_exists)) {
    raw_data <- fd_fetch(
      con = ctx$con,
      sql = sql_clean,
      key = ctx$key,
      feat_table_name = ctx$feat_table_name,
      use_max_key = use_max_key,
      verbose = verbose
    )
    fetch_source <- "fd_fetch"
    fetch_attr <- attr(raw_data, "fd_fetch")
  } else {
    raw_data <- DBI::dbGetQuery(ctx$con, sql_clean)
    fetch_source <- if (identical(fetch_mode, "all")) "all" else "initial_all"
    fetch_attr <- NULL
  }

  n_before_limit <- nrow(raw_data)
  limit_applied <- !is.null(fetch_limit) && n_before_limit > fetch_limit

  if (isTRUE(limit_applied)) {
    raw_data <- raw_data[seq_len(fetch_limit), , drop = FALSE]
  }

  info <- list(
    mode = fetch_mode,
    source = fetch_source,
    table_exists = table_exists,
    use_max_key = use_max_key,
    limit = fetch_limit,
    limit_applied = limit_applied,
    n_rows_before_limit = as.integer(n_before_limit),
    n_rows = as.integer(nrow(raw_data)),
    fd_fetch = fetch_attr
  )

  list(ok = TRUE, data = raw_data, info = info)
}


fd_run_stage_error <- function(stage, err, report = NULL, data = NULL) {
  list(
    ok = FALSE,
    stage = stage,
    error = err,
    message = conditionMessage(err),
    class = class(err),
    report = report,
    data = data
  )
}


fd_run_handle_stage_failure <- function(err,
                                        fail_fast,
                                        started_at,
                                        ctx,
                                        sql_clean,
                                        fetch_mode,
                                        table_exists,
                                        use_max_key,
                                        fetch_limit,
                                        preview_n,
                                        raw_data = NULL,
                                        compute_report = NULL,
                                        features_df = NULL,
                                        return_data = "none") {
  if (isTRUE(fail_fast)) {
    stop(err$error)
  }

  fd_run_build_report(
    success = FALSE,
    stage = err$stage,
    started_at = started_at,
    ctx = ctx,
    sql_clean = sql_clean,
    fetch_mode = fetch_mode,
    table_exists = table_exists,
    use_max_key = use_max_key,
    fetch_limit = fetch_limit,
    fetch_info = NULL,
    raw_data = raw_data,
    compute_report = compute_report,
    features_df = features_df,
    upsert_report = NULL,
    error = err,
    return_data = return_data,
    preview_n = preview_n
  )
}


fd_run_build_report <- function(success,
                                stage,
                                started_at,
                                ctx,
                                sql_clean,
                                fetch_mode,
                                table_exists,
                                use_max_key,
                                fetch_limit,
                                fetch_info = NULL,
                                raw_data = NULL,
                                compute_report = NULL,
                                features_df = NULL,
                                upsert_report = NULL,
                                error = NULL,
                                return_data = "none",
                                preview_n = 10L) {
  finished_at <- Sys.time()

  raw_n <- if (is.null(raw_data)) NA_integer_ else as.integer(nrow(raw_data))
  feature_n <- if (is.null(features_df)) NA_integer_ else as.integer(nrow(features_df))
  feature_cols <- if (is.null(features_df)) character() else setdiff(names(features_df), ctx$key)

  out <- list(
    success = success,
    stage = stage,
    started_at = started_at,
    finished_at = finished_at,
    time_sec = as.numeric(difftime(finished_at, started_at, units = "secs")),
    key = ctx$key,
    feat_table_name = ctx$feat_table_name,
    dialect = ctx$dialect,
    sql = sql_clean,
    fetch = fetch_info %||% list(
      mode = fetch_mode,
      table_exists = table_exists,
      use_max_key = use_max_key,
      limit = fetch_limit,
      n_rows = raw_n
    ),
    compute = list(
      n_rows = feature_n,
      n_features = length(feature_cols),
      feature_names = feature_cols,
      report = compute_report
    ),
    upsert = upsert_report,
    preview = list(
      raw = fd_run_preview(raw_data, preview_n),
      features = fd_run_preview(features_df, preview_n)
    ),
    data = fd_run_select_return_data(raw_data, features_df, return_data),
    error = if (is.null(error)) NULL else list(
      stage = error$stage,
      message = error$message,
      class = error$class
    )
  )

  class(out) <- c("fd_run_report", "list")
  out
}


fd_run_select_return_data <- function(raw_data, features_df, return_data) {
  if (identical(return_data, "none")) {
    return(NULL)
  }

  out <- list()

  if (return_data %in% c("raw", "both")) {
    out$raw <- raw_data
  }

  if (return_data %in% c("features", "both")) {
    out$features <- features_df
  }

  out
}


fd_run_preview <- function(x, n) {
  if (is.null(x) || n == 0L) {
    return(NULL)
  }

  utils::head(x, n)
}


fd_run_validate_optional_positive_int <- function(x, name) {
  if (is.null(x)) {
    return(NULL)
  }

  if (!is.numeric(x) || length(x) != 1L || is.na(x) || x <= 0) {
    stop("`", name, "` must be NULL or a positive number.", call. = FALSE)
  }

  as.integer(x)
}


fd_run_validate_nonnegative_int <- function(x, name) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x) || x < 0) {
    stop("`", name, "` must be a non-negative number.", call. = FALSE)
  }

  as.integer(x)
}


#' Print a featdelta run report
#'
#' @param x An `fd_run_report` object.
#' @param ... Unused.
#'
#' @return The input object invisibly.
#'
#' @export
print.fd_run_report <- function(x, ...) {
  cat("<fd_run_report>\n")
  cat(" success:   ", x$success, "\n", sep = "")
  cat(" stage:     ", x$stage, "\n", sep = "")
  cat(" table:     ", x$feat_table_name, "\n", sep = "")
  cat(" dialect:   ", x$dialect, "\n", sep = "")
  cat(" fetched:   ", x$fetch$n_rows %||% NA_integer_, " rows\n", sep = "")
  cat(" computed:  ", x$compute$n_rows %||% NA_integer_, " rows", sep = "")
  cat(", ", x$compute$n_features %||% 0L, " feature columns\n", sep = "")

  if (!is.null(x$upsert) && !is.null(x$upsert$counts)) {
    cat(
      " upsert:    would_insert=", x$upsert$counts$would_insert,
      ", would_update=", x$upsert$counts$would_update,
      "\n",
      sep = ""
    )
  }

  if (!is.null(x$error)) {
    cat(" error:     [", x$error$stage, "] ", x$error$message, "\n", sep = "")
  }

  invisible(x)
}

`%||%` <- function(x, y) if (is.null(x)) y else x

clean_sql <- function(sql) {
  sql_clean <- trimws(sql)
  sql_clean <- sub(";\\s*$", "", sql_clean)

  return(sql_clean)
}

#' Check whether object is a DBI connection
#'
#' Internal helper used to distinguish plain DBI connections from
#' featdelta-specific connection/context objects.
#'
#' @param x Any R object.
#'
#' @return Logical scalar. `TRUE` if `x` inherits from `"DBIConnection"`,
#'   otherwise `FALSE`.
#'
#' @noRd
is_dbi_connection <- function(x) {
  inherits(x, "DBIConnection")
}


#' Check whether object is a featdelta connection/context
#'
#' Identifies objects created by `fd_connect()` (or future equivalents)
#' that carry featdelta-specific defaults and metadata in addition to
#' a DBI connection.
#'
#' @param x Any R object.
#'
#' @return Logical scalar. `TRUE` if `x` inherits from `"featdelta_con"`,
#'   otherwise `FALSE`.
#'
#' @noRd
is_featdelta_con <- function(x) {
  inherits(x, "featdelta_con")
}


#' Convert table name string to DBI table identifier
#'
#' Converts a table name like `"table"` or `"schema.table"` into a value suitable
#' for DBI functions such as `dbExistsTable()` and `dbListFields()`.
#'
#' For Postgres and MySQL, `"schema.table"` is converted to
#' `DBI::Id(schema=..., table=...)`.
#' For SQLite, schema-qualified names are rejected.
#'
#' @noRd
parse_table_id <- function(dialect, table_name) {
  parts <- strsplit(table_name, ".", fixed = TRUE)[[1]]
  if (length(parts) == 1L) return(table_name)

  if (dialect == "sqlite") {
    stop("SQLite does not support schema-qualified `feat_table_name` ('schema.table').")
  }
  if (length(parts) != 2L) {
    stop("`feat_table_name` must be 'table' or 'schema.table'.")
  }
  DBI::Id(schema = parts[1], table = parts[2])
}

quote_table <- function(con, table_name, dialect) {
  as.character(DBI::dbQuoteIdentifier(con, parse_table_id(dialect, table_name)))
}

supported_dialects <- function() {
  c("postgres", "sqlite", "mysql")
}

ensure_supported_dialect <- function(dialect) {
  if (!dialect %in% supported_dialects()) {
    stop(
      "Unsupported/unknown dialect: ", dialect,
      ". Supported: ", paste(supported_dialects(), collapse = ", "), ".",
      call. = FALSE
    )
  }

  invisible(TRUE)
}


#' Detect database dialect from a DBI connection
#'
#' Attempts to infer the database backend (dialect) from DBI metadata and/or
#' the class of a DBI connection. Used to select dialect-specific SQL.
#'
#' @param con A `DBIConnection` object.
#'
#' @return Character scalar identifying the dialect:
#'   one of `"postgres"`, `"sqlite"`, `"mysql"`, or `"unknown"`.
#'
#' @details
#' Detection uses (1) `DBI::dbGetInfo(con)$dbms.name` when available, and
#' falls back to connection class names (e.g. `PqConnection`, `SQLiteConnection`,
#' `MariaDBConnection`) if metadata is unavailable or inconclusive.
#'
#' If detection fails, returns `"unknown"`; callers can decide whether to error.
#'
#' @noRd
detect_backend <- function(con) {
  if (!inherits(con, "DBIConnection")) {
    stop("`con` must be a DBIConnection (from DBI::dbConnect()).", call. = FALSE)
  }
  if (!DBI::dbIsValid(con)) {
    stop("`con` is not a valid (live) DBIConnection.", call. = FALSE)
  }

  # 1) Try DBI metadata
  info <- tryCatch(DBI::dbGetInfo(con), error = function(e) NULL)

  dbms_name <- ""
  if (!is.null(info) && !is.null(info$dbms.name) && length(info$dbms.name) > 0) {
    dbms_name <- tolower(as.character(info$dbms.name)[1])
  }

  # 2) Fall back to connection class
  con_classes <- tolower(class(con))

  is_sqlite <- grepl("sqlite", dbms_name) || any(grepl("sqliteconnection", con_classes))
  if (is_sqlite) return("sqlite")

  is_postgres <- grepl("postgres|postgresql|redshift", dbms_name) ||
    any(grepl("pqconnection|postgresqlconnection", con_classes))
  if (is_postgres) return("postgres")

  is_mysql <- grepl("mysql|mariadb", dbms_name) ||
    any(grepl("mariadbconnection|mysqlconnection", con_classes))
  if (is_mysql) return("mysql")

  # If you prefer strict behavior, error here; otherwise return "unknown".
  "unknown"
}

#' Define a multi-column feature block
#'
#' Creates a block definition object for use inside `fd_define()`.
#'
#' A block represents one ordered definition step that may produce multiple
#' feature columns when later evaluated by `fd_compute()`. Unlike ordinary
#' named single-column definitions, a block may return a whole `data.frame`
#' whose column names become the produced feature names.
#'
#' Supported block inputs are:
#' \itemize{
#'   \item an inline expression, including braced expressions such as
#'         `{ data.frame(...) }`
#'   \item a quosure
#'   \item an `expression()` object of length 1
#'   \item a quoted call or symbol
#'   \item a function, intended to receive the current working data and return a
#'         `data.frame`
#' }
#'
#' `fd_block()` does not compute anything by itself. It only stores the block
#' definition in a normalized form so that `fd_define()` can include it in the
#' ordered definitions object.
#'
#' @param x Block definition input. Must be either a function or a single
#'   expression-like input that later evaluates to a `data.frame`.
#' @param expected_names Optional character vector of expected output column
#'   names. If supplied, names must be non-empty, non-`NA`, and unique.
#'   These names are not required for normal operation, but they can later be
#'   used by `fd_compute()` for validation and for optional NA-completion of
#'   expected-but-missing outputs.
#' @param envir Optional environment used when normalizing expression-like
#'   inputs. If `NULL`, the calling environment is used.
#'
#' @return An object of class `"featdelta_block"`.
#'
#' @examples
#' # Inline expression block
#' blk <- fd_block({
#'   data.frame(
#'     hp_per_cyl = hp / cyl,
#'     disp_per_cyl = disp / cyl
#'   )
#' })
#'
#' # Function-based block
#' make_engine_features <- function(data) {
#'   data.frame(
#'     hp_per_cyl = data$hp / data$cyl,
#'     disp_per_cyl = data$disp / data$cyl
#'   )
#' }
#'
#' blk <- fd_block(make_engine_features)
#'
#' # Optional expected names
#' blk <- fd_block(
#'   {
#'     data.frame(
#'       hp_per_cyl = hp / cyl
#'     )
#'   },
#'   expected_names = c("hp_per_cyl", "disp_per_cyl")
#' )
#'
#' @family featdelta defs helpers
#' @export
fd_block <- function(x, expected_names = NULL, envir = NULL) {

  if (missing(x)) {
    stop("`x` must be supplied.", call. = FALSE)
  }

  if (!is.null(envir) && !is.environment(envir)) {
    stop("`envir` must be an environment or NULL.", call. = FALSE)
  }

  if (!is.null(expected_names)) {
    fd_block_validate_expected_names(expected_names)
  }

  q <- rlang::enquo(x)
  captured <- fd_block_resolve_captured_input(q)

  target_env <- envir %||%
    if (rlang::is_quosure(captured)) rlang::get_env(captured) else rlang::caller_env()

  out <- normalize_fd_block_input(
    x = captured,
    env = target_env,
    expected_names = expected_names
  )

  class(out) <- c("featdelta_block", "list")
  out
}


#' Resolve captured block input
#'
#' Internal helper for `fd_block()`. It keeps inline expressions as quosures,
#' while also resolving symbols that point to supported quoted-expression
#' containers or functions in the calling environment.
#'
#' @param q A quosure captured from `fd_block(x = ...)`.
#'
#' @return Either a quosure or a function.
#'
#' @noRd
fd_block_resolve_captured_input <- function(q) {
  if (!rlang::is_quosure(q)) {
    stop("`q` must be a quosure.", call. = FALSE)
  }

  expr <- rlang::get_expr(q)
  env  <- rlang::get_env(q)

  # Inline expressions, including `{ ... }`, stay as quosures.
  if (!is.symbol(expr)) {
    return(q)
  }

  name <- as.character(expr)

  if (!rlang::env_has(env, name, inherit = TRUE)) {
    return(q)
  }

  value <- rlang::env_get(env, name, inherit = TRUE)

  # Function object supplied through a symbol.
  if (is.function(value)) {
    return(value)
  }

  # Quoted-expression containers supplied through a symbol.
  if (rlang::is_quosure(value)) {
    return(value)
  }

  if (inherits(value, "expression")) {
    if (length(value) != 1L) {
      stop(
        sprintf(
          "`fd_block()` input `%s` refers to an `expression()` object of length %d. ",
          name, length(value)
        ),
        "A block expression must be length 1.",
        call. = FALSE
      )
    }
    return(rlang::new_quosure(value[[1]], env = env))
  }

  if (is.call(value) || is.name(value)) {
    return(rlang::new_quosure(value, env = env))
  }

  q
}


#' Normalize block input to canonical stored form
#'
#' Internal helper used by `fd_block()` to normalize supported block inputs.
#'
#' @param x Block input.
#' @param env Environment used for expression-like inputs.
#' @param expected_names Optional expected output names.
#'
#' @return A normalized list suitable for class `"featdelta_block"`.
#'
#' @noRd
normalize_fd_block_input <- function(x, env, expected_names = NULL) {

  # Case 1: function block
  if (is.function(x)) {
    fn_text <- paste0("<function: ", fd_block_fn_label(x), ">")

    return(list(
      type = "block",
      mode = "function",
      fn = x,
      quo = NULL,
      expr = NULL,
      env = environment(x) %||% env,
      text = fn_text,
      expected_names = expected_names
    ))
  }

  # Case 2: quosure block (includes inline `{ ... }`)
  if (rlang::is_quosure(x)) {
    expr <- rlang::get_expr(x)

    return(list(
      type = "block",
      mode = "expression",
      fn = NULL,
      quo = rlang::new_quosure(expr, env = env),
      expr = expr,
      env = env,
      text = rlang::expr_text(expr),
      expected_names = expected_names
    ))
  }

  # Case 3: base expression() object
  if (inherits(x, "expression")) {
    if (length(x) != 1L) {
      stop(
        sprintf(
          "`fd_block()` received `expression(...)` with length %d. ",
          length(x)
        ),
        "A block expression must be length 1.",
        call. = FALSE
      )
    }

    expr <- x[[1]]

    return(list(
      type = "block",
      mode = "expression",
      fn = NULL,
      quo = rlang::new_quosure(expr, env = env),
      expr = expr,
      env = env,
      text = rlang::expr_text(expr),
      expected_names = expected_names
    ))
  }

  # Case 4: quoted call / symbol
  if (is.call(x) || is.name(x)) {
    return(list(
      type = "block",
      mode = "expression",
      fn = NULL,
      quo = rlang::new_quosure(x, env = env),
      expr = x,
      env = env,
      text = rlang::expr_text(x),
      expected_names = expected_names
    ))
  }

  stop(
    paste0(
      "`fd_block()` requires either a function or a single expression-like input ",
      "(including `{ ... }`, quosure, expression of length 1, call, or symbol)."
    ),
    call. = FALSE
  )
}


#' Validate expected output names for a block
#'
#' @param expected_names Character vector of expected output names.
#'
#' @return Invisibly returns `TRUE` on success.
#'
#' @noRd
fd_block_validate_expected_names <- function(expected_names) {
  if (!is.character(expected_names)) {
    stop("`expected_names` must be a character vector or NULL.", call. = FALSE)
  }

  if (!length(expected_names)) {
    stop("`expected_names` must not be empty if supplied.", call. = FALSE)
  }

  if (any(is.na(expected_names)) || any(!nzchar(expected_names))) {
    stop(
      "`expected_names` must contain only non-empty, non-NA names.",
      call. = FALSE
    )
  }

  if (anyDuplicated(expected_names)) {
    dups <- unique(expected_names[duplicated(expected_names)])
    stop(
      sprintf(
        "Duplicate names in `expected_names`: %s",
        paste(dups, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}


#' Create a readable label for a function block
#'
#' @param fn A function.
#'
#' @return Character scalar label.
#'
#' @noRd
fd_block_fn_label <- function(fn) {
  nm <- tryCatch(deparse(substitute(fn)), error = function(e) NULL)

  if (is.character(nm) && length(nm) == 1L && nzchar(nm) && nm != "x") {
    return(nm)
  }

  "anonymous"
}


#' @export
print.featdelta_block <- function(x, ...) {
  cat("<featdelta_block>", "\n", sep = "")
  cat("Mode: ", x$mode, "\n", sep = "")

  if (!is.null(x$expected_names)) {
    cat(
      "Expected names: ",
      paste(x$expected_names, collapse = ", "),
      "\n",
      sep = ""
    )
  }

  cat("Definition: ", x$text, "\n", sep = "")
  invisible(x)
}

#' Connect to a database with featdelta defaults
#'
#' Creates a `featdelta_con` context object from a DBI driver and connection
#' arguments. The returned object stores the live DBI connection plus optional
#' featdelta defaults such as the raw table, feature table, key column, and
#' metadata settings.
#'
#' @param driver A DBI driver object, such as `RSQLite::SQLite()`,
#'   `RPostgres::Postgres()`, or `RMariaDB::MariaDB()`.
#' @param ... Additional arguments passed to `DBI::dbConnect()`.
#' @param raw_table Optional character scalar naming the raw/source table.
#' @param raw_table_name Deprecated alias for `raw_table`.
#' @param feat_table_name Optional character scalar naming the feature table.
#' @param key Optional character scalar naming the primary key column.
#' @param meta_enabled Logical. Whether featdelta metadata tracking is enabled.
#' @param meta_schema Optional list of metadata schema settings.
#'
#' @return A `featdelta_con` object.
#'
#' @family featdelta context helpers
#' @export
fd_connect <- function(driver,
                       ...,
                       raw_table = NULL,
                       raw_table_name = NULL,
                       feat_table_name = NULL,
                       key = NULL,
                       meta_enabled = FALSE,
                       meta_schema = NULL) {

  if (!is.null(raw_table_name)) {
    if (!is.null(raw_table)) {
      stop("Use only one of `raw_table` or `raw_table_name`.", call. = FALSE)
    }
    warning(
      "`raw_table_name` is deprecated; use `raw_table` instead.",
      call. = FALSE
    )
    raw_table <- raw_table_name
  }

  con <- DBI::dbConnect(driver, ...)

  ctx <- make_ctx_from_args(
    con = con,
    raw_table = raw_table,
    feat_table_name = feat_table_name,
    key = key,
    meta_enabled = meta_enabled,
    meta_schema = meta_schema
  )
  class(ctx) <- c("featdelta_con", class(ctx))
  ctx
}

#' Print a featdelta connection context
#'
#' @param x A `featdelta_con` object.
#' @param ... Unused.
#'
#' @return The input object invisibly.
#'
#' @export
print.featdelta_con <- function(x, ...) {
  cat("<featdelta_con>\n")
  cat(" dialect:   ", x$dialect, "\n", sep = "")
  cat(" raw_table: ", x$raw_table %||% "<unset>", "\n", sep = "")
  cat(" feat_table_name:", x$feat_table_name %||% "<unset>", "\n", sep = "")
  cat(" key:       ", x$key %||% "<unset>", "\n", sep = "")
  cat(" meta:      ", if (isTRUE(x$meta_enabled)) "enabled" else "disabled", "\n", sep = "")
  invisible(x)
}

#' Upsert computed features into a database features table
#'
#' Takes a feature `data.frame` produced by `fd_compute()` (one row per entity id)
#' and writes it into a database table, using explicit incremental semantics:
#' new keys are inserted; existing keys are optionally updated.
#'
#' The function is set-based (staging + merge) and avoids "read all ids into R"
#' patterns. Writes are performed in chunks (optional) inside a transaction.
#'
#' Schema evolution is supported in a restricted, safe form: when `alter_table = TRUE`,
#' missing feature columns are added to the target table via `ALTER TABLE ... ADD COLUMN ...`.
#' No column drops/renames/type changes are performed. Columns that exist in
#' the target table but are not present in `features_df` are left untouched and
#' reported as `extra_columns`.
#'
#' @param con A `DBIConnection`. Must be a live connection to the database where
#'   `feat_table_name` exists (or will be created).
#'
#' @param features_df A `data.frame` containing at minimum the primary key column
#'   specified by `key`, plus one or more feature columns to be stored.
#'   Requirements:
#'   * `key` must exist in `names(features_df)`.
#'   * `features_df[[key]]` must contain no `NA` and no duplicate values.
#'   * Each feature column should be an atomic vector (numeric, integer, logical,
#'     character, Date/POSIXct) that can be written by DBI for the chosen dialect.
#'
#' @param feat_table_name A single string. Name of the database table where features
#'   are stored (the "target" table). For dialects supporting schemas (e.g. Postgres),
#'   this may be schema-qualified (e.g. `"public.my_features"`), subject to dialect rules.
#'
#' @param key A single string. Name of the primary key column in both `features_df`
#'   and `feat_table_name`. This column uniquely identifies each entity/row.
#'
#' @param create_table Logical, or character value `auto`. If `TRUE`, create `feat_table_name` when it does not exist.
#'   The created table includes all columns present in `features_df` and defines
#'   `PRIMARY KEY(key)`. If `FALSE` and `feat_table_name` is missing, the function errors.If `auto`,
#'   create the table it hasn't been already created.
#'
#' @param alter_table Logical. If `TRUE`, add missing feature columns found in
#'   `features_df` but not present in `feat_table_name` using `ALTER TABLE ... ADD COLUMN ...`.
#'   Only additive schema changes are performed. If `FALSE`, the function errors
#'   when `features_df` contains columns not present in `feat_table_name`.
#'
#' @param update_table  Logical. Controls incremental behavior for keys that already exist
#'   in `feat_table_name`:
#'   * If `TRUE` (default), existing keys are updated (upsert).
#'   * If `FALSE`, only new keys are inserted; any conflict with existing keys
#'     results in an error (insert-only mode).
#'
#' @param chunk_size Optional integer batch size. If provided, `features_df` is written
#'   in batches of up to `chunk_size` rows, each merged via a staging table.
#'   Use this to limit memory/packet sizes for large writes. If `NULL`, write in one batch.
#' @param verbose Logical. If `TRUE`, emit progress messages (e.g., chunk progress,
#'   table creation/alteration actions). Does not affect returned results.
#' @param return_report Logical. If `TRUE`, return a structured
#'   `fd_upsert_report`; if `FALSE`, return `TRUE` invisibly after successful
#'   side effects.
#' @param dialect Optional dialect override. Supported values are `"postgres"`,
#'   `"sqlite"`, and `"mysql"`.
#'
#' @details
#' **Incremental semantics**
#' \itemize{
#'   \item Insert new keys (keys in `features_df` not present in `feat_table_name`).
#'   \item Update existing keys only when `update_table = TRUE`.
#'   \item When `update_table = FALSE`, any overlap between staged keys and existing keys
#'         is treated as a conflict and aborts the write.
#' }
#'
#' **Counts**
#' The returned report contains `would_insert` / `would_update` counts, computed as
#' existence-based counts *prior to the merge* (within the transaction). These are
#' not "rows whose values changed", only "rows targeted as insert/update".
#'
#' **Transactions**
#' The operation runs in a transaction. If any step fails (schema change, staging write,
#' merge), the transaction is rolled back and the target table is unchanged.
#' Existing target tables used with `update_table = TRUE` must have a primary
#' key or unique constraint/index on `key`. Tables created by `fd_upsert()` get
#' this primary key automatically.
#'
#' **Concurrency**
#' `fd_upsert()` does not currently take an explicit table-level lock. Avoid
#' running concurrent writes to the same feature table. Concurrent writes to
#' different feature tables are independent.
#'
#' @return
#' An `fd_upsert_report` object (S3) with a structured summary of actions performed:
#' \itemize{
#'   \item `table_created` (logical)
#'   \item `columns_added` (character vector)
#'   \item `extra_columns` (character vector): target-table columns not present
#'         in the incoming features data, excluding `key`
#'   \item `counts$would_insert` (integer)
#'   \item `counts$would_update` (integer; `0` when `update_table = FALSE`)
#'   \item optional per-chunk breakdown (if chunking is used)
#' }
#'
#' @examples
#' if (requireNamespace("RSQLite", quietly = TRUE)) {
#' con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
#' on.exit(DBI::dbDisconnect(con), add = TRUE)
#'
#' feats <- data.frame(
#'   id = c(1, 2, 3),
#'   f_age = c(10, 20, 30),
#'   f_flag = c(TRUE, FALSE, TRUE)
#' )
#'
#' # Create table and insert initial rows
#' r1 <- fd_upsert(
#'   con = con,
#'   features_df = feats,
#'   feat_table_name = "features_tbl",
#'   key = "id",
#'   create_table = TRUE,
#'   alter_table = FALSE,
#'   update_table = TRUE,
#'   verbose = FALSE
#' )
#'
#' # Upsert: update ids 2-3, insert id 4
#' feats2 <- data.frame(
#'   id = c(2, 3, 4),
#'   f_age = c(21, 31, 40),
#'   f_flag = c(FALSE, TRUE, FALSE)
#' )
#'
#' r2 <- fd_upsert(
#'   con = con,
#'   features_df = feats2,
#'   feat_table_name = "features_tbl",
#'   key = "id",
#'   update_table = TRUE,
#'   verbose = FALSE
#' )
#'
#' # Schema evolution: add a new feature column
#' feats3 <- data.frame(
#'   id = c(4, 5),
#'   f_age = c(41, 50),
#'   f_flag = c(TRUE, TRUE),
#'   f_new = c("A", "B")
#' )
#'
#' r3 <- fd_upsert(
#'   con = con,
#'   features_df = feats3,
#'   feat_table_name = "features_tbl",
#'   key = "id",
#'   alter_table = TRUE,
#'   update_table = TRUE,
#'   verbose = FALSE
#' )
#' }
#'
#' @export
fd_upsert <- function(
    con,
    features_df,
    feat_table_name,
    key,
    create_table = "auto",
    alter_table  = TRUE,
    update_table = TRUE,
    chunk_size   = NULL,
    verbose      = TRUE,
    return_report = TRUE,
    dialect      = NULL) {

  ctx <- resolve_ctx(
    con_or_fd = con,
    feat_table_name = feat_table_name,
    key = key,
    dialect = dialect
    #meta_enabled = meta_enabled,
    #meta_schema = meta_schema
  )

  ensure_supported_dialect(ctx$dialect)

  ## validation of inputs
  stopifnot(!is.null(features_df))
  validate_general_args(
    features_df = features_df,
    key = key,
    logicals = list(
      alter_table = alter_table,
      update_table = update_table,
      verbose = verbose,
      return_report = return_report
    ),
    logicals_with_text = list(
      create_table = create_table
    )
  )

  if (!is.null(chunk_size)) {
    if (!is.numeric(chunk_size) || length(chunk_size) != 1L || is.na(chunk_size) || chunk_size <= 0) {
      stop("`chunk_size` must be NULL or a positive number.")
    }
    chunk_size <- as.integer(chunk_size)
  }

  # Normalize common problematic types for DB writes
  features_df <- normalize_features_df(features_df)

  ## MAIN transaction here (schema + all chunks)
  report <- DBI::dbWithTransaction(ctx$con, {

    # 1) Make sure the target table is ready for writing:
    #    - resolve create_table = "auto"
    #    - create table if requested/missing
    #    - add missing columns if allowed
    target_state <- fd_upsert_prepare_target(
      ctx = ctx,
      features_df = features_df,
      create_table = create_table,
      alter_table = alter_table,
      update_table = update_table,
      verbose = verbose
    )

    # 2) Nothing to write.
    if (nrow(features_df) == 0L) {
      if (!isTRUE(return_report)) {
        if (isTRUE(verbose)) message("No rows to write.")
        TRUE
      } else {
        fd_upsert_empty_report(
          feat_table_name = feat_table_name,
          key = key,
          dialect = ctx$dialect,
          table_created = target_state$table_created,
          columns_added = target_state$columns_added,
          extra_columns = target_state$extra_columns
        )
      }
    } else {
      # 3) Write the incoming features in chunks.
      if (isTRUE(verbose)) {
        message(
          "Writing features to ", feat_table_name,
          " (dialect=", ctx$dialect, "): ",
          nrow(features_df), " rows. Mode: ",
          if (isTRUE(update_table)) "upsert" else "insert-only", "."
        )
      }

      chunk_state <- fd_upsert_execute_chunks(
        con = ctx$con,
        dialect = ctx$dialect,
        target_q = target_state$target_q,
        key = key,
        features_df = features_df,
        update_table = update_table,
        chunk_size = chunk_size,
        verbose = verbose,
        return_report = return_report
      )

      # 4) If the caller only wants side effects, finish here.
      if (!isTRUE(return_report)) {
        if (isTRUE(verbose)) message("Done.")
        TRUE
      } else {
        # 5) Build the final structured report.
        out <- fd_upsert_build_report(
          feat_table_name = feat_table_name,
          key = key,
          dialect = ctx$dialect,
          n_rows = nrow(features_df),
          n_chunks = chunk_state$n_chunks,
          table_created = target_state$table_created,
          columns_added = target_state$columns_added,
          extra_columns = target_state$extra_columns,
          totals_would_insert = chunk_state$totals_would_insert,
          totals_would_update = chunk_state$totals_would_update,
          chunk_details = chunk_state$chunk_details
        )

        if (isTRUE(verbose)) {
          message(
            "Done. would_insert=", out$counts$would_insert,
            if (isTRUE(update_table)) paste0(", would_update=", out$counts$would_update) else ""
          )
          if (length(out$columns_added) > 0L) {
            message("Columns added: ", paste(out$columns_added, collapse = ", "))
          }
          if (length(out$extra_columns) > 0L) {
            message("Extra target columns left untouched: ", paste(out$extra_columns, collapse = ", "))
          }
        }

        out
      }
    }
  })

  if (isTRUE(return_report)) {
    report
  } else {
    invisible(TRUE)
  }
}


fd_upsert_prepare_target <- function(
    ctx,
    features_df,
    create_table,
    alter_table,
    update_table,
    verbose
) {
  table_id <- ctx$feat_table_id
  table_exists <- DBI::dbExistsTable(ctx$con, table_id)

  if (is.character(create_table) && identical(create_table, "auto")) {
    create_table <- !table_exists
  }

  if (isTRUE(create_table) && table_exists) {
    drop_sql <- render_sql(
      fd_sql_templates$drop_table[[ctx$dialect]],
      list(target = ctx$feat_table_q)
    )

    stop(
      paste0(
        "create_table=TRUE but table already exists: ", ctx$feat_table_name, "\n\n",
        "Overwriting is not allowed.\n",
        "If you are sure you want to delete the table, run:\n",
        "DBI::dbExecute(con, ", shQuote(drop_sql), ")\n"
      ),
      call. = FALSE
    )
  }

  if (!table_exists && !isTRUE(create_table)) {
    stop(
      paste0(
        "Table does not exist: ", ctx$feat_table_name,
        ". Set create_table=TRUE or create_table='auto' to create it."
      ),
      call. = FALSE
    )
  }

  table_created <- FALSE

  if (!table_exists && isTRUE(create_table)) {
    if (isTRUE(verbose)) {
      message("Creating table: ", ctx$feat_table_name)
    }

    create_features_table(
      con = ctx$con,
      dialect = ctx$dialect,
      feat_table_name = ctx$feat_table_name,
      target_q = ctx$feat_table_q,
      features_df = features_df,
      key = ctx$key
    )

    table_created <- TRUE
  }

  columns_added <- ensure_columns(
    con = ctx$con,
    dialect = ctx$dialect,
    table_id = ctx$feat_table_id,
    table_q = ctx$feat_table_q,
    features_df = features_df,
    key = ctx$key,
    alter_table = alter_table,
    verbose = verbose
  )
  extra_columns <- attr(columns_added, "extra_columns") %||% character()
  attr(columns_added, "extra_columns") <- NULL

  if (isTRUE(update_table) && !isTRUE(table_created)) {
    ensure_key_constraint(
      con = ctx$con,
      dialect = ctx$dialect,
      table_name = ctx$feat_table_name,
      table_id = ctx$feat_table_id,
      table_q = ctx$feat_table_q,
      key = ctx$key
    )
  }

  list(
    target_q = ctx$feat_table_q,
    table_created = table_created,
    columns_added = columns_added,
    extra_columns = extra_columns
  )
}

ensure_key_constraint <- function(con, dialect, table_name, table_id, table_q, key) {
  ok <- switch(
    dialect,
    sqlite = sqlite_has_key_constraint(con, table_q, key),
    postgres = postgres_has_key_constraint(con, table_name, key),
    mysql = mysql_has_key_constraint(con, table_name, key),
    FALSE
  )

  if (!isTRUE(ok)) {
    stop(
      sprintf(
        "Target table %s must have a PRIMARY KEY or UNIQUE constraint on %s when update_table=TRUE.",
        table_q,
        shQuote(key)
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}

sqlite_has_key_constraint <- function(con, table_q, key) {
  table_info <- DBI::dbGetQuery(con, paste0("PRAGMA table_info(", table_q, ")"))
  if (nrow(table_info) > 0L && any(table_info$name == key & table_info$pk > 0L)) {
    return(TRUE)
  }

  index_list <- DBI::dbGetQuery(con, paste0("PRAGMA index_list(", table_q, ")"))
  if (nrow(index_list) == 0L) {
    return(FALSE)
  }

  unique_indexes <- index_list$name[index_list$unique == 1L]
  for (idx in unique_indexes) {
    idx_q <- as.character(DBI::dbQuoteIdentifier(con, idx))
    idx_info <- DBI::dbGetQuery(con, paste0("PRAGMA index_info(", idx_q, ")"))
    if (identical(idx_info$name, key)) {
      return(TRUE)
    }
  }

  FALSE
}

postgres_has_key_constraint <- function(con, table_name, key) {
  table_lit <- as.character(DBI::dbQuoteLiteral(con, table_name))
  key_lit <- as.character(DBI::dbQuoteLiteral(con, key))

  sql <- paste0(
    "SELECT 1 ",
    "FROM pg_constraint c ",
    "WHERE c.conrelid = ", table_lit, "::regclass ",
    "AND c.contype IN ('p', 'u') ",
    "AND (",
    "  SELECT array_agg(a.attname ORDER BY ck.ord) ",
    "  FROM unnest(c.conkey) WITH ORDINALITY AS ck(attnum, ord) ",
    "  JOIN pg_attribute a ON a.attrelid = c.conrelid AND a.attnum = ck.attnum",
    ") = ARRAY[", key_lit, "]::name[] ",
    "LIMIT 1"
  )

  nrow(DBI::dbGetQuery(con, sql)) > 0L
}

mysql_has_key_constraint <- function(con, table_name, key) {
  parts <- strsplit(table_name, ".", fixed = TRUE)[[1]]
  if (length(parts) == 1L) {
    schema_expr <- "DATABASE()"
    table_lit <- as.character(DBI::dbQuoteLiteral(con, parts[1]))
  } else if (length(parts) == 2L) {
    schema_expr <- as.character(DBI::dbQuoteLiteral(con, parts[1]))
    table_lit <- as.character(DBI::dbQuoteLiteral(con, parts[2]))
  } else {
    stop("`feat_table_name` must be 'table' or 'schema.table'.", call. = FALSE)
  }

  key_lit <- as.character(DBI::dbQuoteLiteral(con, key))

  sql <- paste0(
    "SELECT 1 ",
    "FROM information_schema.table_constraints tc ",
    "JOIN information_schema.key_column_usage kcu ",
    "  ON kcu.constraint_schema = tc.constraint_schema ",
    " AND kcu.constraint_name = tc.constraint_name ",
    " AND kcu.table_schema = tc.table_schema ",
    " AND kcu.table_name = tc.table_name ",
    "WHERE tc.constraint_type IN ('PRIMARY KEY', 'UNIQUE') ",
    "AND tc.table_schema = ", schema_expr, " ",
    "AND tc.table_name = ", table_lit, " ",
    "GROUP BY tc.constraint_schema, tc.constraint_name ",
    "HAVING COUNT(*) = 1 AND SUM(kcu.column_name = ", key_lit, ") = 1 ",
    "LIMIT 1"
  )

  nrow(DBI::dbGetQuery(con, sql)) > 0L
}

fd_upsert_execute_chunks <- function(
    con,
    dialect,
    target_q,
    key,
    features_df,
    update_table,
    chunk_size,
    verbose,
    return_report
) {
  chunks <- make_chunks(nrow(features_df), chunk_size)
  n_chunks <- length(chunks)

  totals_would_insert <- 0L
  totals_would_update <- 0L
  chunk_details <- if (isTRUE(return_report)) vector("list", n_chunks) else NULL

  for (i in seq_len(n_chunks)) {
    chunk_result <- fd_upsert_apply_chunk(
      con = con,
      dialect = dialect,
      target_q = target_q,
      key = key,
      chunk_df = features_df[chunks[[i]], , drop = FALSE],
      chunk_id = i,
      update_table = update_table,
      verbose = verbose,
      return_report = return_report
    )

    if (isTRUE(return_report)) {
      totals_would_insert <- totals_would_insert + chunk_result$would_insert
      totals_would_update <- totals_would_update + chunk_result$would_update

      chunk_details[[i]] <- data.frame(
        chunk = i,
        n = chunk_result$n,
        would_insert = chunk_result$would_insert,
        would_update = chunk_result$would_update,
        stringsAsFactors = FALSE
      )
    }
  }

  list(
    n_chunks = n_chunks,
    totals_would_insert = totals_would_insert,
    totals_would_update = totals_would_update,
    chunk_details = if (isTRUE(return_report)) do.call(rbind, chunk_details) else NULL
  )
}

fd_upsert_apply_chunk <- function(
    con,
    dialect,
    target_q,
    key,
    chunk_df,
    chunk_id,
    update_table,
    verbose,
    return_report
) {
  stage_name <- get_stage_name(chunk_id)
  stage_q <- as.character(DBI::dbQuoteIdentifier(con, stage_name))
  key_q <- as.character(DBI::dbQuoteIdentifier(con, key))

  on.exit({
    if (DBI::dbExistsTable(con, stage_name)) {
      DBI::dbRemoveTable(con, stage_name)
    }
  }, add = TRUE)

  if (isTRUE(verbose)) {
    message("Chunk ", chunk_id, " (", nrow(chunk_df), " rows): staging -> merge")
  }

  fd_upsert_write_stage_table(
    con = con,
    dialect = dialect,
    stage_name = stage_name,
    chunk_df = chunk_df
  )

  if (!isTRUE(update_table)) {
    conflicts <- find_conflicts(
      con = con,
      dialect = dialect,
      target_q = target_q,
      stage_q = stage_q,
      key_q = key_q,
      limit = 50L
    )

    if (length(conflicts) > 0L) {
      stop(
        sprintf(
          "Conflicts detected (update_table=FALSE). Example key values: %s",
          paste(conflicts, collapse = ", ")
        ),
        call. = FALSE
      )
    }
  }

  would_insert <- 0L
  would_update <- 0L

  if (isTRUE(return_report)) {
    would_insert <- count_scalars(
      con,
      fd_sql_templates$upsert_count_would_insert[[dialect]],
      target_q = target_q,
      stage_q = stage_q,
      key_q = key_q
    )

    would_update <- if (isTRUE(update_table)) {
      count_scalars(
        con,
        fd_sql_templates$upsert_count_would_update[[dialect]],
        target_q = target_q,
        stage_q = stage_q,
        key_q = key_q
      )
    } else {
      0L
    }
  }

  DBI::dbExecute(
    con,
    fd_upsert_build_merge_sql(
      con = con,
      dialect = dialect,
      target_q = target_q,
      stage_q = stage_q,
      key = key,
      cols = names(chunk_df),
      update_table = update_table
    )
  )

  if (isTRUE(return_report) && isTRUE(verbose)) {
    if (isTRUE(update_table)) {
      message("  would_insert=", would_insert, ", would_update=", would_update)
    } else {
      message("  would_insert=", would_insert, " (insert-only; conflicts would error)")
    }
  }

  list(
    n = nrow(chunk_df),
    would_insert = as.integer(would_insert),
    would_update = as.integer(would_update)
  )
}

fd_upsert_write_stage_table <- function(con, dialect, stage_name, chunk_df) {
  chunk_df <- as.data.frame(chunk_df)

  if (identical(dialect, "sqlite")) {
    DBI::dbCreateTable(
      conn = con,
      name = stage_name,
      fields = chunk_df,
      temporary = TRUE
    )
    DBI::dbAppendTable(con, stage_name, chunk_df)
    return(invisible(TRUE))
  }

  DBI::dbWriteTable(
    con,
    name = stage_name,
    value = chunk_df,
    overwrite = TRUE,
    append = FALSE,
    temporary = TRUE,
    row.names = FALSE
  )

  invisible(TRUE)
}

fd_upsert_build_merge_sql <- function(
    con,
    dialect,
    target_q,
    stage_q,
    key,
    cols,
    update_table
) {
  key_q <- as.character(DBI::dbQuoteIdentifier(con, key))
  cols_q <- as.character(DBI::dbQuoteIdentifier(con, cols))
  cols_all_q <- paste(cols_q, collapse = ", ")
  cols_select_q <- paste(paste0("s.", cols_q), collapse = ", ")

  feature_cols <- setdiff(cols, key)

  if (length(feature_cols) == 0L) {
    excluded_prefix <- fd_upsert_excluded_prefix(dialect)
    set_clause <- paste0(key_q, " = ", excluded_prefix, key_q)
  } else {
    feature_cols_q <- as.character(DBI::dbQuoteIdentifier(con, feature_cols))
    excluded_prefix <- fd_upsert_excluded_prefix(dialect)
    set_clause <- paste(paste0(feature_cols_q, " = ", excluded_prefix, feature_cols_q), collapse = ", ")
  }

  template_name <- if (isTRUE(update_table)) {
    "upsert_merge_upsert"
  } else {
    "upsert_merge_insert_only"
  }

  render_sql(
    fd_sql_templates[[template_name]][[dialect]],
    list(
      target = target_q,
      stage = stage_q,
      cols_all = cols_all_q,
      cols_insert = cols_all_q,
      cols_select = cols_select_q,
      key = key_q,
      set_clause = set_clause,
      t_key = paste0("t.", key_q),
      s_key = paste0("s.", key_q)
    )
  )
}

fd_upsert_excluded_prefix <- function(dialect) {
  switch(
    dialect,
    postgres = "EXCLUDED.",
    sqlite = "excluded.",
    mysql = "fd_src.",
    stop("Unsupported/unknown dialect: ", dialect, call. = FALSE)
  )
}

fd_upsert_empty_report <- function(
    feat_table_name,
    key,
    dialect,
    table_created,
    columns_added,
    extra_columns = character()
) {
  structure(
    list(
      feat_table_name = feat_table_name,
      key = key,
      dialect = dialect,
      n_rows = 0L,
      n_chunks = 0L,
      table_created = table_created,
      columns_added = columns_added,
      extra_columns = extra_columns,
      counts = list(
        would_insert = 0L,
        would_update = 0L
      ),
      chunk_details = data.frame(
        chunk = integer(),
        n = integer(),
        would_insert = integer(),
        would_update = integer()
      )
    ),
    class = "fd_upsert_report"
  )
}

fd_upsert_build_report <- function(
    feat_table_name,
    key,
    dialect,
    n_rows,
    n_chunks,
    table_created,
    columns_added,
    extra_columns = character(),
    totals_would_insert,
    totals_would_update,
    chunk_details
) {
  structure(
    list(
      feat_table_name = feat_table_name,
      key = key,
      dialect = dialect,
      n_rows = as.integer(n_rows),
      n_chunks = as.integer(n_chunks),
      table_created = table_created,
      columns_added = columns_added,
      extra_columns = extra_columns,
      counts = list(
        would_insert = as.integer(totals_would_insert),
        would_update = as.integer(totals_would_update)
      ),
      chunk_details = chunk_details
    ),
    class = "fd_upsert_report"
  )
}


#' Normalize feature data types before DB write
#'
#' Performs minimal type normalization to improve DBI write compatibility.
#' Currently converts factor columns to character. Leaves Date/POSIXct unchanged.
#'
#' @noRd
normalize_features_df <- function(df) {
  for (nm in names(df)) {
    if (is.factor(df[[nm]])) df[[nm]] <- as.character(df[[nm]])
    # keep Date/POSIXct as-is (DBI drivers decide mapping)
  }
  df
}

#' Generate a unique staging table name
#'
#' Creates a unique, dialect-safe staging table name for use inside a transaction/chunk.
#' Names use only `[A-Za-z0-9_]` to avoid quoting issues across DBs.
#'
#' @noRd
get_stage_name <- function(i) {
  # Only [A-Za-z0-9_] to avoid quoting surprises across dialects.
  rnd <- paste(sample(c(letters, LETTERS, 0:9), 8, replace = TRUE), collapse = "")
  paste0("fd_stage_", format(Sys.time(), "%Y%m%d%H%M%S"), "_", i, "_", rnd)
}

#' Split row indices into chunks
#'
#' Returns a list of integer index vectors partitioning `1:n` into chunks of size
#' `chunk_size`. If `chunk_size` is NULL, returns a single chunk containing all rows.
#'
#' @noRd
make_chunks <- function(n, chunk_size) {
  if (n == 0L) return(list(integer(0)))
  if (is.null(chunk_size)) return(list(seq_len(n)))
  chunk_size <- as.integer(chunk_size)
  split(seq_len(n), ceiling(seq_len(n) / chunk_size))
}


create_features_table <- function(con, dialect, feat_table_name, target_q, features_df, key) {
  cols <- names(features_df)

  # Column definitions using dbDataType
  col_defs <- vapply(cols, function(nm) {
    type <- DBI::dbDataType(con, features_df[[nm]])
    col_q <- as.character(DBI::dbQuoteIdentifier(con, nm))
    paste(col_q, type)
  }, character(1))

  key_q <- as.character(DBI::dbQuoteIdentifier(con, key))

  sql <- paste0(
    "CREATE TABLE ", target_q, " (",
    paste(col_defs, collapse = ", "),
    ", PRIMARY KEY (", key_q, ")",
    ")"
  )

  DBI::dbExecute(con, sql)
  invisible(TRUE)
}

ensure_columns <- function(
    con,
    dialect,
    table_id,
    table_q,
    features_df,
    key,
    alter_table,
    verbose
) {
  existing <- DBI::dbListFields(con, table_id)
  extra_columns <- setdiff(existing, names(features_df))
  extra_columns <- setdiff(extra_columns, key)

  if (!key %in% existing) {
    stop(
      sprintf(
        "Target table %s does not contain key column %s.",
        table_q,
        shQuote(key)
      ),
      call. = FALSE
    )
  }

  missing <- setdiff(names(features_df), existing)
  if (length(missing) == 0L) {
    out <- character()
    if (length(extra_columns) > 0L) {
      attr(out, "extra_columns") <- extra_columns
    }
    return(out)
  }

  if (!isTRUE(alter_table)) {
    stop(
      sprintf(
        "Target table %s is missing columns: %s. Set alter_table=TRUE to add them.",
        table_q,
        paste(missing, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  if (isTRUE(verbose)) {
    message("Altering table to add missing columns: ", paste(missing, collapse = ", "))
  }

  added <- character()
  for (nm in missing) {
    if (identical(nm, key)) {
      stop(
        "Refusing to add key column via alter_table; key must already exist in target.",
        call. = FALSE
      )
    }

    col_q <- as.character(DBI::dbQuoteIdentifier(con, nm))
    type  <- DBI::dbDataType(con, features_df[[nm]])

    sql <- render_sql(
      fd_sql_templates$alter_add_column[[dialect]],
      list(
        target = table_q,
        col = col_q,
        type = as.character(type)
      )
    )

    DBI::dbExecute(con, sql)
    added <- c(added, nm)
  }

  if (length(extra_columns) > 0L) {
    attr(added, "extra_columns") <- extra_columns
  }
  added
}

count_scalars <- function(con, template, target_q, stage_q, key_q) {
  t_key <- paste0("t.", key_q)
  s_key <- paste0("s.", key_q)
  sql <- render_sql(
    template,
    list(target = target_q, stage = stage_q, t_key = t_key, s_key = s_key)
  )
  res <- DBI::dbGetQuery(con, sql)
  as.integer(res[[1]][[1]])
}

find_conflicts <- function(con, dialect, target_q, stage_q, key_q, limit = 50L) {
  t_key <- paste0("t.", key_q)
  s_key <- paste0("s.", key_q)

  sql <- render_sql(
    fd_sql_templates$upsert_find_conflicts[[dialect]],
    list(
      target = target_q,
      stage  = stage_q,
      t_key  = t_key,
      s_key  = s_key,
      limit  = as.character(as.integer(limit))
    )
  )
  out <- DBI::dbGetQuery(con, sql)
  if (nrow(out) == 0L) return(character())
  as.character(out[[1]])
}

upsert_ns <- function(name) {
  getFromNamespace(name, "featdelta")
}

test_that("render_sql replaces placeholders and rejects invalid values", {
  render_sql <- upsert_ns("render_sql")

  out <- render_sql(
    "INSERT INTO {target} ({cols}) SELECT {cols} FROM {stage}",
    list(
      target = "`features_tbl`",
      cols = "`id`, `score`",
      stage = "`stage_tbl`"
    )
  )

  expect_identical(
    out,
    "INSERT INTO `features_tbl` (`id`, `score`) SELECT `id`, `score` FROM `stage_tbl`"
  )

  expect_error(
    render_sql("SELECT {missing}", list(missing = NULL)),
    "invalid replacement value"
  )
  expect_error(
    render_sql("SELECT {missing}", list(missing = c("a", "b"))),
    "invalid replacement value"
  )
  expect_error(
    render_sql("SELECT {missing}", list(missing = NA_character_)),
    "invalid replacement value"
  )
})

test_that("fd_upsert_build_merge_sql renders SQLite upsert SQL", {
  build_merge_sql <- upsert_ns("fd_upsert_build_merge_sql")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  sql <- build_merge_sql(
    con = con,
    dialect = "sqlite",
    target_q = "`features_tbl`",
    stage_q = "`stage_tbl`",
    key = "id",
    cols = c("id", "score", "flag"),
    update_table = TRUE
  )

  expect_match(sql, "INSERT INTO `features_tbl` (`id`, `score`, `flag`)", fixed = TRUE)
  expect_match(sql, "SELECT `id`, `score`, `flag` FROM `stage_tbl`", fixed = TRUE)
  expect_match(sql, "ON CONFLICT(`id`) DO UPDATE SET", fixed = TRUE)
  expect_match(sql, "`score` = excluded.`score`", fixed = TRUE)
  expect_match(sql, "`flag` = excluded.`flag`", fixed = TRUE)
})

test_that("fd_upsert_build_merge_sql renders SQLite insert-only SQL", {
  build_merge_sql <- upsert_ns("fd_upsert_build_merge_sql")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  sql <- build_merge_sql(
    con = con,
    dialect = "sqlite",
    target_q = "`features_tbl`",
    stage_q = "`stage_tbl`",
    key = "id",
    cols = c("id", "score"),
    update_table = FALSE
  )

  expect_match(sql, "INSERT INTO `features_tbl` (`id`, `score`)", fixed = TRUE)
  expect_match(sql, "SELECT s.`id`, s.`score`", fixed = TRUE)
  expect_match(sql, "FROM `stage_tbl` s", fixed = TRUE)
  expect_match(sql, "LEFT JOIN `features_tbl` t ON t.`id` = s.`id`", fixed = TRUE)
  expect_match(sql, "WHERE t.`id` IS NULL", fixed = TRUE)
  expect_false(grepl("ON CONFLICT", sql, fixed = TRUE))
})

test_that("fd_upsert_build_merge_sql renders MySQL upsert SQL", {
  build_merge_sql <- upsert_ns("fd_upsert_build_merge_sql")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  sql <- build_merge_sql(
    con = con,
    dialect = "mysql",
    target_q = "`features_tbl`",
    stage_q = "`stage_tbl`",
    key = "id",
    cols = c("id", "score", "flag"),
    update_table = TRUE
  )

  expect_match(sql, "INSERT INTO `features_tbl` (`id`, `score`, `flag`)", fixed = TRUE)
  expect_match(sql, "FROM (SELECT `id`, `score`, `flag` FROM `stage_tbl`) AS fd_src", fixed = TRUE)
  expect_match(sql, "ON DUPLICATE KEY UPDATE", fixed = TRUE)
  expect_match(sql, "`score` = fd_src.`score`", fixed = TRUE)
  expect_match(sql, "`flag` = fd_src.`flag`", fixed = TRUE)
  expect_false(grepl("VALUES(", sql, fixed = TRUE))
})

test_that("all SQL templates include supported dialects", {
  fd_sql_templates <- upsert_ns("fd_sql_templates")
  supported_dialects <- upsert_ns("supported_dialects")

  expect_true(all(vapply(fd_sql_templates, function(template) {
    all(supported_dialects() %in% names(template))
  }, logical(1))))
})

test_that("fd_upsert_build_merge_sql handles key-only updates", {
  build_merge_sql <- upsert_ns("fd_upsert_build_merge_sql")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  sql <- build_merge_sql(
    con = con,
    dialect = "sqlite",
    target_q = "`features_tbl`",
    stage_q = "`stage_tbl`",
    key = "id",
    cols = "id",
    update_table = TRUE
  )

  expect_match(sql, "ON CONFLICT(`id`) DO UPDATE SET `id` = excluded.`id`", fixed = TRUE)
})

test_that("fd_upsert report constructors return structured report objects", {
  empty_report <- upsert_ns("fd_upsert_empty_report")
  build_report <- upsert_ns("fd_upsert_build_report")

  empty <- empty_report(
    feat_table_name = "features_tbl",
    key = "id",
    dialect = "sqlite",
    table_created = TRUE,
    columns_added = "score"
  )

  expect_s3_class(empty, "fd_upsert_report")
  expect_identical(empty$feat_table_name, "features_tbl")
  expect_identical(empty$key, "id")
  expect_identical(empty$dialect, "sqlite")
  expect_identical(empty$n_rows, 0L)
  expect_identical(empty$n_chunks, 0L)
  expect_true(empty$table_created)
  expect_identical(empty$columns_added, "score")
  expect_identical(empty$counts$would_insert, 0L)
  expect_identical(empty$counts$would_update, 0L)
  expect_equal(nrow(empty$chunk_details), 0L)

  chunk_details <- data.frame(
    chunk = 1:2,
    n = c(2L, 3L),
    would_insert = c(2L, 1L),
    would_update = c(0L, 2L)
  )
  report <- build_report(
    feat_table_name = "features_tbl",
    key = "id",
    dialect = "sqlite",
    n_rows = 5,
    n_chunks = 2,
    table_created = FALSE,
    columns_added = character(),
    totals_would_insert = 3,
    totals_would_update = 2,
    chunk_details = chunk_details
  )

  expect_s3_class(report, "fd_upsert_report")
  expect_identical(report$n_rows, 5L)
  expect_identical(report$n_chunks, 2L)
  expect_false(report$table_created)
  expect_identical(report$counts$would_insert, 3L)
  expect_identical(report$counts$would_update, 2L)
  expect_equal(report$chunk_details, chunk_details)
})

test_that("normalize_features_df converts factors and preserves date-like columns", {
  normalize_features_df <- upsert_ns("normalize_features_df")

  df <- data.frame(
    id = 1:2,
    group = factor(c("a", "b")),
    day = as.Date(c("2026-05-01", "2026-05-02")),
    stringsAsFactors = FALSE
  )
  df$time <- as.POSIXct(c("2026-05-01", "2026-05-02"), tz = "UTC")

  out <- normalize_features_df(df)

  expect_type(out$group, "character")
  expect_identical(out$group, c("a", "b"))
  expect_s3_class(out$day, "Date")
  expect_s3_class(out$time, "POSIXct")
})

test_that("get_stage_name creates unique dialect-safe names", {
  get_stage_name <- upsert_ns("get_stage_name")

  names <- replicate(20L, get_stage_name(3L))

  expect_true(all(grepl("^fd_stage_[0-9]{14}_3_[A-Za-z0-9]{8}$", names)))
  expect_equal(length(unique(names)), length(names))
})

test_that("make_chunks partitions row indices", {
  make_chunks <- upsert_ns("make_chunks")

  expect_equal(make_chunks(0L, NULL), list(integer()))
  expect_equal(make_chunks(5L, NULL), list(1:5))
  expect_equal(unname(make_chunks(5L, 2L)), list(1:2, 3:4, 5L))
  expect_equal(unname(make_chunks(3L, 10L)), list(1:3))
})

test_that("create_features_table creates a SQLite table with expected fields", {
  create_features_table <- upsert_ns("create_features_table")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  features <- data.frame(
    id = 1:2,
    score = c(10.5, 20.5),
    flag = c(TRUE, FALSE),
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )

  create_features_table(
    con = con,
    dialect = "sqlite",
    feat_table_name = "features_tbl",
    target_q = "`features_tbl`",
    features_df = features,
    key = "id"
  )

  expect_true(DBI::dbExistsTable(con, "features_tbl"))
  expect_equal(DBI::dbListFields(con, "features_tbl"), names(features))

  DBI::dbAppendTable(con, "features_tbl", features)
  roundtrip <- DBI::dbReadTable(con, "features_tbl")
  expect_equal(roundtrip$id, features$id)
  expect_equal(roundtrip$score, features$score)
  expect_equal(roundtrip$flag, as.integer(features$flag))
  expect_equal(roundtrip$label, features$label)
  expect_error(DBI::dbAppendTable(con, "features_tbl", features[1, , drop = FALSE]))
})

test_that("ensure_columns returns no additions when target schema already matches", {
  ensure_columns <- upsert_ns("ensure_columns")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  features <- data.frame(id = 1:2, score = c(10, 20))
  DBI::dbWriteTable(con, "features_tbl", features[0, , drop = FALSE], overwrite = TRUE)

  added <- ensure_columns(
    con = con,
    dialect = "sqlite",
    table_id = "features_tbl",
    table_q = "`features_tbl`",
    features_df = features,
    key = "id",
    alter_table = TRUE,
    verbose = FALSE
  )

  expect_identical(added, character())
})

test_that("ensure_columns adds missing columns when allowed", {
  ensure_columns <- upsert_ns("ensure_columns")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  DBI::dbWriteTable(
    con,
    "features_tbl",
    data.frame(id = integer(), score = numeric()),
    overwrite = TRUE
  )
  features <- data.frame(
    id = 1:2,
    score = c(10, 20),
    label = c("a", "b"),
    flag = c(TRUE, FALSE),
    stringsAsFactors = FALSE
  )

  added <- ensure_columns(
    con = con,
    dialect = "sqlite",
    table_id = "features_tbl",
    table_q = "`features_tbl`",
    features_df = features,
    key = "id",
    alter_table = TRUE,
    verbose = FALSE
  )

  expect_identical(added, c("label", "flag"))
  expect_equal(DBI::dbListFields(con, "features_tbl"), names(features))
})

test_that("ensure_columns errors for missing columns when alteration is disabled", {
  ensure_columns <- upsert_ns("ensure_columns")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  DBI::dbWriteTable(
    con,
    "features_tbl",
    data.frame(id = integer(), score = numeric()),
    overwrite = TRUE
  )

  expect_error(
    ensure_columns(
      con = con,
      dialect = "sqlite",
      table_id = "features_tbl",
      table_q = "`features_tbl`",
      features_df = data.frame(id = 1L, score = 10, label = "a"),
      key = "id",
      alter_table = FALSE,
      verbose = FALSE
    ),
    "missing columns"
  )
})

test_that("ensure_columns errors when target table lacks the key column", {
  ensure_columns <- upsert_ns("ensure_columns")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  DBI::dbWriteTable(
    con,
    "features_tbl",
    data.frame(other_id = integer(), score = numeric()),
    overwrite = TRUE
  )

  expect_error(
    ensure_columns(
      con = con,
      dialect = "sqlite",
      table_id = "features_tbl",
      table_q = "`features_tbl`",
      features_df = data.frame(id = 1L, score = 10),
      key = "id",
      alter_table = TRUE,
      verbose = FALSE
    ),
    "does not contain key column"
  )
})

test_that("count_scalars counts staged inserts and updates", {
  count_scalars <- upsert_ns("count_scalars")
  fd_sql_templates <- upsert_ns("fd_sql_templates")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  DBI::dbWriteTable(
    con,
    "features_tbl",
    data.frame(id = c(1L, 3L), score = c(10, 30)),
    overwrite = TRUE
  )
  DBI::dbWriteTable(
    con,
    "stage_tbl",
    data.frame(id = c(2L, 3L, 4L), score = c(20, 31, 40)),
    overwrite = TRUE
  )

  would_insert <- count_scalars(
    con,
    fd_sql_templates$upsert_count_would_insert$sqlite,
    target_q = "`features_tbl`",
    stage_q = "`stage_tbl`",
    key_q = "`id`"
  )
  would_update <- count_scalars(
    con,
    fd_sql_templates$upsert_count_would_update$sqlite,
    target_q = "`features_tbl`",
    stage_q = "`stage_tbl`",
    key_q = "`id`"
  )

  expect_identical(would_insert, 2L)
  expect_identical(would_update, 1L)
})

test_that("find_conflicts returns staged keys already present in target", {
  find_conflicts <- upsert_ns("find_conflicts")

  con <- fd_test_sqlite_con()
  on.exit(fd_test_disconnect(con), add = TRUE)

  DBI::dbWriteTable(
    con,
    "features_tbl",
    data.frame(id = c(1L, 3L, 5L), score = c(10, 30, 50)),
    overwrite = TRUE
  )
  DBI::dbWriteTable(
    con,
    "stage_tbl",
    data.frame(id = c(2L, 3L, 5L, 6L), score = c(20, 31, 51, 60)),
    overwrite = TRUE
  )

  conflicts <- find_conflicts(
    con = con,
    dialect = "sqlite",
    target_q = "`features_tbl`",
    stage_q = "`stage_tbl`",
    key_q = "`id`",
    limit = 1L
  )
  all_conflicts <- find_conflicts(
    con = con,
    dialect = "sqlite",
    target_q = "`features_tbl`",
    stage_q = "`stage_tbl`",
    key_q = "`id`",
    limit = 50L
  )

  expect_length(conflicts, 1L)
  expect_true(conflicts %in% c("3", "5"))
  expect_setequal(all_conflicts, c("3", "5"))
})

# Validation layer
#
# This file contains the package's central argument validator.
# The design intentionally keeps one main validation function
# (validate_general_args()) rather than splitting validation into many
# top-level functions.
#
# Rule of thumb:
# - use this validator for common scalar/type/range checks;
# - do not use it for operation-specific semantics such as table existence,
#   feature result shape, or SQL execution outcomes.




#' Validate common featdelta function arguments
#'
#' Central internal validator used across featdelta functions to check common
#' argument classes, scalar constraints, simple value ranges, and keyed-data
#' invariants.
#'
#' The function is intentionally broad: callers pass only the arguments they
#' want checked and use `NULL` for the rest. This keeps validation logic in a
#' single place while allowing different callers to validate different subsets
#' of arguments.
#'
#' The validator covers common categories such as:
#' \itemize{
#'   \item data frames (`data`, `features_df`)
#'   \item DB connections (`con`)
#'   \item SQL text (`sql`)
#'   \item key/table names
#'   \item defs objects
#'   \item grouped logical flags
#'   \item grouped environment arguments
#'   \item grouped positive or non-negative integer-like controls
#'   \item keyed-data invariants such as "key column exists", "key unique",
#'         and "key contains no NA"
#' }
#'
#' This function does not perform database-side checks such as whether a table
#' exists, whether a column exists in a database table, or whether an SQL query
#' can actually be executed.
#'
#' @param data Optional data frame/tibble to validate.
#' @param con Optional `DBIConnection` to validate.
#' @param sql Optional SQL query string to validate.
#' @param key Optional character scalar naming the primary key column.
#' @param defs Optional featdelta specification object to validate.
#' @param feat_table_name Optional character scalar naming the feature table.
#' @param features_df Optional data frame of computed features.
#' @param logicals_with_text Optional named list of scalar flags allowed to be
#'   either `TRUE`/`FALSE` or a small sentinel string such as `"auto"`.
#' @param logicals Optional named list of scalar logical flags.
#'
#' @return Invisibly returns `TRUE` on success. Throws an error on the first
#'   failed validation.
#'
#' @details
#' This function is designed to be read and maintained as one main validation
#' routine. Internal assertion closures may be used inside the function body
#' to reduce repetition, but the package should continue to expose and use a
#' single central validator rather than many top-level validation functions.
#'
#' @family featdelta validation helpers
#' @noRd
validate_general_args <- function(
    data = NULL,
    con = NULL,
    sql = NULL,
    key = NULL,
    defs = NULL,
    feat_table_name = NULL,
    features_df = NULL,
    compute_envir = NULL,
    logicals_with_text = NULL,
    logicals = NULL
) {

  ## data
  if(!is.null(data)) {
    if (!is.data.frame(data)) {
      stop("`data` must be a data.frame or tibble.")
    }
  }

  ## defs
  if(!is.null(defs)) {
    if(!inherits(defs, "featdelta_defs")) {
      stop("`defs` must be a `featdelta_defs`.")
    }
  }

  ## compute_envir
  if (!is.null(compute_envir) && !is.environment(compute_envir)) {
    stop("`compute_envir` must be an environment or NULL.")
  }

  ## con
  if(!is.null(con)) {
    if (!inherits(con, "DBIConnection")) {
      stop("`con` must be a DBIConnection (from DBI::dbConnect()).")
    }
    if (!DBI::dbIsValid(con)) {
      stop("`con` is not a valid (open) DBIConnection.")
    }
  }

  ## sql
  if(!is.null(sql)) {
    if (!is.character(sql) || length(sql) != 1L || !nzchar(sql)) {
      stop("`sql` must be a non-empty character scalar.")
    }

    if (!grepl("^\\s*(SELECT|WITH)\\b", sql, ignore.case = TRUE)) {
      stop("`sql` must be a SELECT query (may start with SELECT or WITH).")
    }
  }

  ## feat_table_name
  if(!is.null(feat_table_name)) {
    if (!is.character(feat_table_name) || length(feat_table_name) != 1L || !nzchar(feat_table_name)) {
      stop("`feat_table_name` must be a non-empty character scalar naming the features table.")
    }
  }

  ## key
  if(!is.null(key)) {
    if (!is.character(key) || length(key) != 1L || !nzchar(key)) {
      stop("`key` must be a non-empty string.")
    }
  }

  ## features_df
  if(!is.null(features_df)) {
    if (!is.data.frame(features_df)) {
      stop("`features_df` must be a data.frame.")
    }
  }

  ## key && features_df
  if(!is.null(key) && !is.null(features_df)) {
    if (!key %in% names(features_df)) {
      stop("`key` must be a column in `features_df`.")
    }
    if(anyDuplicated(features_df[[key]])) {
      stop("`features_df[[key]]` contains duplicates; keys must be unique.")
    }
    if(anyNA(features_df[[key]])) {
      stop("`features_df[[key]]` must not contain NAs.")
    }
  }

  ## key && data
  if(!is.null(key) && !is.null(data)) {
    if (!key %in% names(data)) {
      stop("`key` must be a column in `data`.")
    }
    if(anyDuplicated(data[[key]])) {
      stop("`data[[key]]` contains duplicates; keys must be unique.")
    }
    if(anyNA(data[[key]])) {
      stop("`data[[key]]` must not contain NAs.")
    }
  }

  ## handling logicals
  if(!is.null(logicals)) {
    for(i in 1:length(logicals)) {
      var_name <- names(logicals)[[i]]
      value <- logicals[[i]]
      if (!is.logical(value) || length(value) != 1L || is.na(value)) {
        stop(var_name, " must be TRUE/FALSE.")
      }
    }
  }

  ## handling logicals with text
  if(!is.null(logicals_with_text)) {
    for(i in 1:length(logicals_with_text)) {
      var_name <- names(logicals_with_text)[[i]]
      value <- logicals_with_text[[i]]

      if(length(value) != 1L || is.na(value)) {
        stop(var_name, " must be TRUE/FALSE, or 'auto'.")
      }
      if(!is.logical(value) && !is.character(value)) {
        stop(var_name, " must be TRUE/FALSE, or 'auto'.")
      }
      if(is.character(value) && value != 'auto') {
        stop(var_name, " must be TRUE/FALSE, or 'auto'.")
      }

    }
  }

  invisible(TRUE)
}

# Definition-definition layer
#
# This file defines the user-facing definitions object.
# fd_define() is responsible for definition-level normalization and guardrails.
#
# Rule of thumb:
# - fd_define() rejects malformed definitions and invalid step/output names;
# - fd_compute() evaluates a valid defs object against data;
# - fd_define() does not know the future key column or row count and should
#   not try to enforce compute-time invariants.



#' Define featdelta feature definitions
#'
#' Creates a `featdelta_defs` object that stores ordered definition steps in a
#' normalized internal representation suitable for later evaluation by
#' `fd_compute()`.
#'
#' `fd_define()` is the definitions-construction step of the package. It does
#' not compute features and does not access the database. Its role is to:
#' \itemize{
#'   \item capture user definitions,
#'   \item normalize supported input styles to a canonical internal structure,
#'   \item enforce definition-level guardrails such as name validity and
#'         duplicate handling,
#'   \item store the resulting ordered definition steps for later execution.
#' }
#'
#' Definitions are evaluated later by `fd_compute()` in the order they are
#' stored. This means later definitions may depend on columns created by
#' earlier steps. Reordering or removing earlier steps may therefore break
#' downstream definitions.
#'
#' @param ... Definitions supplied inline.
#'   Ordinary named inputs such as `new_feature = hp / cyl` are stored as
#'   single-column definition steps.
#' @param defs Optional list of programmatically supplied definitions.
#'   Supported inputs are the same as for `...`.
#' @param description Optional character scalar describing the definitions.
#' @param overwrite Logical. If `FALSE`, duplicate step names are rejected.
#'   If `TRUE`, later steps replace earlier ones.
#' @param envir Optional environment used as the explicit evaluation
#'   environment for normalized definitions. If `NULL`, environments are taken
#'   from the captured inputs.
#'
#' @return A list of class `"featdelta_defs"` containing:
#'   \itemize{
#'     \item `steps`: ordered normalized definition steps
#'     \item `description`: optional description
#'     \item `created_at`: creation time
#'     \item `defs_version`: internal version marker
#'     \item `envir_policy`: whether environments were inherited or explicit
#'     \item `envir`: the explicit environment, if supplied
#'   }
#'
#' @details
#' Supported definition inputs include:
#' \itemize{
#'   \item inline expressions captured from `...`
#'   \item quosures
#'   \item `expression()` objects of length 1
#'   \item quoted calls or names
#'   \item direct scalar constants
#'   \item `fd_block()` objects for multi-column feature generation
#' }
#'
#' Ordinary named expressions define one output column per step.
#'
#' `fd_block()` defines a multi-column step. At compute time, a block is
#' expected to return a `data.frame` with one row per input row. The returned
#' column names become the produced feature names. Optionally, expected output
#' names may be declared in advance via `fd_block(expected_names = ...)`.
#'
#' Direct atomic constants must be scalar. If vectorized behavior is desired,
#' provide an expression instead of embedding a fixed-length atomic vector in
#' the definitions.
#'
#' Duplicate step handling is part of definitions construction:
#' \itemize{
#'   \item with `overwrite = FALSE`, duplicate step names error;
#'   \item with `overwrite = TRUE`, the last definition wins.
#' }
#'
#' @examples
#' # Single-column definitions
#' defs <- fd_define(
#'   hp_per_cyl = hp / cyl,
#'   wt_per_hp = wt / hp
#' )
#'
#' # Definitions can also be supplied programmatically
#' predef <- expression(log(hp))
#' defs <- fd_define(
#'   log_hp = predef
#' )
#'
#' # Multi-column block using an inline expression
#' defs <- fd_define(
#'   engine_ratios = fd_block({
#'     data.frame(
#'       hp_per_cyl = hp / cyl,
#'       disp_per_cyl = disp / cyl
#'     )
#'   })
#' )
#'
#' # Multi-column block with declared expected names
#' defs <- fd_define(
#'   engine_ratios = fd_block(
#'     {
#'       data.frame(
#'         hp_per_cyl = hp / cyl,
#'         disp_per_cyl = disp / cyl
#'       )
#'     },
#'     expected_names = c("hp_per_cyl", "disp_per_cyl")
#'   )
#' )
#'
#' # Function-based block
#' make_engine_features <- function(data) {
#'   data.frame(
#'     hp_per_cyl = data$hp / data$cyl,
#'     disp_per_cyl = data$disp / data$cyl
#'   )
#' }
#'
#' defs <- fd_define(
#'   engine_ratios = fd_block(make_engine_features)
#' )
#'
#' # A block can contain a small feature script with temporary variables
#' defs <- fd_define(
#'   engine_script = fd_block({
#'     hp_per_cyl <- hp / cyl
#'     disp_per_cyl <- disp / cyl
#'
#'     data.frame(
#'       hp_per_cyl = hp_per_cyl,
#'       disp_per_cyl = disp_per_cyl,
#'       engine_index = hp_per_cyl + disp_per_cyl
#'     )
#'   })
#' )
#'
#' # A function-based block can create a variable number of columns in a loop
#' make_scaled_features <- function(data) {
#'   vars <- c("hp", "disp", "wt")
#'   out <- list()
#'
#'   for (var in vars) {
#'     center <- mean(data[[var]], na.rm = TRUE)
#'     spread <- stats::sd(data[[var]], na.rm = TRUE)
#'     out[[paste0(var, "_scaled")]] <- (data[[var]] - center) / spread
#'   }
#'
#'   as.data.frame(out)
#' }
#'
#' defs <- fd_define(
#'   scaled_inputs = fd_block(make_scaled_features)
#' )
#'
#' @family featdelta defs helpers
#' @export
fd_define <- function(...,
                      defs = NULL,
                      description = NULL,
                      overwrite = FALSE,
                      envir = NULL) {

  if (!is.null(description)) {
    if (!is.character(description) || length(description) != 1L) {
      stop("`description` must be a character scalar or NULL.")
    }
  }

  if (!is.logical(overwrite) || length(overwrite) != 1L || is.na(overwrite)) {
    stop("`overwrite` must be TRUE/FALSE.")
  }

  if (!is.null(envir) && !is.environment(envir)) {
    stop("`envir` must be an environment or NULL.")
  }

  dots <- rlang::enquos(..., .ignore_empty = "all")

  defs_in <- list()
  if (!is.null(defs)) {
    if (!is.list(defs)) {
      stop("`defs` must be a list or NULL.")
    }
    if (is.null(names(defs)) || any(!nzchar(names(defs)))) {
      stop("`defs` must be a *named* list.")
    }
    defs_in <- defs
  }

  all_inputs <- c(defs_in, dots)
  if (!length(all_inputs)) {
    stop("No definitions supplied.")
  }

  nm <- names(all_inputs)
  fd_define_validate_step_names(nm, allow_duplicates = isTRUE(overwrite))

  if (isTRUE(overwrite) && anyDuplicated(nm)) {
    all_inputs <- all_inputs[!duplicated(names(all_inputs), fromLast = TRUE)]
  }

  nm <- names(all_inputs)
  fd_define_validate_step_names(nm, allow_duplicates = FALSE)

  step_defs <- lapply(names(all_inputs), function(step_name) {
    x <- all_inputs[[step_name]]

    if (rlang::is_quosure(x)) {
      x <- maybe_unwrap_symbol(x)
      if (rlang::is_quosure(x)) {
        x <- maybe_eval_fd_block_call(x)
      }
      default_env <- if (rlang::is_quosure(x)) {
        rlang::get_env(x)
      } else if (inherits(x, "featdelta_block")) {
        x$env %||% rlang::caller_env()
      } else {
        rlang::caller_env()
      }
    } else {
      default_env <- rlang::caller_env()
    }

    target_env <- envir %||% default_env

    normalize_definition_step(
      x = x,
      env = target_env,
      step_name = step_name
    )
  })

  names(step_defs) <- names(all_inputs)
  fd_define_validate_step_defs(step_defs)

  out <- list(
    steps = step_defs,
    description = description,
    created_at = Sys.time(),
    defs_version = 2L,
    envir_policy = if (is.null(envir)) "caller" else "explicit",
    envir = envir
  )

  class(out) <- c("featdelta_defs", "list")
  out
}

#' Print a featdelta definitions object
#'
#' Prints a concise human-readable summary of a `featdelta_defs` object.
#'
#' @param x A `featdelta_defs` object.
#' @param ... Unused.
#'
#' @return The input object invisibly.
#'
#' @family featdelta defs helpers
#' @export
print.featdelta_defs <- function(x, ...) {
  cat("<featdelta_defs>", "\n", sep = "")

  if (!is.null(x$description)) {
    cat("Description: ", x$description, "\n", sep = "")
  }

  cat("Definition steps (", length(x$steps), "):\n", sep = "")

  for (i in seq_along(x$steps)) {
    step <- x$steps[[i]]
    step_name <- names(x$steps)[[i]]

    if (identical(step$type, "column")) {
      cat(
        "  - [column] ",
        step_name,
        " -> ",
        step$text,
        "\n",
        sep = ""
      )

    } else if (identical(step$type, "block")) {
      cat(
        "  - [block]  ",
        step_name,
        " -> ",
        step$text,
        "\n",
        sep = ""
      )

      if (!is.null(step$output_names)) {
        cat(
          "      expected: ",
          paste(step$output_names, collapse = ", "),
          "\n",
          sep = ""
        )
      }

    } else {
      cat("  - [unknown] ", step_name, "\n", sep = "")
    }
  }

  invisible(x)
}


#' Summarize a featdelta definitions object
#'
#' Returns a compact data-frame summary of the ordered definition steps stored
#' in a `featdelta_defs` object.
#'
#' @param object A `featdelta_defs` object.
#' @param ... Unused.
#'
#' @return A data frame with one row per definition step and columns such as
#'   step name, type, mode, declared outputs, and stored expression text.
#'
#' @family featdelta defs helpers
#' @export
summary.featdelta_defs <- function(object, ...) {
  data.frame(
    step_name = names(object$steps),
    type = vapply(object$steps, function(s) s$type, character(1)),
    mode = vapply(
      object$steps,
      function(s) s$mode %||% NA_character_,
      character(1)
    ),
    outputs = vapply(
      object$steps,
      function(s) paste(s$output_names %||% character(0), collapse = ", "),
      character(1)
    ),
    text = vapply(
      object$steps,
      function(s) s$text %||% NA_character_,
      character(1)
    ),
    stringsAsFactors = FALSE
  )
}


#' Normalize one user-supplied definition into a step object
#'
#' Internal helper used by `fd_define()` to convert one supported input style
#' into the package's canonical stored step representation.
#'
#' At present, ordinary inputs are normalized to `type = "column"` steps.
#' The returned structure is intentionally general enough to support future
#' step types such as `type = "block"`.
#'
#' @param x A single user-supplied definition.
#' @param env Environment used when constructing the normalized quosure.
#' @param step_name Step name used for clearer error messages.
#'
#' @return A normalized step object.
#'
#' @family featdelta defs helpers
#' @noRd
normalize_definition_step <- function(x, env, step_name) {

  if (inherits(x, "featdelta_block")) {
    return(normalize_block_step(x = x, env = env, step_name = step_name))
  }

  q <- normalize_feature_input(x, env = env, feature_name = step_name)
  e <- rlang::get_expr(q)

  list(
    type = "column",
    mode = "expression",
    step_name = step_name,
    output_names = step_name,
    fn = NULL,
    quo = q,
    expr = e,
    env = rlang::get_env(q),
    text = rlang::expr_text(e)
  )
}


#' Normalize a block definition into a step object
#'
#' Internal placeholder for future `fd_block()` support.
#'
#' @param x A block definition object.
#' @param env Environment used for evaluation.
#' @param step_name Step name used for clearer error messages.
#'
#' @return A normalized block step object.
#'
#' @family featdelta defs helpers
#' @noRd
normalize_block_step <- function(x, env, step_name) {
  list(
    type = "block",
    mode = x$mode,
    step_name = step_name,
    output_names = x$expected_names %||% NULL,
    fn = x$fn %||% NULL,
    quo = x$quo %||% NULL,
    expr = x$expr %||% NULL,
    env = x$env %||% env,
    text = x$text %||% "<block>"
  )
}


#' Normalize a single feature-like input to a canonical quosure representation
#'
#' Internal helper used by `fd_define()` to convert one supported single-column
#' definition input style into the package's canonical stored form.
#'
#' Supported input styles include:
#' \itemize{
#'   \item quosures
#'   \item `expression()` objects of length 1
#'   \item quoted calls
#'   \item quoted names
#'   \item direct scalar constants
#' }
#'
#' Direct atomic constants must be scalar. Multi-element atomic constants are
#' rejected because they encode row-count assumptions into the definitions and
#' belong in compute-time expressions instead.
#'
#' @param x A single user-supplied single-column definition.
#' @param env Environment used when constructing the normalized quosure.
#' @param feature_name Optional name used for clearer error messages.
#'
#' @return A quosure representing the normalized single-column definition.
#'
#' @family featdelta defs helpers
#' @noRd
normalize_feature_input <- function(x, env, feature_name = NULL) {
  nm <- feature_name %||% "<unknown>"

  if (rlang::is_quosure(x)) {
    expr <- rlang::get_expr(x)
    return(rlang::new_quosure(expr, env = env))
  }

  if (inherits(x, "expression")) {
    if (length(x) != 1L) {
      stop(
        sprintf(
          "Definition `%s` was supplied as `expression(...)` with length %d. ",
          nm, length(x)
        ),
        "Each single-column definition must correspond to exactly one expression.",
        call. = FALSE
      )
    }
    return(rlang::new_quosure(x[[1]], env = env))
  }

  if (is.call(x) || is.name(x)) {
    return(rlang::new_quosure(x, env = env))
  }

  if (is.atomic(x)) {
    if (length(x) != 1L) {
      stop(
        sprintf(
          "Definition `%s` was supplied as an atomic constant of length %d. ",
          nm, length(x)
        ),
        "Direct constants must be scalar. ",
        "If you want vectorized behavior, supply an expression instead.",
        call. = FALSE
      )
    }
    return(rlang::new_quosure(x, env = env))
  }

  stop(
    sprintf(
      "Definition `%s` must be an expression, quosure, call, symbol, or scalar constant.",
      nm
    )
  )
}


#' Maybe unwrap a captured symbol that points to a quoted expression container
#'
#' Internal helper used by `fd_define()` when a user writes a definition
#' through an intermediate object, for example:
#'
#' `predef <- expression(hp / cyl)`
#'
#' `fd_define(x = predef)`
#'
#' In such cases, tidy capture initially sees the symbol `predef` rather than
#' the expression stored inside it. This helper resolves the symbol only when
#' it refers to a recognized quoted-expression container such as:
#' \itemize{
#'   \item a quosure
#'   \item an `expression()` object
#'   \item a quoted language object
#' }
#'
#' Ordinary values are intentionally left untouched so that `fd_define()` does
#' not silently rewrite arbitrary objects from the calling environment.
#'
#' @param q A quosure captured from `...`.
#'
#' @return Either the original quosure or a normalized quosure built from the
#'   referenced quoted-expression container.
#'
#' @family featdelta defs helpers
#' @noRd
maybe_unwrap_symbol <- function(q) {
  if (!rlang::is_quosure(q)) {
    stop("`q` must be a quosure.", call. = FALSE)
  }

  expr <- rlang::get_expr(q)
  env  <- rlang::get_env(q)

  if (!is.symbol(expr)) {
    return(q)
  }

  name <- as.character(expr)

  if (!rlang::env_has(env, name, inherit = TRUE)) {
    return(q)
  }

  value <- rlang::env_get(env, name, inherit = TRUE)

  if (inherits(value, "featdelta_block")) {
    return(value)
  }

  if (rlang::is_quosure(value)) {
    return(value)
  }

  if (inherits(value, "expression")) {
    if (length(value) != 1L) {
      stop(
        sprintf(
          "Definition `%s` refers to an `expression()` object of length %d. ",
          name, length(value)
        ),
        "Each single-column definition must correspond to exactly one expression.",
        call. = FALSE
      )
    }
    return(rlang::new_quosure(value[[1]], env = env))
  }

  if (is.call(value) || is.name(value)) {
    return(rlang::new_quosure(value, env = env))
  }

  q
}


#' Maybe evaluate a captured `fd_block()` call
#'
#' Internal helper used by `fd_define()` to support inline block definitions
#' such as `fd_define(x = fd_block({ data.frame(...) }))`.
#'
#' Ordinary feature expressions are intentionally left unevaluated. Only direct
#' calls to `fd_block()` are evaluated at definition-construction time.
#'
#' @param q A quosure captured from `...`.
#'
#' @return Either a `featdelta_block` object or the original quosure.
#'
#' @family featdelta defs helpers
#' @noRd
maybe_eval_fd_block_call <- function(q) {
  if (!rlang::is_quosure(q)) {
    stop("`q` must be a quosure.", call. = FALSE)
  }

  expr <- rlang::get_expr(q)

  if (!is.call(expr)) {
    return(q)
  }

  head <- expr[[1]]
  is_fd_block_call <- is.symbol(head) && identical(as.character(head), "fd_block")

  if (!is_fd_block_call) {
    return(q)
  }

  value <- rlang::eval_tidy(expr, env = rlang::get_env(q))

  if (!inherits(value, "featdelta_block")) {
    stop("A direct `fd_block()` call must return a `featdelta_block` object.", call. = FALSE)
  }

  value
}


#' Validate step names at definition-construction time
#'
#' Internal helper used by `fd_define()` to enforce the package's step-name
#' policy before a defs object is created.
#'
#' @param nm Character vector of proposed step names.
#' @param allow_duplicates Logical. If `TRUE`, duplicate names are permitted
#'   temporarily (for example before `overwrite = TRUE` resolution). If
#'   `FALSE`, duplicates error.
#'
#' @return Invisibly returns `TRUE` on success. Errors on invalid names.
#'
#' @family featdelta defs helpers
#' @noRd
fd_define_validate_step_names <- function(nm, allow_duplicates = FALSE) {
  if (is.null(nm) || any(is.na(nm)) || any(!nzchar(nm))) {
    stop("All definition steps must be named with non-empty, non-NA names.", call. = FALSE)
  }

  reserved <- c(".key", ".rowid", ".featdelta")
  bad <- nm %in% reserved
  if (any(bad)) {
    stop(
      sprintf(
        "Reserved definition step name(s): %s",
        paste(unique(nm[bad]), collapse = ", ")
      ),
      call. = FALSE
    )
  }

  if (!isTRUE(allow_duplicates) && anyDuplicated(nm)) {
    dups <- unique(nm[duplicated(nm)])
    stop(
      sprintf(
        "Duplicate definition step name(s): %s",
        paste(dups, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}


#' Validate normalized step definitions before returning defs
#'
#' Internal helper used by `fd_define()` after normalization to verify that the
#' final `step_defs` object is internally consistent.
#'
#' @param step_defs Named list of normalized definition step objects.
#'
#' @return Invisibly returns `TRUE` on success. Errors on malformed normalized
#'   steps.
#'
#' @family featdelta defs helpers
#' @noRd
fd_define_validate_step_defs <- function(step_defs) {
  if (!is.list(step_defs) || !length(step_defs)) {
    stop("`step_defs` must be a non-empty list.", call. = FALSE)
  }

  nm <- names(step_defs)
  fd_define_validate_step_names(nm, allow_duplicates = FALSE)

  for (i in seq_along(step_defs)) {
    step <- step_defs[[i]]
    step_name_i <- nm[[i]]

    if (!is.list(step)) {
      stop(
        "Definition step `", step_name_i, "` must be stored as a list.",
        call. = FALSE
      )
    }

    required_common <- c("type", "step_name", "output_names", "env", "text")
    missing_common <- setdiff(required_common, names(step))
    if (length(missing_common)) {
      stop(
        "Definition step `", step_name_i, "` is malformed; missing field(s): ",
        paste(missing_common, collapse = ", "),
        call. = FALSE
      )
    }

    if (!identical(step$step_name, step_name_i)) {
      stop(
        "Definition step `", step_name_i,
        "` is malformed: stored `step_name` does not match list name.",
        call. = FALSE
      )
    }

    if (!is.character(step$type) || length(step$type) != 1L || is.na(step$type)) {
      stop(
        "Definition step `", step_name_i, "` has invalid `type`.",
        call. = FALSE
      )
    }

    if (!is.environment(step$env)) {
      stop(
        "Definition step `", step_name_i, "` has invalid `env`.",
        call. = FALSE
      )
    }

    if (!is.character(step$text) || length(step$text) != 1L || is.na(step$text)) {
      stop(
        "Definition step `", step_name_i, "` has invalid `text`.",
        call. = FALSE
      )
    }

    if (!is.null(step$output_names)) {
      if (!is.character(step$output_names) ||
          any(is.na(step$output_names)) ||
          any(!nzchar(step$output_names))) {
        stop(
          "Definition step `", step_name_i, "` has invalid `output_names`.",
          call. = FALSE
        )
      }

      if (anyDuplicated(step$output_names)) {
        stop(
          "Definition step `", step_name_i, "` has duplicate `output_names`.",
          call. = FALSE
        )
      }
    }

    if (identical(step$type, "column")) {
      required_column <- c("quo", "expr")
      missing_column <- setdiff(required_column, names(step))
      if (length(missing_column)) {
        stop(
          "Column step `", step_name_i, "` is malformed; missing field(s): ",
          paste(missing_column, collapse = ", "),
          call. = FALSE
        )
      }

      if (!rlang::is_quosure(step$quo)) {
        stop(
          "Column step `", step_name_i, "` has invalid `quo`.",
          call. = FALSE
        )
      }

      if (!identical(step$output_names, step_name_i)) {
        stop(
          "Column step `", step_name_i,
          "` must have exactly one output name equal to the step name.",
          call. = FALSE
        )
      }

    } else if (identical(step$type, "block")) {
      required_block <- c("mode", "fn", "quo", "expr")
      missing_block <- setdiff(required_block, names(step))
      if (length(missing_block)) {
        stop(
          "Block step `", step_name_i, "` is malformed; missing field(s): ",
          paste(missing_block, collapse = ", "),
          call. = FALSE
        )
      }

      if (!is.character(step$mode) || length(step$mode) != 1L || is.na(step$mode)) {
        stop(
          "Block step `", step_name_i, "` has invalid `mode`.",
          call. = FALSE
        )
      }

      if (identical(step$mode, "function")) {
        if (!is.function(step$fn)) {
          stop(
            "Block step `", step_name_i, "` has invalid `fn` for `mode = 'function'`.",
            call. = FALSE
          )
        }

        if (!is.null(step$quo)) {
          stop(
            "Block step `", step_name_i,
            "` must have `quo = NULL` for `mode = 'function'`.",
            call. = FALSE
          )
        }

      } else if (identical(step$mode, "expression")) {
        if (!rlang::is_quosure(step$quo)) {
          stop(
            "Block step `", step_name_i,
            "` has invalid `quo` for `mode = 'expression'`.",
            call. = FALSE
          )
        }

        if (!is.null(step$fn)) {
          stop(
            "Block step `", step_name_i,
            "` must have `fn = NULL` for `mode = 'expression'`.",
            call. = FALSE
          )
        }

      } else {
        stop(
          "Block step `", step_name_i, "` has unsupported `mode`: ",
          step$mode,
          call. = FALSE
        )
      }

    } else {
      stop(
        "Definition step `", step_name_i, "` has unsupported `type`: ",
        step$type,
        call. = FALSE
      )
    }
  }

  invisible(TRUE)
}

test_that("fd_define constructs ordered column definitions from dots", {
  defs <- fd_define(
    ratio = x / y,
    doubled = ratio * 2,
    description = "basic definitions"
  )

  expect_s3_class(defs, "featdelta_defs")
  expect_named(defs$steps, c("ratio", "doubled"))
  expect_identical(defs$description, "basic definitions")
  expect_identical(defs$defs_version, 2L)
  expect_identical(defs$envir_policy, "caller")
  expect_null(defs$envir)
  expect_s3_class(defs$created_at, "POSIXct")

  expect_identical(defs$steps$ratio$type, "column")
  expect_identical(defs$steps$ratio$mode, "expression")
  expect_identical(defs$steps$ratio$step_name, "ratio")
  expect_identical(defs$steps$ratio$output_names, "ratio")
  expect_true(rlang::is_quosure(defs$steps$ratio$quo))
  expect_true(is.environment(defs$steps$ratio$env))

  expect_identical(defs$steps$doubled$type, "column")
  expect_identical(defs$steps$doubled$step_name, "doubled")
  expect_identical(defs$steps$doubled$output_names, "doubled")
})

test_that("fd_define accepts programmatic named defs inputs", {
  defs <- fd_define(defs = list(
    ratio = expression(x / y),
    total = quote(x + y),
    constant = 1,
    label = "ok"
  ))

  expect_s3_class(defs, "featdelta_defs")
  expect_named(defs$steps, c("ratio", "total", "constant", "label"))
  expect_identical(
    vapply(defs$steps, `[[`, character(1), "type"),
    c(ratio = "column", total = "column", constant = "column", label = "column")
  )
  expect_match(defs$steps$ratio$text, "x")
  expect_match(defs$steps$ratio$text, "y")
  expect_match(defs$steps$total$text, "x")
  expect_match(defs$steps$total$text, "y")
  expect_identical(defs$steps$constant$text, "1")
  expect_match(defs$steps$label$text, "ok")
})

test_that("fd_define unwraps captured symbols that hold quoted expressions", {
  ratio_expr <- expression(x / y)
  total_call <- quote(x + y)

  defs <- fd_define(
    ratio = ratio_expr,
    total = total_call
  )

  expect_named(defs$steps, c("ratio", "total"))
  expect_match(defs$steps$ratio$text, "x")
  expect_match(defs$steps$ratio$text, "y")
  expect_false(grepl("ratio_expr", defs$steps$ratio$text, fixed = TRUE))
  expect_match(defs$steps$total$text, "x")
  expect_match(defs$steps$total$text, "y")
  expect_false(grepl("total_call", defs$steps$total$text, fixed = TRUE))
})

test_that("fd_define uses explicit environments when supplied", {
  env <- new.env(parent = baseenv())
  env$offset <- 10

  defs <- fd_define(
    shifted = x + offset,
    envir = env
  )

  expect_identical(defs$envir_policy, "explicit")
  expect_identical(defs$envir, env)
  expect_identical(defs$steps$shifted$env, env)
})

test_that("fd_define handles duplicate names according to overwrite", {
  expect_error(
    fd_define(score = x, score = y),
    "Duplicate definition step"
  )

  defs <- fd_define(score = x, score = y, overwrite = TRUE)

  expect_named(defs$steps, "score")
  expect_length(defs$steps, 1L)
  expect_match(defs$steps$score$text, "y")
  expect_false(grepl("^x$", defs$steps$score$text))
})

test_that("fd_define rejects malformed construction inputs", {
  expect_error(fd_define(), "No definitions supplied")
  expect_error(fd_define(x), "All definition steps")
  expect_error(fd_define(defs = 1), "`defs` must be a list")
  expect_error(fd_define(defs = list(1)), "`defs` must be a \\*named\\* list")
  expect_error(fd_define(.key = x), "Reserved definition step name")
  expect_error(fd_define(defs = list(value = 1:2)), "Direct constants must be scalar")
  expect_error(fd_define(defs = list(value = expression(x, y))), "exactly one expression")
  expect_error(fd_define(value = x, description = c("a", "b")), "`description`")
  expect_error(fd_define(value = x, overwrite = NA), "`overwrite`")
  expect_error(fd_define(value = x, envir = list()), "`envir`")
})

test_that("summary.featdelta_defs returns one row per definition step", {
  defs <- fd_define(
    ratio = x / y,
    constant = 1
  )

  out <- summary(defs)

  expect_s3_class(out, "data.frame")
  expect_named(out, c("step_name", "type", "mode", "outputs", "text"))
  expect_equal(out$step_name, c("ratio", "constant"))
  expect_equal(out$type, c("column", "column"))
  expect_equal(out$mode, c("expression", "expression"))
  expect_equal(out$outputs, c("ratio", "constant"))
  expect_true(all(nzchar(out$text)))
})

test_that("print.featdelta_defs returns input invisibly and shows key fields", {
  defs <- fd_define(
    ratio = x / y,
    description = "printable definitions"
  )

  printed <- capture.output(ret <- print(defs))

  expect_identical(ret, defs)
  expect_true(any(grepl("<featdelta_defs>", printed, fixed = TRUE)))
  expect_true(any(grepl("Description: printable definitions", printed, fixed = TRUE)))
  expect_true(any(grepl("Definition steps (1):", printed, fixed = TRUE)))
  expect_true(any(grepl("[column] ratio", printed, fixed = TRUE)))
})

# Compute layer
#
# This file contains in-memory feature computation. The functions here do not
# know anything about database context objects and should not call resolve_ctx().
#
# Rule of thumb:
# - fd_define() creates and normalizes a defs;
# - fd_compute() evaluates that defs against an in-memory data frame;
# - DB-facing functions such as fd_fetch()/fd_upsert()/fd_run() belong to the
#   context/database layer, not here.



#' Compute features from featdelta definitions on in-memory data
#'
#' Evaluates a `featdelta_defs` object created by `fd_define()` against an
#' in-memory data frame/tibble and returns a feature data frame containing the
#' key column plus computed feature columns.
#'
#' `fd_compute()` is the package's pure in-memory execution step. It does not
#' access the database and does not use the featdelta context system.
#'
#' Definition steps are evaluated sequentially in the order stored in `defs`.
#' This means later steps may depend on columns created by earlier steps in the
#' same compute call.
#'
#' Guardrails implemented by `fd_compute()` are compute-specific:
#' \itemize{
#'   \item validate the structure of the supplied defs object
#'   \item validate result type and row alignment for each step
#'   \item preserve 1:1 row alignment with the input data
#'   \item reject output-name collisions with `key` and previously produced names
#'   \item attach a per-step computation report
#' }
#'
#' Guardrails that belong at definition time, such as malformed construction of
#' individual definition steps, should be handled by `fd_define()`.
#'
#' @param data A data frame/tibble containing the raw input variables used to
#'   compute features.
#' @param defs A `featdelta_defs` object created by `fd_define()`.
#' @param key Character scalar naming the primary key column in `data`.
#' @param compute_envir Optional environment. If supplied, it overrides the
#'   stored evaluation environment for expression-based steps.
#' @param compute_strict Logical. If `TRUE`, any failed step causes the overall
#'   compute call to error after the per-step report is assembled. If `FALSE`,
#'   failed column steps return `NA` columns, and failed block steps return
#'   `NA` columns only when `output_names` are known in advance.
#' @param verbose Logical. If `TRUE`, print progress messages and a short
#'   summary.
#' @param return_report Logical. If `TRUE`, return a list with elements
#'   `data` and `report`. If `FALSE`, return only the computed data frame and
#'   attach the report as `attr(x, "fd_report")`.
#'
#' @return Either:
#'   \itemize{
#'     \item a data frame with class `"featdelta_features"` and attached report
#'           attribute, or
#'     \item a list `{data, report}` with class
#'           `"featdelta_compute_result"` when `return_report = TRUE`
#'   }
#'
#' The output preserves row order and row count relative to `data`.
#'
#' @details
#' Supported definition step types are:
#' \itemize{
#'   \item ordinary single-column steps
#'   \item `fd_block()` multi-column steps
#' }
#'
#' For single-column steps, the result must be a supported vector-like output.
#'
#' For block steps:
#' \itemize{
#'   \item the result must be a `data.frame`
#'   \item the number of rows must match `nrow(data)`
#'   \item returned column names must be non-empty and unique
#'   \item returned names must not collide with `key`
#'   \item returned names must not collide with feature names already produced
#'         earlier in the same compute run
#' }
#'
#' If a block declares `expected_names`, then:
#' \itemize{
#'   \item returned names must be a subset of `expected_names`
#'   \item missing expected names are added as `NA`
#'   \item final output order follows `expected_names`
#' }
#'
#' Because definitions are evaluated sequentially, later steps may use:
#' \itemize{
#'   \item raw input columns from `data`
#'   \item columns produced by earlier single-column steps
#'   \item columns produced by earlier block steps
#' }
#'
#' Reordering or removing earlier steps may therefore break downstream
#' definitions. When evaluation fails with "object not found"-style errors,
#' `fd_compute()` augments the message with likely reasons.
#'
#' @examples
#' raw_df <- mtcars
#' raw_df$car_id <- seq_len(nrow(raw_df))
#'
#' defs <- fd_define(
#'   hp_per_cyl = hp / cyl,
#'   engine_ratios = fd_block({
#'     data.frame(
#'       disp_per_cyl = disp / cyl,
#'       wt_per_hp = wt / hp
#'     )
#'   }),
#'   double_ratio = disp_per_cyl * 2
#' )
#'
#' out <- fd_compute(
#'   data = raw_df,
#'   defs = defs,
#'   key = "car_id",
#'   compute_strict = TRUE
#' )
#'
#' head(out)
#'
#' @family featdelta compute helpers
#' @export
fd_compute <- function(data,
                       defs,
                       key,
                       compute_envir = NULL,
                       compute_strict = TRUE,
                       verbose = FALSE,
                       return_report = FALSE) {

  ## validation of inputs
  validate_general_args(
    data = data,
    key = key,
    defs = defs,
    compute_envir = compute_envir,

    logicals = list(
      compute_strict = compute_strict,
      verbose = verbose,
      return_report = return_report
    )
  )

  fd_compute_validate_defs_structure(defs, key)


  ##
  steps <- defs$steps
  n <- nrow(data)
  if (verbose) {
    message("fd_compute(): n = ", n, ", steps = ", length(steps))
  }

  # working data is what function-blocks receive
  working_data <- data

  # Evaluation data contains raw columns plus features produced so far.
  # Keep this as a data frame/list, not an environment, because passing a plain
  # environment as `data` to rlang::eval_tidy() is deprecated.
  mask <- working_data


  # evaluate features sequentially
  produced_cols <- list()
  produced_names <- character(0)
  report <- vector("list", length(steps))

  step_names <- names(steps)

  for (i in seq_along(steps)) {
    step <- steps[[i]]
    step_name_i <- step_names[[i]]

    t0 <- Sys.time()

    step_result <- fd_compute_eval_step(
      step = step,
      step_name = step_name_i,
      working_data = working_data,
      mask = mask,
      n = n,
      key = key,
      produced_names = produced_names,
      compute_envir = compute_envir,
      compute_strict = compute_strict
    )

    t1 <- Sys.time()

    # In non-strict mode failed steps may provide NA fallback columns.
    if ((step_result$ok || !isTRUE(compute_strict)) && length(step_result$cols)) {
      new_names <- names(step_result$cols)

      for (j in seq_along(step_result$cols)) {
        nm_j <- new_names[[j]]
        col_j <- step_result$cols[[j]]

        produced_cols[[nm_j]] <- col_j
        produced_names <- c(produced_names, nm_j)
        working_data[[nm_j]] <- col_j
        mask[[nm_j]] <- col_j
      }
    }

    report[[i]] <- list(
      step_name = step_name_i,
      type = step$type %||% NA_character_,
      mode = step$mode %||% NA_character_,
      ok = step_result$ok,
      error = if (step_result$ok) NA_character_ else step_result$error,
      error_raw = if (step_result$ok) NA_character_ else step_result$error_raw,
      error_class = if (step_result$ok) NA_character_ else step_result$error_class,
      hint = if (step_result$ok) NA_character_ else step_result$hint,
      time_sec = as.numeric(difftime(t1, t0, units = "secs")),
      text = step$text %||% NA_character_,
      output_names = paste(names(step_result$cols), collapse = ", ")
    )

    if (verbose) {
      message(sprintf("  [%s] %s", if (step_result$ok) "OK" else "FAIL", step_name_i))
    }
  }

  report_df <- data.frame(
    step_name = vapply(report, `[[`, character(1), "step_name"),
    type = vapply(report, `[[`, character(1), "type"),
    mode = vapply(report, `[[`, character(1), "mode"),
    ok = vapply(report, `[[`, logical(1), "ok"),
    error = vapply(report, `[[`, character(1), "error"),
    error_raw = vapply(report, `[[`, character(1), "error_raw"),
    error_class = vapply(report, `[[`, character(1), "error_class"),
    hint = vapply(report, `[[`, character(1), "hint"),
    time_sec = vapply(report, `[[`, numeric(1), "time_sec"),
    text = vapply(report, `[[`, character(1), "text"),
    output_names = vapply(report, `[[`, character(1), "output_names"),
    stringsAsFactors = FALSE
  )

  any_fail <- any(!report_df$ok)
  if (verbose) {
    message("fd_compute(): success = ", !any_fail)
  }

  out <- data.frame(
    c(stats::setNames(list(data[[key]]), key), produced_cols),
    check.names = FALSE,
    stringsAsFactors = FALSE
  )

  if (nrow(out) != n) {
    stop(
      "Internal fd_compute error: output row count does not match input row count.",
      call. = FALSE
    )
  }

  if (length(produced_names)) {
    expected_names <- c(key, produced_names)
  } else {
    expected_names <- key
  }

  if (!identical(names(out), expected_names)) {
    stop(
      "Internal fd_compute error: output column names do not match expected names.",
      call. = FALSE
    )
  }

  attr(out, "fd_report") <- report_df
  attr(out, "fd_key") <- key
  class(out) <- c("featdelta_features", class(out))

  if (any_fail && isTRUE(compute_strict)) {
    e <- simpleError("fd_compute(): one or more definition steps failed (see report).")
    attr(e, "fd_report") <- report_df
    attr(e, "fd_partial") <- out
    class(e) <- c("featdelta_compute_error", class(e))
    stop(e)
  }

  if (isTRUE(return_report)) {
    res <- list(data = out, report = report_df)
    class(res) <- c("featdelta_compute_result", "list")
    return(res)
  }

  out
}



#' Validate the internal structure of a featdelta defs for computation
#'
#' Defensive internal helper used by `fd_compute()` to verify that the supplied
#' defs object is structurally sound before evaluation begins.
#'
#' This validation is intentionally compute-focused. It checks the shape and
#' consistency of `defs$features`, including:
#' \itemize{
#'   \item non-empty named feature list
#'   \item unique feature names
#'   \item no feature name collision with `key`
#'   \item presence and basic validity of stored fields such as `quo`, `expr`,
#'         `env`, and `text`
#' }
#'
#' Although `fd_define()` should normally construct a valid defs, this helper
#' protects `fd_compute()` against manually modified or malformed defs objects.
#'
#' @param defs A `featdelta_defs`.
#' @param key Character scalar naming the key column used for the current
#'   compute call.
#'
#' @return Invisibly returns `TRUE` on success. Errors on malformed defs.
#'
#' @family featdelta compute helpers
#' @noRd
fd_compute_validate_defs_structure <- function(defs, key = NULL) {
  steps <- defs$steps

  if (!is.list(steps) || !length(steps)) {
    stop("`defs$steps` must be a non-empty list.", call. = FALSE)
  }

  nm <- names(steps)
  if (is.null(nm) || any(is.na(nm)) || any(!nzchar(nm))) {
    stop("`defs$steps` must be a named list.", call. = FALSE)
  }

  if (anyDuplicated(nm)) {
    stop("`defs$steps` contains duplicate step names.", call. = FALSE)
  }

  for (i in seq_along(steps)) {
    step <- steps[[i]]
    step_name_i <- nm[[i]]

    if (!is.list(step)) {
      stop(
        "Definition step `", step_name_i, "` must be stored as a list.",
        call. = FALSE
      )
    }

    if (!is.character(step$type) || length(step$type) != 1L || is.na(step$type)) {
      stop(
        "Definition step `", step_name_i, "` has malformed `type`.",
        call. = FALSE
      )
    }

    if (!is.character(step$step_name) || length(step$step_name) != 1L || is.na(step$step_name)) {
      stop(
        "Definition step `", step_name_i, "` has malformed `step_name`.",
        call. = FALSE
      )
    }

    if (!identical(step$step_name, step_name_i)) {
      stop(
        "Definition step `", step_name_i, "` has inconsistent `step_name`.",
        call. = FALSE
      )
    }

    if (!is.null(step$output_names)) {
      if (!is.character(step$output_names) ||
          any(is.na(step$output_names)) ||
          any(!nzchar(step$output_names))) {
        stop(
          "Definition step `", step_name_i, "` has malformed `output_names`.",
          call. = FALSE
        )
      }

      if (anyDuplicated(step$output_names)) {
        stop(
          "Definition step `", step_name_i, "` has duplicate `output_names`.",
          call. = FALSE
        )
      }

      if (!is.null(key) && any(step$output_names %in% key)) {
        stop(
          "Definition step `", step_name_i,
          "` declares output name(s) that collide with `key`.",
          call. = FALSE
        )
      }
    }

    if (!is.environment(step$env)) {
      stop(
        "Definition step `", step_name_i, "` has malformed `env`.",
        call. = FALSE
      )
    }

    if (!is.character(step$text) || length(step$text) != 1L || is.na(step$text)) {
      stop(
        "Definition step `", step_name_i, "` has malformed `text`.",
        call. = FALSE
      )
    }

    if (identical(step$type, "column")) {
      if (!rlang::is_quosure(step$quo)) {
        stop(
          "Column step `", step_name_i, "` has malformed `quo`.",
          call. = FALSE
        )
      }

      if (!identical(step$output_names, step_name_i)) {
        stop(
          "Column step `", step_name_i,
          "` must have exactly one output name equal to the step name.",
          call. = FALSE
        )
      }

    } else if (identical(step$type, "block")) {
      if (!is.character(step$mode) || length(step$mode) != 1L || is.na(step$mode)) {
        stop(
          "Block step `", step_name_i, "` has malformed `mode`.",
          call. = FALSE
        )
      }

      if (identical(step$mode, "function")) {
        if (!is.function(step$fn)) {
          stop(
            "Block step `", step_name_i, "` has malformed `fn`.",
            call. = FALSE
          )
        }
      } else if (identical(step$mode, "expression")) {
        if (!rlang::is_quosure(step$quo)) {
          stop(
            "Block step `", step_name_i, "` has malformed `quo`.",
            call. = FALSE
          )
        }
      } else {
        stop(
          "Block step `", step_name_i, "` has unsupported `mode`: ",
          step$mode,
          call. = FALSE
        )
      }

    } else {
      stop(
        "Definition step `", step_name_i, "` has unsupported `type`: ",
        step$type,
        call. = FALSE
      )
    }
  }

  invisible(TRUE)
}


#' Evaluate one definition step
#'
#' Internal dispatcher used by `fd_compute()`.
#'
#' @param step Normalized definition step.
#' @param step_name Step name.
#' @param working_data Current working data frame.
#' @param mask Current evaluation mask.
#' @param n Expected row count.
#' @param key Key column name.
#' @param produced_names Names already produced in prior steps.
#' @param compute_envir Optional overriding environment.
#' @param compute_strict Logical strictness flag.
#'
#' @return A list with elements `ok`, `error`, and `cols`.
#'
#' @noRd
fd_compute_eval_step <- function(step,
                                 step_name,
                                 working_data,
                                 mask,
                                 n,
                                 key,
                                 produced_names,
                                 compute_envir = NULL,
                                 compute_strict = TRUE) {

  if (identical(step$type, "column")) {
    return(
      fd_compute_eval_column_step(
        step = step,
        step_name = step_name,
        mask = mask,
        n = n,
        compute_envir = compute_envir,
        compute_strict = compute_strict
      )
    )
  }

  if (identical(step$type, "block")) {
    return(
      fd_compute_eval_block_step(
        step = step,
        step_name = step_name,
        working_data = working_data,
        mask = mask,
        n = n,
        key = key,
        produced_names = produced_names,
        compute_envir = compute_envir,
        compute_strict = compute_strict
      )
    )
  }

  list(
    ok = FALSE,
    error = paste0("Unsupported step type: ", step$type),
    cols = list()
  )
}


#' Evaluate one column step
#'
#' @param step Column step.
#' @param step_name Step name.
#' @param mask Current evaluation mask.
#' @param n Expected row count.
#' @param compute_envir Optional overriding environment.
#' @param compute_strict Logical strictness flag.
#'
#' @return A list with elements `ok`, `error`, and `cols`.
#'
#' @noRd
fd_compute_eval_column_step <- function(step,
                                        step_name,
                                        mask,
                                        n,
                                        compute_envir = NULL,
                                        compute_strict = TRUE) {

  expr_i <- step$expr %||% rlang::get_expr(step$quo)
  env_i  <- compute_envir %||% step$env %||% rlang::get_env(step$quo)

  ok <- TRUE
  diag <- NULL
  res <- NULL

  res <- tryCatch(
    rlang::eval_tidy(expr_i, data = mask, env = env_i),
    error = function(e) {
      ok <<- FALSE
      diag <<- fd_compute_diagnose_error(e)
      NULL
    }
  )

  if (ok) {
    validation <- fd_compute_validate_column_result(res, n = n, name = step_name)
    ok <- validation$ok

    if (!ok) {
      diag <- list(
        class = "invalid_result",
        raw_message = validation$error,
        message = validation$error,
        hint = NA_character_
      )
    }

    res <- validation$result
  }

  if (!ok) {
    msg <- fd_compute_format_diagnostic_message(diag)
    cols <- stats::setNames(list(rep(NA, n)), step_name)

    if (!isTRUE(compute_strict)) {
      warning(sprintf("fd_compute(): step '%s' failed: %s", step_name, msg), call. = FALSE)
    }

    return(list(
      ok = FALSE,
      error = msg,
      error_raw = diag$raw_message %||% msg,
      error_class = diag$class %||% "other_error",
      hint = diag$hint %||% NA_character_,
      cols = cols
    ))
  }

  list(
    ok = TRUE,
    error = NA_character_,
    error_raw = NA_character_,
    error_class = NA_character_,
    hint = NA_character_,
    cols = stats::setNames(list(res), step_name)
  )
}


#' Evaluate one block step
#'
#' @param step Block step.
#' @param step_name Step name.
#' @param working_data Current working data frame.
#' @param mask Current evaluation mask.
#' @param n Expected row count.
#' @param key Key column name.
#' @param produced_names Names already produced in prior steps.
#' @param compute_envir Optional overriding environment.
#' @param compute_strict Logical strictness flag.
#'
#' @return A list with elements `ok`, `error`, and `cols`.
#'
#' @noRd
fd_compute_eval_block_step <- function(step,
                                       step_name,
                                       working_data,
                                       mask,
                                       n,
                                       key,
                                       produced_names,
                                       compute_envir = NULL,
                                       compute_strict = TRUE) {

  ok <- TRUE
  diag <- NULL
  res <- NULL

  if (identical(step$mode, "function")) {
    res <- tryCatch(
      fd_compute_call_block_function(step$fn, working_data),
      error = function(e) {
        ok <<- FALSE
        diag <<- fd_compute_diagnose_error(e)
        NULL
      }
    )

  } else if (identical(step$mode, "expression")) {
    expr_i <- step$expr %||% rlang::get_expr(step$quo)
    env_i  <- compute_envir %||% step$env %||% rlang::get_env(step$quo)

    res <- tryCatch(
      rlang::eval_tidy(expr_i, data = mask, env = env_i),
      error = function(e) {
        ok <<- FALSE
        diag <<- fd_compute_diagnose_error(e)
        NULL
      }
    )

  } else {
    ok <- FALSE
    diag <- list(
      class = "unsupported_mode",
      raw_message = paste0("Unsupported block mode: ", step$mode),
      message = paste0("Unsupported block mode: ", step$mode),
      hint = NA_character_
    )
  }

  if (ok) {
    validation <- fd_compute_validate_block_result(
      res = res,
      n = n,
      step_name = step_name,
      key = key,
      produced_names = produced_names,
      expected_names = step$output_names %||% NULL
    )
    ok <- validation$ok

    if (!ok) {
      diag <- list(
        class = "invalid_block_result",
        raw_message = validation$error,
        message = validation$error,
        hint = NA_character_
      )
    }

    res <- validation$result
  }

  if (!ok) {
    msg <- fd_compute_format_diagnostic_message(diag)
    fallback_cols <- list()

    if (length(step$output_names)) {
      fallback_cols <- stats::setNames(
        replicate(length(step$output_names), rep(NA, n), simplify = FALSE),
        step$output_names
      )
    }

    if (!isTRUE(compute_strict)) {
      warning(sprintf("fd_compute(): step '%s' failed: %s", step_name, msg), call. = FALSE)
    }

    return(list(
      ok = FALSE,
      error = msg,
      error_raw = diag$raw_message %||% msg,
      error_class = diag$class %||% "other_error",
      hint = diag$hint %||% NA_character_,
      cols = fallback_cols
    ))
  }

  cols <- as.list(res)
  names(cols) <- names(res)

  list(
    ok = TRUE,
    error = NA_character_,
    error_raw = NA_character_,
    error_class = NA_character_,
    hint = NA_character_,
    cols = cols
  )
}

#' Validate one computed column result
#'
#' @param x Computed result.
#' @param n Expected row count.
#' @param name Output name used in error messages.
#'
#' @return A list with elements `ok`, `error`, and `result`.
#'
#' @noRd
fd_compute_validate_column_result <- function(x, n, name) {
  if (is.null(x)) {
    return(list(
      ok = FALSE,
      error = paste0("Result for `", name, "` is NULL; expected a vector."),
      result = NULL
    ))
  }

  if (!fd_compute_is_supported_result(x)) {
    return(list(
      ok = FALSE,
      error = paste(
        "Result for `", name, "` has unsupported type/class:",
        paste(class(x), collapse = "/"),
        "- expected an atomic/Date/POSIXct/difftime vector."
      ),
      result = NULL
    ))
  }

  if (n == 0L) {
    if (length(x) %in% c(0L, 1L)) {
      return(list(ok = TRUE, error = NA_character_, result = x[0]))
    }

    return(list(
      ok = FALSE,
      error = sprintf(
        "Result for `%s` has length %s; expected 0 (or scalar recyclable to 0-row output).",
        name, length(x)
      ),
      result = NULL
    ))
  }

  if (length(x) == 1L) {
    return(list(ok = TRUE, error = NA_character_, result = rep(x, n)))
  }

  if (length(x) != n) {
    return(list(
      ok = FALSE,
      error = sprintf(
        "Result for `%s` has length %s; expected %s (or 1).",
        name, length(x), n
      ),
      result = NULL
    ))
  }

  list(ok = TRUE, error = NA_character_, result = x)
}


#' Validate one computed block result
#'
#' @param res Computed result.
#' @param n Expected row count.
#' @param step_name Step name used in error messages.
#' @param key Key column name.
#' @param produced_names Names already produced before this block.
#'
#' @return A list with elements `ok`, `error`, and `result`.
#'
#' @noRd
fd_compute_validate_block_result <- function(res,
                                             n,
                                             step_name,
                                             key,
                                             produced_names,
                                             expected_names = NULL) {

  if (is.null(res)) {
    return(list(
      ok = FALSE,
      error = paste0("Block step `", step_name, "` returned NULL; expected a data.frame."),
      result = NULL
    ))
  }

  if (!is.data.frame(res)) {
    return(list(
      ok = FALSE,
      error = paste0(
        "Block step `", step_name, "` returned unsupported type/class: ",
        paste(class(res), collapse = "/"),
        ". Expected a data.frame."
      ),
      result = NULL
    ))
  }

  if (nrow(res) != n) {
    return(list(
      ok = FALSE,
      error = sprintf(
        "Block step `%s` returned %s rows; expected %s.",
        step_name, nrow(res), n
      ),
      result = NULL
    ))
  }

  nm <- names(res)
  if (is.null(nm) || any(is.na(nm)) || any(!nzchar(nm))) {
    return(list(
      ok = FALSE,
      error = paste0("Block step `", step_name, "` returned unnamed or empty-named columns."),
      result = NULL
    ))
  }

  if (anyDuplicated(nm)) {
    dups <- unique(nm[duplicated(nm)])
    return(list(
      ok = FALSE,
      error = paste0(
        "Block step `", step_name, "` returned duplicate column names: ",
        paste(dups, collapse = ", ")
      ),
      result = NULL
    ))
  }

  if (key %in% nm) {
    return(list(
      ok = FALSE,
      error = paste0(
        "Block step `", step_name, "` returned column name(s) colliding with `key`: ",
        key
      ),
      result = NULL
    ))
  }

  colliding <- intersect(nm, produced_names)
  if (length(colliding)) {
    return(list(
      ok = FALSE,
      error = paste0(
        "Block step `", step_name, "` returned column name(s) already produced earlier: ",
        paste(colliding, collapse = ", ")
      ),
      result = NULL
    ))
  }

  if (!is.null(expected_names)) {
    unexpected <- setdiff(nm, expected_names)
    if (length(unexpected)) {
      return(list(
        ok = FALSE,
        error = paste0(
          "Block step `", step_name, "` returned unexpected column name(s): ",
          paste(unexpected, collapse = ", "),
          ". Allowed names are: ",
          paste(expected_names, collapse = ", ")
        ),
        result = NULL
      ))
    }
  }

  for (j in seq_along(res)) {
    col_j <- res[[j]]
    nm_j <- nm[[j]]

    validation <- fd_compute_validate_column_result(col_j, n = n, name = nm_j)
    if (!validation$ok) {
      return(list(ok = FALSE, error = validation$error, result = NULL))
    }

    res[[j]] <- validation$result
  }

  if (!is.null(expected_names)) {
    res <- fd_compute_complete_expected_block_outputs(
      res = res,
      expected_names = expected_names,
      n = n
    )

    # after completion, re-check collisions against already produced names
    nm2 <- names(res)

    if (key %in% nm2) {
      return(list(
        ok = FALSE,
        error = paste0(
          "Block step `", step_name, "` would produce column name(s) colliding with `key`: ",
          key
        ),
        result = NULL
      ))
    }

    colliding2 <- intersect(nm2, produced_names)
    if (length(colliding2)) {
      return(list(
        ok = FALSE,
        error = paste0(
          "Block step `", step_name, "` would produce column name(s) already produced earlier: ",
          paste(colliding2, collapse = ", ")
        ),
        result = NULL
      ))
    }
  }

  list(ok = TRUE, error = NA_character_, result = res)
}



#' Complete missing expected outputs from a block with `NA`
#'
#' Internal helper used by `fd_compute()` for block steps that declare
#' `expected_names`.
#'
#' When a block returns only a subset of its declared outputs, this helper adds
#' the missing expected columns as `NA` vectors and then reorders the result to
#' match the declared `expected_names`.
#'
#' This implements featdelta's lenient block-output completion policy:
#' \itemize{
#'   \item returned block columns must be a subset of `expected_names`
#'   \item any missing expected columns are added as `NA`
#'   \item final output order follows `expected_names`
#' }
#'
#' The helper does not validate whether unexpected columns are present; that
#' check belongs earlier in `fd_compute_validate_block_result()`.
#'
#' @param res A data frame returned by a block step.
#' @param expected_names Character vector of declared expected output names.
#' @param n Expected row count.
#'
#' @return A data frame containing exactly the columns in `expected_names`,
#'   ordered accordingly, with missing expected columns filled by `NA`.
#'
#' @family featdelta compute helpers
#' @noRd
fd_compute_complete_expected_block_outputs <- function(res, expected_names, n) {
  if (is.null(expected_names)) {
    return(res)
  }

  missing_names <- setdiff(expected_names, names(res))

  if (length(missing_names)) {
    for (nm in missing_names) {
      res[[nm]] <- rep(NA, n)
    }
  }

  res <- res[, expected_names, drop = FALSE]
  res
}


#' Call a block function against the current working data
#'
#' @param fn Function stored in a function-mode block.
#' @param working_data Current working data.
#'
#' @return Block result.
#'
#' @noRd
fd_compute_call_block_function <- function(fn, working_data) {
  fmls <- formals(fn)

  if (length(fmls) == 0L) {
    return(fn())
  }

  fn(working_data)
}


#' Diagnose an evaluation error from a compute step
#'
#' Internal helper used by `fd_compute()` to convert a raw evaluation error into
#' a structured diagnostic record that is more useful for users and for the
#' computation report.
#'
#' The function is intentionally lightweight. It does not attempt full static
#' dependency analysis. Instead, it classifies a few high-value failure patterns
#' and adds a concise hint when possible.
#'
#' In particular, "object not found" style errors are expanded with likely
#' causes relevant to featdelta's ordered step semantics:
#' \itemize{
#'   \item the raw input column is missing
#'   \item an earlier definition step was removed
#'   \item definition steps were reordered
#'   \item an earlier block no longer returns the expected column
#' }
#'
#' @param err An error object or error message.
#'
#' @return A list with fields:
#'   \itemize{
#'     \item `class`: diagnostic class label
#'     \item `raw_message`: original error message
#'     \item `message`: enriched user-facing message
#'     \item `hint`: short additional hint or `NA_character_`
#'   }
#'
#' @family featdelta compute helpers
#' @noRd
fd_compute_diagnose_error <- function(err) {
  raw_msg <- if (inherits(err, "error")) {
    conditionMessage(err)
  } else {
    as.character(err)[1]
  }

  if (!is.character(raw_msg) || length(raw_msg) != 1L || is.na(raw_msg)) {
    raw_msg <- "Unknown compute error."
  }

  if (grepl("object .* not found", raw_msg)) {
    hint <- paste(
      "Possible reasons:",
      "the raw input column is missing;",
      "an earlier definition step was removed;",
      "the order of definition steps was changed;",
      "or an earlier block no longer returns the expected column."
    )

    return(list(
      class = "missing_object",
      raw_message = raw_msg,
      message = paste(raw_msg, hint),
      hint = hint
    ))
  }

  if (grepl("could not find function", raw_msg, fixed = TRUE)) {
    hint <- paste(
      "Possible reasons:",
      "the function is not available in the stored evaluation environment;",
      "the compute environment was changed;",
      "or a required package/function was not attached or referenced explicitly."
    )

    return(list(
      class = "missing_function",
      raw_message = raw_msg,
      message = paste(raw_msg, hint),
      hint = hint
    ))
  }

  if (grepl("argument .* is missing", raw_msg)) {
    hint <- paste(
      "Possible reasons:",
      "a helper function inside the definition now requires different arguments;",
      "or a function-based block no longer matches the expected calling pattern."
    )

    return(list(
      class = "missing_argument",
      raw_message = raw_msg,
      message = paste(raw_msg, hint),
      hint = hint
    ))
  }

  list(
    class = "other_error",
    raw_message = raw_msg,
    message = raw_msg,
    hint = NA_character_
  )
}


#' Format a step failure for warnings and reports
#'
#' @param diag Diagnostic list from `fd_compute_diagnose_error()`.
#'
#' @return Character scalar.
#'
#' @family featdelta compute helpers
#' @noRd
fd_compute_format_diagnostic_message <- function(diag) {
  if (!is.list(diag) || is.null(diag$message)) {
    return("Unknown compute error.")
  }
  diag$message
}



#' Check whether a computed feature result is supported for column output
#'
#' Internal helper used by `fd_compute()` to reject unsupported evaluation
#' results before they propagate into the final feature frame.
#'
#' Supported result classes are vector-like column outputs suitable for
#' inclusion in a data frame. Tabular or list-like results should be rejected
#' here rather than failing later during DB write or report construction.
#'
#' @param x A single feature evaluation result.
#'
#' @return Logical scalar. `TRUE` if `x` is a supported result type for
#'   `fd_compute()`, otherwise `FALSE`.
#'
#' @family featdelta compute helpers
#' @noRd
fd_compute_is_supported_result <- function(x) {
  is.atomic(x) ||
    inherits(x, "Date") ||
    inherits(x, "POSIXct") ||
    inherits(x, "difftime")
}

---
title: "Database pipeline workshop"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Database pipeline workshop}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Purpose of this workshop

The previous vignettes introduced the two central ideas of `featdelta`:

1. feature logic can be written and organized in R;
2. the computed features can be stored in a database table and refreshed over
   time.

This workshop focuses on the database side of the package. It shows how
`featdelta` decides which rows need features, how it writes computed features
back to the database, and how the full pipeline is orchestrated by `fd_run()`.

The database pipeline is useful when raw observations arrive repeatedly and the
feature table should be kept up to date without rebuilding everything by hand.
The package is designed around three stages:

1. `fd_fetch()` finds raw rows that are missing from the feature table;
2. `fd_compute()` computes R feature definitions on those rows;
3. `fd_upsert()` inserts or updates the computed features in the database.

Most users will call `fd_run()`, which combines these stages. The individual
functions are still useful when you want to inspect, debug, or test one stage at
a time.

## Supported database dialects

`featdelta` uses DBI connections and selects SQL templates for the database
dialect. The supported dialects are:

* SQLite;
* PostgreSQL;
* MySQL or MariaDB.

In this vignette, the runnable examples use an in-memory SQLite database because
it does not require credentials or an external server. The same pipeline is
intended to work with PostgreSQL and MySQL through DBI-compatible drivers.

```{r, eval = FALSE}
# SQLite
con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")

# PostgreSQL
con <- DBI::dbConnect(
  RPostgres::Postgres(),
  host = "localhost",
  dbname = "analytics",
  user = "analyst",
  password = "secret"
)

# MySQL or MariaDB
con <- DBI::dbConnect(
  RMariaDB::MariaDB(),
  host = "localhost",
  dbname = "analytics",
  user = "analyst",
  password = "secret"
)
```

Usually the dialect is detected from the connection. If detection is not
possible, you can pass `dialect = "sqlite"`, `dialect = "postgres"`, or
`dialect = "mysql"` to `fd_run()` or `fd_upsert()`.

## Workshop database

We will use a small e-commerce order dataset. The data is intentionally simple,
but it gives us a realistic database workflow:

1. raw orders arrive in a table;
2. R definitions calculate transformed features;
3. features are written to a separate feature table;
4. new raw orders are processed incrementally.

```{r}
library(DBI)
library(RSQLite)
library(featdelta)

con <- dbConnect(SQLite(), ":memory:")

orders <- data.frame(
  order_id = 1:7,
  customer_id = c(101, 102, 103, 101, 104, 105, 102),
  gross_amount = c(120, 250, 80, 310, 45, 520, 160),
  discount_amount = c(0, 25, 5, 30, 0, 60, 10),
  shipping_fee = c(8, 0, 6, 0, 5, 0, 7),
  order_to_ship_days = c(1, 3, 2, 5, 1, 4, 2),
  stringsAsFactors = FALSE
)

day_one <- 1:4
day_two <- 5:7

dbWriteTable(
  con,
  "raw_orders",
  orders[day_one, ],
  overwrite = TRUE
)

dbGetQuery(con, "SELECT * FROM raw_orders ORDER BY order_id")
```

The raw table contains day-one observations only. The feature table does not
exist yet.

## Define the source query

`featdelta` does not require the source data to come from one table. The source
dataset is defined by a SQL query. In real projects, this query may join several
tables, filter to a modelling population, or select only specific records.

```{r}
source_sql <- "
  SELECT
    order_id,
    customer_id,
    gross_amount,
    discount_amount,
    shipping_fee,
    order_to_ship_days
  FROM raw_orders
  ORDER BY order_id
"

key <- "order_id"

dbGetQuery(con, source_sql)
```

There is one important requirement: the query must return the key column. The
key connects the raw dataset to the feature table. It is how `featdelta` knows
whether a raw row has already been processed.

## Define features in R

The feature logic is still ordinary R. Here we define a few transformed
variables that are more useful for analytics or reporting than the raw columns.

```{r}
defs <- fd_define(
  net_revenue = gross_amount - discount_amount + shipping_fee,
  discount_rate = discount_amount / gross_amount,
  free_shipping = shipping_fee == 0,
  slow_fulfillment = order_to_ship_days > 3
)

defs
```

The database pipeline does not require you to write these transformations in
SQL. You can keep them in R, test them with `fd_compute()`, and then use the
same definitions in the database pipeline.

## First run: create and populate the feature table

On the first run, the feature table is missing. `fd_run()` executes the source
query, computes the features, creates the feature table, and inserts the result.

```{r}
run_day_one <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "order_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")
```

The feature table is now a database table. Other R scripts, dashboards, model
training jobs, or monitoring jobs can read it without rerunning the feature
engineering code.

## Second run: fetch only rows that are missing features

Now add more raw orders. This simulates new observations arriving after
the first feature run.

```{r}
dbAppendTable(con, "raw_orders", orders[day_two, ])

dbGetQuery(con, "SELECT * FROM raw_orders ORDER BY order_id")
```

The raw table contains seven orders, but the feature table still contains
features for only the first four.

```{r}
dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")
```

When we run the pipeline again with the default `fetch_mode = "new_only"`,
`featdelta` finds the raw keys missing from the feature table and computes only
those rows.

```{r}
run_day_two <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "order_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")
```

The existing feature rows were not duplicated. Only the missing keys were
inserted.

## Inspect fd_fetch() directly

Most of the time, `fd_run()` is enough. During development, it can be useful to
look at the fetch stage directly. `fd_fetch()` returns raw rows whose key is not
present in the feature table.

Let's demonstrate that with an example: first, add one more raw order.

```{r}
new_order <- data.frame(
  order_id = 8,
  customer_id = 106,
  gross_amount = 275,
  discount_amount = 20,
  shipping_fee = 0,
  order_to_ship_days = 6
)

dbAppendTable(con, "raw_orders", new_order)

new_rows <- fd_fetch(
  con = con,
  sql = source_sql,
  key = key,
  feat_table_name = "order_features"
)

new_rows
```

This is the row that would be computed on the next incremental run. The fetch
result also carries metadata, including the SQL that was executed.

```{r}
attr(new_rows, "fd_fetch")
```

This can help when you want to verify that the package is selecting the rows you
expect.

## Inspect fd_upsert() directly

`fd_upsert()` writes a feature data frame into the database. It is useful when
you already have computed features and want to use only the persistence part of
the package.

```{r}
new_features <- fd_compute(
  data = new_rows,
  defs = defs,
  key = key
)

upsert_report <- fd_upsert(
  con = con,
  features_df = new_features,
  feat_table_name = "order_features",
  key = key,
  verbose = FALSE
)

upsert_report

dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")
```

The upsert report summarizes what the write attempted to do. The `would_insert`
and `would_update` counts are existence-based counts computed before the merge.
They tell you how many incoming keys were new and how many already existed.

## Add a new feature column

Feature tables often evolve. Suppose we add a new feature, `high_value_order`.
If we run the pipeline with `fetch_mode = "new_only"`, only new rows would be
processed. Existing rows would not receive the new column values.

When you want to backfill or refresh existing rows, use `fetch_mode = "all"`.

```{r}
defs_v2 <- fd_define(
  net_revenue = gross_amount - discount_amount + shipping_fee,
  discount_rate = discount_amount / gross_amount,
  free_shipping = shipping_fee == 0,
  slow_fulfillment = order_to_ship_days > 3,
  high_value_order = gross_amount >= 250
)

refresh_report <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "order_features",
  fetch_mode = "all",
  verbose = FALSE
)

dbListFields(con, "order_features")

dbGetQuery(con, "SELECT * FROM order_features ORDER BY order_id")
```

With the default `alter_table = TRUE`, `fd_upsert()` adds missing feature
columns to the target table. This schema evolution is intentionally conservative:
columns can be added, but columns are not dropped, renamed, or type-changed
automatically.

## Understand create_table and alter_table

The default `create_table = "auto"` creates the feature table when it does not
exist and uses the existing table when it does. This is convenient for ordinary
incremental pipelines.

If you want to be stricter, use:

```{r, eval = FALSE}
# Error if the table does not already exist
fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "order_features",
  create_table = FALSE
)

# Error if new feature columns are missing from the existing table
fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "order_features",
  alter_table = FALSE
)
```

This is useful in production environments where table creation or schema changes
must be handled through a separate approval process.

## Choose between new_only and all

The most important database-pipeline decision is the fetch mode.

Use `fetch_mode = "new_only"` for ordinary incremental processing. This mode
uses the feature table to identify keys that have not yet been processed. It is
the default.

Use `fetch_mode = "all"` when existing feature rows should be recomputed. Common
reasons include:

1. you changed a feature definition;
2. you added a new feature and want old rows to receive values;
3. you are performing an explicit backfill;
4. you want to rebuild a feature table from the current source query.

`fetch_mode = "new_only"` is key-based. It does not know whether your feature
logic changed. If the definitions changed and existing rows should be updated,
use `fetch_mode = "all"`.

## Insert-only mode

By default, `fd_upsert()` updates existing keys and inserts new keys. This is
usually what you want for refreshes and backfills.

If you want a stricter insert-only workflow, set `update_table = FALSE`. In that
mode, any incoming key that already exists in the feature table is treated as a
conflict.

```{r, eval = FALSE}
fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "order_features",
  fetch_mode = "all",
  update_table = FALSE
)
```

This can be useful when your pipeline is intended only to append new feature
rows and you want accidental refreshes to fail loudly.

## Process large writes in chunks

For larger feature frames, use `chunk_size` to write the data in batches. This
can reduce memory and packet-size pressure on the database connection.

```{r, eval = FALSE}
fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "order_features",
  chunk_size = 5000
)
```

Chunking does not change the feature logic. It only changes how computed rows
are staged and merged into the target table.

## Dialect notes for PostgreSQL and MySQL

SQLite is convenient for local examples. In production, many users will connect
to PostgreSQL, MySQL, or MariaDB.

For PostgreSQL and MySQL, you can use schema-qualified feature table names when
the driver supports them through DBI identifiers:

```{r, eval = FALSE}
fd_run(
  con = con,
  sql = "SELECT * FROM raw_schema.orders",
  defs = defs,
  key = "order_id",
  feat_table_name = "feature_schema.order_features",
  dialect = "postgres"
)

fd_run(
  con = con,
  sql = "SELECT * FROM raw_schema.orders",
  defs = defs,
  key = "order_id",
  feat_table_name = "feature_schema.order_features",
  dialect = "mysql"
)
```

Existing target tables used with `update_table = TRUE` must have a primary key
or unique constraint on the key column. Tables created by `fd_upsert()` include
that primary key automatically.

If your DBI connection is correctly identified, you usually do not need the
`dialect` argument. It is available as an explicit override when automatic
detection is not sufficient.

## Read the run report

The database table is the main result of the pipeline, but the run report is
useful for checking what happened.

```{r}
names(refresh_report)

refresh_report$fetch

refresh_report$compute$feature_names

refresh_report$upsert$counts

refresh_report$upsert$columns_added
```

The report is especially useful in development and monitoring. It shows which
stage ran, how many rows were fetched and computed, and what the upsert stage
planned to insert or update.

## What to remember

The database pipeline exists to keep a feature table current while letting the
feature logic stay in R.

Use `fd_run()` for the regular end-to-end workflow. Use `fd_fetch()` when you
want to inspect which raw rows still need features. Use `fd_upsert()` when you
already have a feature data frame and want to persist it safely.

For daily or repeated processing, start with the default
`fetch_mode = "new_only"`. When feature definitions change and old rows need to
be refreshed, switch to `fetch_mode = "all"`. For production databases, make
sure the feature table has a stable key and choose the dialect that matches the
DBI connection: SQLite, PostgreSQL, or MySQL/MariaDB.

```{r cleanup, include = FALSE}
dbDisconnect(con)
```

---
title: "Getting started with featdelta"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with featdelta}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Why featdelta exists

`featdelta` is for R users who create features in R but need those features to
live in a database.

In many analytics projects, the raw observations are stored in a database and
new rows arrive over time. The feature engineering logic, however, may be easier
to write, test, and maintain in R. This creates a practical problem: after the R
code computes the features, those values still need to be stored in a database
table so they can be reused by modelling scripts, dashboards, monitoring jobs,
or other downstream systems.

`featdelta` provides an automated pipeline for that workflow. You define feature
expressions in R, and the package handles the database-oriented steps around
them:

1. storing feature definitions in a reusable object;
2. finding raw rows that do not yet have features;
3. computing features for those missing rows;
4. creating or extending the database feature table;
5. inserting or updating the computed feature values.

The goal is to reduce the amount of repeated code needed to refresh feature
tables. Instead of rebuilding an analysis dataset by hand each time new raw data
arrives, you can keep a persistent feature table in the database and update it
incrementally.

## Where this helps

This pattern appears in many applied data workflows.

For example, in a daily market-data workflow, a database table may be updated
with the latest closing prices after each trading day. Once the new prices are
available, an analyst may calculate indicators in R and store them in a separate
feature table. On the next day, the pipeline should process only the newly
available observations rather than rebuilding the entire history.

In a credit-risk workflow, new loan or credit-card applications may arrive every
day. Feature engineering often involves transformed variables that are more
useful for modelling than the raw input columns. For example, an analyst may
derive ratios, delay summaries, utilization measures, or other scorecard inputs
from the raw application and bureau data. If these features are recreated only
inside ad-hoc modelling scripts, refreshing the dataset with recent applications
can become a recurring manual task. A persistent feature table makes the
transformed variables easier to reuse for model development, validation, and
monitoring.

In a business-intelligence workflow, a reporting tool may need variables that
are difficult to create directly in the BI layer. These might come from an R
model, a specialized matching algorithm, or domain-specific transformation code.
If the results are pushed back to the database, the BI tool can read them like
ordinary columns.

The details differ across domains, but the workflow is similar: raw data changes
over time, feature logic is maintained in R, and the computed features are more
useful when they are available in the database.

## A small running example

This vignette uses an in-memory SQLite database and the built-in `mtcars`
dataset. The database is tiny and the example features are intentionally simple,
but the workflow mirrors a real production pattern:

1. finding which raw observations are new;
2. recomputing only the features that are missing;
3. keeping feature definitions organized;
4. creating or extending the database table where the features live;
5. pushing the newly computed values back to the database safely.

Before demonstrating `featdelta`, we first set up the database. In this example,
we pretend that the rows arrive in two batches:

* at the end of day one, we have observed the first 20 rows of the `mtcars`
  dataset;
* at the end of day two, we observe the next 10 rows, corresponding to ids 21 to
  30. These rows are not available on day one.

Our goal is to add new features to a database feature table at the end of each
day:

* whether the transmission is automatic or manual, stored as a text field;
* the horsepower-to-cylinder ratio;
* the weight per horsepower unit.

```{r}
library(DBI)
library(RSQLite)
library(featdelta)

cars <- mtcars
cars$id <- seq_len(nrow(cars))

# Name of the unique row identifier used throughout the pipeline.
key <- "id"

day_one <- 1:20
day_two <- 21:30
```

We use an in-memory SQLite database to mimic a real R-to-database workflow. We
establish the database connection with `DBI::dbConnect()`. For this small
example, no database credentials are needed.

Next, `DBI::dbWriteTable()` pushes the first batch of rows from R into the
database and creates a table named `raw_cars`. This table represents the raw
observations for which we want to create separate features.

After the raw data has been written to the database, we can inspect it using a
simple SQL query.

```{r}
con <- dbConnect(SQLite(), ":memory:")
dbWriteTable(con, "raw_cars", cars[day_one, ], overwrite = TRUE)

dbGetQuery(con, "SELECT id, mpg, cyl, hp, wt, am FROM raw_cars ORDER BY id")
```

In real-world scenarios, the data needed for feature engineering is rarely stored
in one simple table. It is often selected from several database tables with
joins, filters, and other selection criteria. For that reason, the
database-to-R interface in this vignette uses a SQL query instead of referring
directly to the `raw_cars` table.

In the snippet below, we define the source dataset as rows where `id > 15`.
This demonstrates that the dataset to be processed can be created with a custom
query. There is one important requirement: each returned row must have a unique
identifier. That is why we defined the `key` variable earlier.

We can bring this source dataset into the R session with
`DBI::dbGetQuery(con, source_sql)`.

```{r}
source_sql <- "SELECT * FROM raw_cars WHERE id > 15 ORDER BY id"

dbGetQuery(con, source_sql)
```

At this point, the source table contains raw observations, but the feature table
does not exist yet.

The rest of the vignette demonstrates how the feature table can be created and
refreshed with `featdelta`. Later, we summarize which manual steps this pipeline
replaces.

## Feature definitions live in R

Start by defining features. These are ordinary R expressions evaluated against
the raw data returned by `source_sql`.

```{r}
defs <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp
)

defs
```

This is one of the first conveniences of `featdelta`: feature definitions are
stored as an object. They can be printed, reviewed, reused, tested, and passed
to the pipeline. The expressions are not scattered across one-off scripts.

Before writing to the database, we can also compute locally to see the shape of
the feature table that will be produced.

```{r}
raw_preview <- dbGetQuery(con, source_sql)

fd_compute(
  data = raw_preview,
  defs = defs,
  key = key
)
```

The output contains the key column plus the computed features. This is the
table shape we want to persist.

In this example, we define features in the simplest possible way: one expression
per feature. In real projects, feature definitions may be generated by longer R
scripts and may produce many columns. Later vignettes will cover richer
definition patterns, including multi-column feature blocks.

## First run: create the feature table

Now let `featdelta` run the full pipeline. The important result is the database
table it creates. In this example, we call the new database table
`car_features`.

```{r}
run_day_one <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "car_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM car_features ORDER BY id")
```

The R expressions were evaluated against the selected raw rows, and the result
was stored as a persistent table. You can inspect that table with an ordinary SQL
query, just as you would inspect any other database table.

That is the central idea: the data scientist can write feature logic in R, while
the computed features are stored in a database table that other workflows can
reuse.

## Second run: process only new rows

Now pretend time passes and more raw observations arrive.

```{r}
dbAppendTable(con, "raw_cars", cars[day_two, ])

dbGetQuery(
  con,
  "SELECT id, mpg, cyl, hp, wt, am FROM raw_cars WHERE id > 15 ORDER BY id DESC"
)
```

The query above shows that new rows have been added to the raw table. The
feature table, however, has not been refreshed yet. If we inspect it now, it
still ends at id 20, even though raw rows up to id 30 are available.

```{r}
dbGetQuery(
  con,
  "SELECT * FROM car_features ORDER BY id DESC"
)
```

This is the point where manual pipelines become repetitive: you need to compare
raw keys with feature-table keys, build feature rows only for the missing raw
observations, and append them without duplicating existing rows.

With `featdelta`, run the same pipeline again and inspect the database table.

```{r}
run_day_two <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "car_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM car_features ORDER BY id DESC")
```

The existing feature rows were not duplicated. `featdelta` detected the raw rows
whose keys were missing from `car_features`, computed features only for those
rows, and inserted them into the database table.

## Add a new feature later

Feature engineering is rarely finished on the first day. Suppose we decide that
we also want a power-to-weight feature.

We can define a new version of the feature set and run the same pipeline again.
This time, the definitions include the new `power_to_weight` feature.

```{r}
defs_v2 <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp,
  power_to_weight = hp / wt
)

run_refresh <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "car_features",
  fetch_mode = "all",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM car_features ORDER BY id DESC")
```

Here we use `fetch_mode = "all"`, which tells `fd_run()` to refresh the features
for all rows returned by `source_sql`, including rows that already existed in
the feature table. This ensures that old rows are not missing the newly added
feature. In a real workflow, this is the difference between maintaining a
database feature table over time and repeatedly rebuilding analysis datasets by
hand.

## What the pipeline replaces

Without `featdelta`, the workflow usually grows into a script that does all of
this manually:

1. run a source SQL query;
2. check which keys already exist in the feature table;
3. keep only the missing rows;
4. evaluate feature expressions in R;
5. create the target table if it does not exist;
6. alter the target table when new feature columns appear;
7. insert or update rows;
8. inspect whether the run did what you expected.

That script can be written, but it tends to become repetitive and fragile.
`featdelta` makes those steps explicit:

```{r, eval = FALSE}
defs <- fd_define(
  feature_a = some_r_expression,
  feature_b = another_r_expression
)

fd_run(
  con = con,
  sql = "SELECT * FROM raw_table",
  defs = defs,
  key = "id",
  feat_table_name = "feature_table"
)
```

For most users, this is the main workflow: define features in R, then run the
database pipeline.

## Inspect the run when needed

The database table is the main result, but `fd_run()` also returns a structured
report. The report is useful when developing, debugging, or monitoring the
pipeline.

```{r}
run_report <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "car_features",
  fetch_mode = "all",
  return_data = "features",
  preview_n = 3,
  verbose = FALSE
)

run_report
```

The report tells you how many rows were fetched, which features were computed,
and how many rows were targeted for insert or update. It is there when you need
visibility, but it does not get in the way of the core workflow.

## Where to go next

This vignette introduced the main story: create feature logic in R, keep a
database feature table up to date, and avoid rebuilding the same feature dataset
by hand.

The next topics are:

1. writing richer feature definitions with `fd_define()` and `fd_block()`;
2. using `fd_fetch()` and `fd_upsert()` directly;
3. production patterns for failures, previews, refreshes, and reports.

```{r cleanup, include = FALSE}
dbDisconnect(con)
```

---
title: "Feature definitions workshop"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Feature definitions workshop}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Purpose of this workshop

The first vignette showed how `featdelta` helps move features created in R into
a database table. This workshop focuses on the part of the workflow that most
data scientists will edit most often: the feature definitions.

Feature definitions are the bridge between exploratory R code and a repeatable
feature pipeline. When features are kept only as lines inside an ad-hoc script,
they are harder to review, reuse, test, and refresh. With `fd_define()`, the
feature logic becomes a structured object. That object can then be computed
locally with `fd_compute()` or passed to `fd_run()` for the full database
pipeline.

In this workshop, we will build feature definitions step by step:

1. simple one-column features;
2. features that depend on earlier features;
3. reusable definitions supplied programmatically;
4. multi-column scripts with `fd_block()`;
5. function-based blocks that generate an unknown number of columns.


## Workshop data

We will use a small version of the built-in `mtcars` dataset. The data is simple
enough to inspect directly, but it lets us demonstrate the same patterns used in
larger feature engineering projects.

```{r}
library(featdelta)

raw_cars <- mtcars
raw_cars$car_id <- seq_len(nrow(raw_cars))

raw_cars <- raw_cars[, c("car_id", "mpg", "cyl", "disp", "hp", "wt", "am")]

head(raw_cars)
```

The `car_id` column is the key. It identifies each row and will be preserved in
the computed feature table.

## Start with simple feature definitions

The most direct use of `fd_define()` is to write one expression per feature.
Each named expression becomes one output column.

```{r}
defs_basic <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp
)

defs_basic
```

The printed object gives a quick overview of the feature set. 
Note that the feature names and expressions are stored together and can be passed
around as one object.

Now compute the definitions on the raw data.

```{r}
features_basic <- fd_compute(
  data = raw_cars,
  defs = defs_basic,
  key = "car_id"
)

head(features_basic)
```

The output contains the key column plus the computed features. This is the
feature table shape that can later be written to the database by `fd_run()`.

## Use earlier features in later features

Definitions are evaluated in order. This means a later feature can use columns
created by earlier definitions in the same `fd_compute()` call.

```{r}
defs_ordered <- fd_define(
  hp_per_cyl = hp / cyl,
  strong_engine = hp_per_cyl > 30,
  engine_label = ifelse(strong_engine, "strong", "regular")
)

features_ordered <- fd_compute(
  data = raw_cars,
  defs = defs_ordered,
  key = "car_id"
)

head(features_ordered)
```

This is useful when your feature engineering naturally has stages. You can first
create a base transformation, then reuse it in flags, labels, scores, or other
derived features. The important habit is to keep the order intentional. If a
later feature uses `hp_per_cyl`, then `hp_per_cyl` must be defined earlier.

## Keep programmatic definitions reusable

Sometimes feature definitions are created outside the `fd_define()` call. For
example, you might keep a small library of expressions, generate definitions
from a configuration file, or reuse the same expression across projects.

```{r}
log_hp_expr <- expression(log(hp))
heavy_car_expr <- expression(wt > 3.5)

defs_programmatic <- fd_define(
  log_hp = log_hp_expr,
  heavy_car = heavy_car_expr
)

features_programmatic <- fd_compute(
  data = raw_cars,
  defs = defs_programmatic,
  key = "car_id"
)

head(features_programmatic)
```

The benefit is administrative: the feature set can be built from named pieces
instead of being rewritten by hand each time. This matters when the feature
catalog becomes larger than a few simple columns.

## Use fd_block() when one feature step returns several columns

One expression per feature is convenient for small feature sets. In real
projects, however, a single conceptual feature step may naturally produce
several columns. That is what `fd_block()` is for.

An `fd_block()` is a multi-column definition step. It must return a
`data.frame`, and each column of that data frame becomes a feature.

```{r}
defs_block <- fd_define(
  engine_ratios = fd_block({
    data.frame(
      hp_per_cyl = hp / cyl,
      disp_per_cyl = disp / cyl,
      wt_per_hp = wt / hp
    )
  })
)

features_block <- fd_compute(
  data = raw_cars,
  defs = defs_block,
  key = "car_id"
)

head(features_block)
```

This is useful when you want one named definition step, such as
`engine_ratios`, to produce a small family of related columns.

## Write a small script inside fd_block()

The block body does not have to be a single `data.frame()` call. It can be a
small R script. You can create temporary variables, reuse intermediate
calculations, and return only the final columns you want to store.

```{r}
defs_script_block <- fd_define(
  engine_script = fd_block({
    hp_per_cyl <- hp / cyl
    disp_per_cyl <- disp / cyl

    ratio_average <- (hp_per_cyl + disp_per_cyl) / 2
    high_ratio <- ratio_average > stats::median(ratio_average, na.rm = TRUE)

    data.frame(
      hp_per_cyl = hp_per_cyl,
      disp_per_cyl = disp_per_cyl,
      engine_ratio_average = ratio_average,
      high_engine_ratio = high_ratio
    )
  })
)

features_script_block <- fd_compute(
  data = raw_cars,
  defs = defs_script_block,
  key = "car_id"
)

head(features_script_block)
```

This pattern is often easier to read than forcing every intermediate expression
into a separate top-level feature. Temporary variables stay inside the block,
while the returned data frame defines the columns that become part of the final
feature table.

## Use function-based blocks for larger feature scripts

As feature logic grows, it is often better to move it into a regular R function.
This is especially helpful when you want to test the feature script separately,
reuse it across projects, or keep the `fd_define()` call compact.

```{r}
make_engine_features <- function(data) {
  hp_per_cyl <- data$hp / data$cyl
  disp_per_cyl <- data$disp / data$cyl

  data.frame(
    hp_per_cyl = hp_per_cyl,
    disp_per_cyl = disp_per_cyl,
    engine_index = hp_per_cyl + disp_per_cyl
  )
}

defs_function_block <- fd_define(
  engine_features = fd_block(make_engine_features)
)

features_function_block <- fd_compute(
  data = raw_cars,
  defs = defs_function_block,
  key = "car_id"
)

head(features_function_block)
```

Function-based blocks are a good fit for code that already looks like a small
feature-engineering script. The function receives the current working data and
returns a `data.frame` of feature columns.

## Generate an unknown number of features in a loop

Some feature sets are not known column by column in advance. For example, you
might want to apply the same transformation to a selected group of numeric
variables. A function-based block can generate those columns in a loop.

```{r}
make_scaled_features <- function(data) {
  vars <- c("hp", "disp", "wt")
  out <- list()

  for (var in vars) {
    center <- mean(data[[var]], na.rm = TRUE)
    spread <- stats::sd(data[[var]], na.rm = TRUE)

    out[[paste0(var, "_scaled")]] <- (data[[var]] - center) / spread
  }

  as.data.frame(out)
}

defs_loop_block <- fd_define(
  scaled_inputs = fd_block(make_scaled_features)
)

features_loop_block <- fd_compute(
  data = raw_cars,
  defs = defs_loop_block,
  key = "car_id"
)

head(features_loop_block)
```

This pattern is useful when the number of output columns depends on a vector of
input names, a configuration object, or another piece of project logic. The
important rule remains the same: the block must return a data frame with one row
per input row.

## Combine ordinary features and blocks

You do not have to choose between ordinary definitions and blocks. A single
definition object can contain both. Later steps can also use columns produced by
earlier steps, including columns produced by blocks.

```{r}
defs_combined <- fd_define(
  transmission = ifelse(am == 1, "automatic", "manual"),
  engine_features = fd_block(make_engine_features),
  scaled_inputs = fd_block(make_scaled_features),
  engine_per_weight = engine_index / wt
)

features_combined <- fd_compute(
  data = raw_cars,
  defs = defs_combined,
  key = "car_id"
)

head(features_combined)
```

This is where feature definitions become a practical organizing tool. You can
keep simple expressions simple, move related feature families into blocks, and
still evaluate the full set as one ordered pipeline.

## Declare expected block outputs when useful

Sometimes you want a block to have an expected output schema. This is useful
when the block may return only some columns in some situations, but the database
feature table should still have a stable set of columns.

```{r}
defs_expected <- fd_define(
  optional_engine_flags = fd_block(
    {
      data.frame(
        high_hp = hp > 150
      )
    },
    expected_names = c("high_hp", "high_disp")
  )
)

features_expected <- fd_compute(
  data = raw_cars,
  defs = defs_expected,
  key = "car_id"
)

head(features_expected)
```

The block returned `high_hp`, but `high_disp` was declared as an expected output.
`fd_compute()` includes the missing expected column and fills it with `NA`. This
can help when you want the feature table to keep a predictable schema.

## What to remember

Feature definitions are where the package lets you turn R feature engineering
into a reusable pipeline component.

Use ordinary `fd_define()` expressions when each feature is simple and readable
on one line. Use `fd_block()` when a feature step naturally produces several
columns, needs temporary variables, or belongs in a reusable function. Use
function-based blocks when the logic is long enough to test separately or when
the output columns are generated programmatically.

Once the definitions object is ready, the same object can be used in two ways:

```{r, eval = FALSE}
# Local computation while developing feature logic
fd_compute(raw_data, defs, key = "id")

# Full database pipeline once the definitions are ready
fd_run(con, sql, defs, key = "id", feat_table_name = "feature_table")
```

That is the main workflow: develop feature logic in R, store it as a definitions
object, test it locally, and then use it in the incremental database pipeline.

---
title: "Production patterns"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Production patterns}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Purpose of this workshop

The earlier vignettes introduced the main workflow:

1. define feature logic in R;
2. compute features from raw rows;
3. store those features in a database table;
4. rerun the pipeline as new raw rows arrive.

This vignette focuses on patterns that become important once the workflow is
used repeatedly. In practice, a production feature pipeline needs more than a
single successful run. You need a way to preview changes, inspect failures,
refresh old rows when definitions change, control schema evolution, and monitor
what the pipeline did.

`featdelta` is not a scheduler, an orchestration platform, or a replacement for
database permissions. It is the R feature-engineering component inside that
larger workflow. This workshop shows how to use the package in a way that is
more comfortable to maintain over time.

## Workshop database

We will use a small device-monitoring example. Imagine a database table that
receives equipment readings over time. Analysts want to derive features in R and
store them in a separate table for dashboards, alerts, or predictive-maintenance
models.

```{r}
library(DBI)
library(RSQLite)
library(featdelta)

con <- dbConnect(SQLite(), ":memory:")

readings <- data.frame(
  reading_id = 1:10,
  device_id = c("A", "A", "B", "C", "A", "B", "C", "A", "B", "C"),
  temperature = c(68, 71, 75, 70, 83, 78, 72, 88, 81, 74),
  vibration = c(0.12, 0.15, 0.31, 0.18, 0.44, 0.39, 0.22, 0.55, 0.41, 0.24),
  pressure = c(31, 33, 36, 32, 39, 37, 34, 42, 38, 35),
  runtime_hours = c(120, 125, 210, 88, 130, 216, 93, 136, 223, 97),
  stringsAsFactors = FALSE
)

day_one <- 1:6
day_two <- 7:10

dbWriteTable(
  con,
  "raw_readings",
  readings[day_one, ],
  overwrite = TRUE
)

source_sql <- "
  SELECT
    reading_id,
    device_id,
    temperature,
    vibration,
    pressure,
    runtime_hours
  FROM raw_readings
  ORDER BY reading_id
"

key <- "reading_id"

dbGetQuery(con, source_sql)
```

This is a small table, but the operating pattern is realistic: raw readings
arrive, R creates transformed features, and the database stores the results.

## Start by testing definitions locally

Before writing to a production database table, test the feature definitions on a
small local extract. This lets you check names, row counts, and basic values
without changing the database feature table.

```{r}
defs <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  pressure_per_hour = pressure / runtime_hours,
  maintenance_flag = temp_above_80 | vibration_score > 80
)

raw_preview <- dbGetQuery(con, source_sql)

features_preview <- fd_compute(
  data = raw_preview,
  defs = defs,
  key = key
)

features_preview
```

This is the first production habit: separate definition development from
database writes. When the output table has the expected shape, use the same
definitions in `fd_run()`.

```{r}
run_initial <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "reading_features",
  verbose = FALSE
)

dbGetQuery(con, "SELECT * FROM reading_features ORDER BY reading_id")
```

## Use a development feature table for dry runs

`fetch_limit`, `preview_n`, and `return_data` are useful during development, but
remember that `fd_run()` is still a database-writing function. If you want to
test a pipeline run without touching the production feature table, use a
development database or a separate feature table.

```{r}
dev_report <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "reading_features_dev",
  fetch_limit = 3,
  return_data = "both",
  preview_n = 2,
  verbose = FALSE
)

dev_report$preview$raw
dev_report$preview$features
dev_report$data$features
```

This pattern is useful when changing definitions. You can inspect a small run
against a scratch table, then run the same definitions against the real feature
table once the output looks right.

## Let the report answer operational questions

Every `fd_run()` returns a structured report. In production, this object is more
useful than the printed table because it can be logged, inspected, or checked by
monitoring code.

```{r}
run_initial$success
run_initial$stage
run_initial$fetch$n_rows
run_initial$compute$feature_names
run_initial$upsert$counts
```

A typical monitoring check can be simple: did the run succeed, which stage
finished, how many rows were fetched, and how many rows were inserted or
updated?

## Handle compute failures without losing context

During development, a broken feature definition should be easy to diagnose. In a
production job, you may also want a failed run to return a report instead of
immediately stopping the whole script.

Set `fail_fast = FALSE` to capture stage failures in the returned
`fd_run_report`.

In the example below we define a feature, named `broken_feature`, that won't compute
as the definition requires a column that does not exist.

```{r}
bad_defs <- fd_define(
  temp_above_80 = temperature > 80,
  broken_feature = missing_sensor_column / 10
)

failed_report <- fd_run(
  con = con,
  sql = source_sql,
  defs = bad_defs,
  key = key,
  feat_table_name = "reading_features_broken",
  fail_fast = FALSE,
  return_data = "both",
  verbose = FALSE
)

failed_report$success
failed_report$stage
failed_report$error
failed_report$compute$report
```

The returned report keeps the failure attached to the compute stage. When
possible, it also keeps partial data and the compute report, which makes it
easier to see which step failed.

In contrast, `fail_fast = TRUE` is useful when the surrounding job scheduler
should treat any stage failure as an immediate error.

```{r, eval = FALSE}
fd_run(
  con = con,
  sql = source_sql,
  defs = bad_defs,
  key = key,
  feat_table_name = "reading_features",
  fail_fast = TRUE
)
```

## Choose the right refresh mode

For regular daily processing, use the default `fetch_mode = "new_only"`. This
mode processes only raw keys that are not yet present in the feature table.

```{r}
dbAppendTable(con, "raw_readings", readings[day_two, ])

run_incremental <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "reading_features",
  verbose = FALSE
)

# The default fetch mode is "new_only".
run_incremental$fetch$mode

# Only the four new day-two readings were fetched.
run_incremental$fetch$n_rows

# Four rows were inserted, and no existing feature rows were updated.
run_incremental$upsert$counts
```

Use `fetch_mode = "all"` when feature definitions changed and existing rows
should be recomputed.

```{r}
defs_v2 <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  pressure_per_hour = pressure / runtime_hours,
  maintenance_flag = temp_above_80 | vibration_score > 80,
  high_pressure = pressure >= 38
)

run_refresh <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v2,
  key = key,
  feat_table_name = "reading_features",
  fetch_mode = "all",
  verbose = FALSE
)

# Refresh mode fetches all rows returned by the source query.
run_refresh$fetch$mode
run_refresh$fetch$n_rows

# All rows already existed in the feature table, so they were updated.
# No new rows were inserted.
run_refresh$upsert$counts
```

This is one of the most important production distinctions. `new_only` is for
incremental append-style processing. `all` is for refreshes, backfills, and
definition changes.

## Treat schema evolution as additive

With `alter_table = TRUE`, missing feature columns are added to the target
table. This is convenient when a new feature is introduced.

The package deliberately does not drop, rename, or retire old columns
automatically. If a column exists in the database feature table but is no longer
produced by the current definitions, it is reported as an extra column and left
untouched.

In the example below, `defs_v3` no longer produces the previously created
`pressure_per_hour` feature. We also use `fetch_mode = "all"` so that all
existing rows in the feature table are refreshed with the current definitions.

```{r}
defs_v3 <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  maintenance_flag = temp_above_80 | vibration_score > 80,
  high_pressure = pressure >= 38
)

run_removed_column <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs_v3,
  key = key,
  feat_table_name = "reading_features",
  fetch_mode = "all",
  verbose = FALSE
)

# All ten source rows already existed in the feature table, so they were
# counted as updates rather than inserts.
run_removed_column$upsert$counts

# The database table still contains the old `pressure_per_hour` column.
dbGetQuery(con, "pragma table_info(reading_features)")
```

This behavior is intentional. Removing a feature column from a shared database
table can affect dashboards, models, and other users. In production, column
removal should usually be handled as an explicit database migration.

If your organization requires approval before any schema change, set
`alter_table = FALSE`. Then a new feature column will cause the run to fail
instead of altering the table. To demonstrate this, we define `defs_v4` with a
new feature, `thermal_load`, which does not yet exist in `reading_features`.

```{r}
defs_v4 <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  maintenance_flag = temp_above_80 | vibration_score > 80,
  high_pressure = pressure >= 38,
  thermal_load = temperature * runtime_hours
)

alter_error <- tryCatch(
  fd_run(
    con = con,
    sql = source_sql,
    defs = defs_v4,
    key = key,
    feat_table_name = "reading_features",
    fetch_mode = "all",
    alter_table = FALSE,
    verbose = FALSE
  ),
  error = function(e) conditionMessage(e)
)

alter_error
```

## Use insert-only mode when duplicates should fail

Some pipelines should only append new feature rows. In those workflows, an
incoming key that already exists in the feature table may indicate a logic or
scheduling problem.

Set `update_table = FALSE` to make existing-key conflicts fail. In the example
below, `fetch_mode = "all"` sends rows whose keys already exist in
`reading_features`, so insert-only mode reports a conflict.

```{r}
insert_only_error <- tryCatch(
  fd_run(
    con = con,
    sql = source_sql,
    defs = defs_v3,
    key = key,
    feat_table_name = "reading_features",
    fetch_mode = "all",
    update_table = FALSE,
    verbose = FALSE
  ),
  error = function(e) conditionMessage(e)
)

insert_only_error
```

This is stricter than the default upsert behavior. Use it when accidental
updates should be blocked.

## Use chunking for large writes

For large feature tables, writing everything in one batch may be inconvenient.
Use `chunk_size` to stage and merge rows in batches.

```{r, eval = FALSE}
fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = key,
  feat_table_name = "reading_features",
  chunk_size = 10000
)
```

Chunking does not change which features are computed. It changes how computed
rows are written to the database.

## Be explicit about keys and constraints

The key column is the contract between the raw query and the feature table. In
production, make sure that:

1. the source query returns the key column;
2. the key values are unique and non-missing in the source data;
3. the feature table has the same key column;
4. existing feature tables used with upsert behavior have a primary key or
   unique constraint on that key.

Tables created by `fd_upsert()` include a primary key automatically. If you are
using a pre-existing table, check the database schema before running the
pipeline.

## Avoid concurrent writes to the same feature table

`featdelta` does not currently coordinate concurrent writers. Avoid running
multiple `fd_run()` or `fd_upsert()` calls against the same feature table at the
same time.

Concurrent writes to different feature tables are independent, but concurrent
writes to the same target table should be handled by your scheduler, database
permissions, or orchestration layer.

## A practical production checklist

Before using a feature pipeline repeatedly, it is worth checking a few things.

1. The source SQL returns exactly the population you intend to process.
2. The source SQL returns the key column.
3. The key is unique and has no missing values.
4. Feature definitions have been tested with `fd_compute()`.
5. A small `fd_run()` has been tested against a development table.
6. The intended refresh mode is clear: `new_only` or `all`.
7. Schema evolution is intentional: `alter_table = TRUE` or `FALSE`.
8. The feature table has a primary key or unique constraint on the key.
9. The run report is logged or inspected.
10. The scheduler prevents overlapping writes to the same feature table.

The package handles the repeated database mechanics, but these operational
decisions still belong to the user or the surrounding production system.

## What to remember

Production use is mostly about making the feature workflow predictable. Develop
definitions locally, use a scratch table for small end-to-end tests, inspect the
run report, choose the correct refresh mode, and be intentional about schema
changes.

For regular repeated jobs, `fd_run()` is the main entry point. The supporting
functions, especially `fd_compute()`, `fd_fetch()`, and `fd_upsert()`, are there
when you need to inspect or test individual stages.

```{r cleanup, include = FALSE}
dbDisconnect(con)
```

---
title: "Scheduling featdelta runs"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Scheduling featdelta runs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Purpose of this tutorial

The earlier vignettes showed how to define features and run the database
pipeline with `fd_run()`. In practice, you usually do not want to open RStudio
and run that command manually every day. You want a script that can be executed
by the operating system.

This tutorial shows how to set up that script and schedule it:

1. create a small project folder;
2. save an R script that connects to the database and calls `fd_run()`;
3. test the script manually with `Rscript`;
4. schedule the script with Windows Task Scheduler or cron;
5. inspect a log file after each run.

This vignette is written as an instruction rather than as executed package code.
The database connection uses imaginary PostgreSQL credentials. You should adapt
the connection details, SQL query, feature definitions, and table names to your
own project.

The example is not meant to be copied without changes. It assumes a specific
database, a specific raw table, and a specific set of columns. Your own script
will differ in those places.

## Before you start

This tutorial assumes that the previous vignettes have already explained what
`fd_run()` does. Here we focus only on scheduling.

You need:

1. R installed on the machine that will run the job;
2. the `featdelta` package installed in that R library;
3. the DBI driver for your database, such as `RPostgres` or `RMariaDB`;
4. permission to write to the target feature table;
5. permission to create scheduled tasks on the machine.

Operating-system schedulers are normally minute-based. Windows Task Scheduler
and cron are good for running jobs every minute, hourly, daily, or on a business
schedule. They are not meant for second-by-second demonstrations, so this
tutorial uses minute-level schedules.

## Create a project folder

Create a small folder for the scheduled pipeline. For example:

```text
featdelta_sensor_pipeline/
  scripts/
    run_featdelta.R
  logs/
```

On Windows, this could be:

```text
C:/featdelta_sensor_pipeline/
```

On Linux or macOS, this could be:

```text
/home/analyst/featdelta_sensor_pipeline/
```

The important point is that the scheduled script should use stable, absolute
paths. A scheduled job may not start in the same working directory as your
interactive R session.

## Create the R script

Create this file:

```text
C:/featdelta_sensor_pipeline/scripts/run_featdelta.R
```

On Linux or macOS, use the equivalent path in your project folder, for example:

```text
/home/analyst/featdelta_sensor_pipeline/scripts/run_featdelta.R
```

Save the following script as `run_featdelta.R`.

Before looking at the script, here is the scenario it assumes:

* there is a PostgreSQL database available at a private company host;
* the database name is `analytics`;
* credentials are stored in the environment variables `ANALYTICS_DB_USER` and
  `ANALYTICS_DB_PASSWORD`;
* the raw data is in `raw_schema.device_readings`;
* the raw query returns one row per equipment reading;
* the unique key column is `reading_id`;
* the raw columns used for features are `temperature`, `vibration`, `pressure`,
  and `runtime_hours`;
* the computed features should be stored in
  `feature_schema.reading_features`;
* during testing, the script will be scheduled every minute.

For this example, the features are:

* `temp_above_80`: whether the reading temperature is above 80;
* `vibration_score`: a simple interaction between vibration and runtime;
* `pressure_high`: whether pressure is at or above a chosen threshold;
* `maintenance_flag`: whether the reading should be flagged for follow-up.

In your own project, replace these with the transformations that make sense for
your raw table.

```r
library(DBI)
library(RPostgres)
library(featdelta)

# Use an absolute project directory so the script works from a scheduler.
project_dir <- "C:/featdelta_sensor_pipeline"
log_path <- file.path(project_dir, "logs", "featdelta-run.log")

log_line <- function(...) {
  cat(
    format(Sys.time(), "%Y-%m-%d %H:%M:%S"),
    " | ",
    paste(..., collapse = ""),
    "\n",
    file = log_path,
    append = TRUE,
    sep = ""
  )
}

con <- dbConnect(
  RPostgres::Postgres(),
  # Replace these connection settings with your own database details.
  host = "database.company.local",
  port = 5432,
  dbname = "analytics",
  user = Sys.getenv("ANALYTICS_DB_USER"),
  password = Sys.getenv("ANALYTICS_DB_PASSWORD")
)

on.exit(dbDisconnect(con), add = TRUE)

# Replace this query with the raw dataset you want to process.
# The query must return the key column and every raw column used below.
source_sql <- "
  SELECT
    reading_id,
    device_id,
    temperature,
    vibration,
    pressure,
    runtime_hours
  FROM raw_schema.device_readings
  WHERE reading_status = 'ready'
  ORDER BY reading_id
"

# Replace these definitions with your own feature logic.
defs <- fd_define(
  temp_above_80 = temperature > 80,
  vibration_score = vibration * runtime_hours,
  pressure_high = pressure >= 38,
  maintenance_flag = temp_above_80 | vibration_score > 80 | pressure_high
)

report <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  # Replace the key and feature table with your project-specific values.
  key = "reading_id",
  feat_table_name = "feature_schema.reading_features",
  fail_fast = FALSE,
  verbose = FALSE
)

would_insert <- if (is.null(report$upsert)) {
  NA_integer_
} else {
  report$upsert$counts$would_insert
}

would_update <- if (is.null(report$upsert)) {
  NA_integer_
} else {
  report$upsert$counts$would_update
}

log_line(
  "success=", report$success,
  ", stage=", report$stage,
  ", fetched=", report$fetch$n_rows,
  ", inserts=", would_insert,
  ", updates=", would_update
)

if (!isTRUE(report$success)) {
  log_line("error=", report$error$message)
  stop(report$error$message)
}
```

This script does four things:

1. opens a database connection;
2. defines the source SQL query;
3. defines the R feature expressions;
4. calls `fd_run()` and writes a compact log line.

The parts you normally change are:

1. `project_dir`, so logs are written to your project folder;
2. `dbConnect()`, so the script reaches your database;
3. `source_sql`, so the script selects your raw data;
4. `defs`, so the script computes your features;
5. `key`, so `featdelta` knows the unique row identifier;
6. `feat_table_name`, so features are written to the correct database table;
7. the scheduler frequency, so the job runs when your raw data is ready.

For MySQL or MariaDB, the connection part would look similar, but with
`RMariaDB`:

```r
library(DBI)
library(RMariaDB)

con <- dbConnect(
  RMariaDB::MariaDB(),
  host = "database.company.local",
  port = 3306,
  dbname = "analytics",
  user = Sys.getenv("ANALYTICS_DB_USER"),
  password = Sys.getenv("ANALYTICS_DB_PASSWORD")
)
```

If automatic dialect detection is not sufficient in your environment, pass the
dialect explicitly:

```r
report <- fd_run(
  con = con,
  sql = source_sql,
  defs = defs,
  key = "reading_id",
  feat_table_name = "feature_schema.reading_features",
  dialect = "postgres",
  fail_fast = FALSE,
  verbose = FALSE
)
```

For MySQL or MariaDB, use `dialect = "mysql"`.

## Store credentials outside the script

The example reads credentials from environment variables:

```r
Sys.getenv("ANALYTICS_DB_USER")
Sys.getenv("ANALYTICS_DB_PASSWORD")
```

This is safer than writing passwords directly into the R file. The exact way to
set environment variables depends on your operating system and security setup.

For local testing, you can set them temporarily inside an R session:

```r
Sys.setenv(ANALYTICS_DB_USER = "analyst")
Sys.setenv(ANALYTICS_DB_PASSWORD = "secret")
```

For production, use your organization's preferred method, such as user-level
environment variables, a secrets manager, or a protected configuration file.

## Test the script manually

Before scheduling anything, run the script manually from a terminal.

On Windows PowerShell:

```powershell
& "C:\Program Files\R\R-4.5.2\bin\Rscript.exe" "C:\featdelta_sensor_pipeline\scripts\run_featdelta.R"
```

On Linux or macOS:

```bash
Rscript /home/analyst/featdelta_sensor_pipeline/scripts/run_featdelta.R
```

Then inspect the log file:

```text
C:/featdelta_sensor_pipeline/logs/featdelta-run.log
```

A successful line might look like this:

```text
2026-05-05 09:00:04 | success=TRUE, stage=complete, fetched=12, inserts=12, updates=0
```

If the script fails manually, fix that first. A scheduler will not make an
unreliable script more reliable. The script must be able to run from a plain
terminal command before you schedule it.

## Windows: schedule with taskscheduleR

On Windows, you can use the `taskscheduleR` package to create a Windows Task
Scheduler job from R.

Install it if needed:

```r
install.packages("taskscheduleR")
```

Then run this once from an interactive R session:

```r
library(taskscheduleR)

taskscheduler_create(
  taskname = "featdelta_sensor_pipeline",
  rscript = "C:/featdelta_sensor_pipeline/scripts/run_featdelta.R",
  schedule = "MINUTE",
  starttime = format(Sys.time() + 70, "%H:%M"),
  modifier = 1
)
```

This creates a task that runs the script every minute. That is a useful testing
frequency because you can watch the log file and database table change shortly
after creating the task.

In this demonstration, "every minute" is only a testing schedule. A real
production schedule should match the arrival pattern of your raw data. For
example, if new sensor readings are loaded every hour, schedule the job after
that hourly load. If a warehouse table is refreshed once per night, schedule
`fd_run()` after the nightly refresh.

To list scheduled tasks created through `taskscheduleR`:

```r
taskscheduler_ls()
```

To remove the test task:

```r
taskscheduler_delete("featdelta_sensor_pipeline")
```

For production, use a less aggressive schedule, such as hourly, daily, or after
the expected raw-data load has finished.

## Linux or macOS: schedule with cronR

On Linux or macOS, you can use the `cronR` package to create a cron job.

Install it if needed:

```r
install.packages("cronR")
```

Then run this once from an interactive R session:

```r
library(cronR)

cmd <- cron_rscript(
  rscript = "/home/analyst/featdelta_sensor_pipeline/scripts/run_featdelta.R",
  rscript_log = "/home/analyst/featdelta_sensor_pipeline/logs/featdelta-cron.log",
  log_append = TRUE
)

cron_add(
  command = cmd,
  frequency = "minutely",
  id = "featdelta_sensor_pipeline",
  description = "Run featdelta sensor feature pipeline"
)
```

This creates a cron job that runs every minute. During testing, inspect:

```text
/home/analyst/featdelta_sensor_pipeline/logs/featdelta-run.log
/home/analyst/featdelta_sensor_pipeline/logs/featdelta-cron.log
```

To list cron jobs:

```r
cron_ls()
```

To remove the test job:

```r
cron_rm(id = "featdelta_sensor_pipeline")
```

For production, use a schedule that matches your data refresh cycle. For
example, a daily run after midnight, an hourly run during business hours, or a
custom cron expression.

## What to monitor

At minimum, monitor the log file written by the script. Each run should tell you:

1. whether the run succeeded;
2. which stage completed;
3. how many rows were fetched;
4. how many rows were inserted;
5. how many rows were updated.

You may also want to query the feature table directly:

```sql
SELECT COUNT(*) AS n_rows
FROM feature_schema.reading_features;
```

Or inspect the most recent keys:

```sql
SELECT *
FROM feature_schema.reading_features
ORDER BY reading_id DESC
LIMIT 20;
```

The exact monitoring setup depends on your database and scheduler, but the
principle is simple: make each scheduled run leave evidence that it ran and what
it changed.

## Common scheduler issues

Scheduled jobs often fail for environmental reasons rather than because of
`fd_run()` itself.

Common issues include:

1. the scheduler uses a different R installation than RStudio;
2. the scheduled R library does not have `featdelta` or the DBI driver
   installed;
3. environment variables are available interactively but not to the scheduled
   process;
4. the script uses relative paths;
5. the scheduler account does not have database access;
6. the scheduler account cannot write to the log folder.

When debugging, first run the exact `Rscript` command from a terminal. Then make
sure the scheduler runs under a user account that has the same package library,
environment variables, file permissions, and database permissions.

## What to remember

Scheduling is not part of `featdelta` itself. The package gives you the
repeatable R feature pipeline. Windows Task Scheduler or cron runs that pipeline
at the chosen time.

The practical setup is:

1. write a standalone `run_featdelta.R`;
2. test it manually with `Rscript`;
3. log each run;
4. schedule it with `taskscheduleR` or `cronR`;
5. monitor the log and feature table.

## References

The scheduling examples use the public documentation for:

* `taskscheduleR::taskscheduler_create()`:
  <https://rdrr.io/cran/taskscheduleR/man/taskscheduler_create.html>
* `cronR::cron_add()` and `cronR::cron_rscript()`:
  <https://search.r-project.org/CRAN/refmans/cronR/html/cron_add.html>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_fetch.R
\name{fd_fetch}
\alias{fd_fetch}
\title{Fetch source rows that are not yet present in the features table}
\usage{
fd_fetch(con, sql, key, feat_table_name, use_max_key = FALSE, verbose = FALSE)
}
\arguments{
\item{con}{A live \code{DBIConnection}, created with \code{\link[DBI:dbConnect]{DBI::dbConnect()}}.}

\item{sql}{A single SQL string (typically a \code{SELECT}) that defines the \strong{source
dataset} you want to process into features. In other words, it should return
the rows you would normally feed into your feature creation pipeline.
The query may join multiple tables and apply filters.

The query must be usable as a derived table: \verb{FROM (<sql>) AS r}.
Trailing semicolons are tolerated and removed.}

\item{key}{Name of the primary key column (character scalar). Must exist in
both the result of \code{sql} and in \code{feat_table_name}. This is the identifier used to
decide whether a source row has already been processed into the features table.}

\item{feat_table_name}{Name of the existing features table in the database
(character scalar). Must exist. This table is used only to identify which
\code{key} values are already present.}

\item{use_max_key}{Logical. If \code{TRUE}, \code{fd_fetch()} may reduce join work by
first computing \code{MAX(feat_table_name.key)} and selecting rows with \code{key > max_key}
(guaranteed not in \code{feat_table_name}). If \code{FALSE}, \code{fd_fetch()} selects all the rows
with any key no in \code{feat_table_name.key} which may be computationally more expensive.}

\item{verbose}{Logical. If \code{TRUE}, prints the executed SQL.}
}
\value{
A \code{data.frame} containing only rows returned by \code{sql} whose \code{key} is not
present in \code{feat_table_name}. The result includes an attribute \code{attr(x, "fd_fetch")}
(a list) with metadata such as \code{key}, \code{feat_table_name}, \code{use_max_key}, \code{max_key}
(if computed), \code{executed_sql}, and \code{n_rows}.
}
\description{
\code{fd_fetch()} executes a user-supplied SQL \code{SELECT} query against a database and
returns \strong{only} those rows whose primary key (\code{key}) is \strong{not present} in an
existing features table (\code{feat_table_name}).
}
\details{
This function is intentionally not a general-purpose query runner. It always
applies a "not yet processed" filter against \code{feat_table_name} and errors if that
cannot be done honestly (e.g., missing tables/columns).
\subsection{Important assumption}{

\code{fd_fetch()} assumes that:
\itemize{
\item the \code{sql} query is executed on the \strong{same database} (same \code{con}) where
\code{feat_table_name} is stored, and
\item the returned \code{key} values are comparable to \code{feat_table_name.key}.
}

Cross-database fetching (e.g., pulling data from one database and comparing to
a features table stored in another database) is not supported, because the
"not present in \code{feat_table_name}" filter must be evaluated by the database engine
in a single query context.
}
}
\examples{
if (requireNamespace("RSQLite", quietly = TRUE)) {
con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
on.exit(DBI::dbDisconnect(con), add = TRUE)

DBI::dbExecute(con, "CREATE TABLE raw (id INTEGER, x INTEGER)")
DBI::dbExecute(con, "CREATE TABLE r_variables_table (id INTEGER)")
DBI::dbExecute(con, "INSERT INTO raw (id, x) VALUES (1,10), (2,20), (3,30), (4,40), (5,50)")
DBI::dbExecute(con, "INSERT INTO r_variables_table (id) VALUES (1), (2), (4)")

# Returns ids 3 and 5 only (rows not yet present in the features table)
new_rows <- fd_fetch(
  con = con,
  sql = "SELECT * FROM raw",
  key = "id",
  feat_table_name = "r_variables_table"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_define.R
\name{print.featdelta_defs}
\alias{print.featdelta_defs}
\title{Print a featdelta definitions object}
\usage{
\method{print}{featdelta_defs}(x, ...)
}
\arguments{
\item{x}{A \code{featdelta_defs} object.}

\item{...}{Unused.}
}
\value{
The input object invisibly.
}
\description{
Prints a concise human-readable summary of a \code{featdelta_defs} object.
}
\seealso{
Other featdelta defs helpers: 
\code{\link{fd_block}()},
\code{\link{fd_define}()},
\code{\link{summary.featdelta_defs}()}
}
\concept{featdelta defs helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_upsert.R
\name{fd_upsert}
\alias{fd_upsert}
\title{Upsert computed features into a database features table}
\usage{
fd_upsert(
  con,
  features_df,
  feat_table_name,
  key,
  create_table = "auto",
  alter_table = TRUE,
  update_table = TRUE,
  chunk_size = NULL,
  verbose = TRUE,
  return_report = TRUE,
  dialect = NULL
)
}
\arguments{
\item{con}{A \code{DBIConnection}. Must be a live connection to the database where
\code{feat_table_name} exists (or will be created).}

\item{features_df}{A \code{data.frame} containing at minimum the primary key column
specified by \code{key}, plus one or more feature columns to be stored.
Requirements:
\itemize{
\item \code{key} must exist in \code{names(features_df)}.
\item \code{features_df[[key]]} must contain no \code{NA} and no duplicate values.
\item Each feature column should be an atomic vector (numeric, integer, logical,
character, Date/POSIXct) that can be written by DBI for the chosen dialect.
}}

\item{feat_table_name}{A single string. Name of the database table where features
are stored (the "target" table). For dialects supporting schemas (e.g. Postgres),
this may be schema-qualified (e.g. \code{"public.my_features"}), subject to dialect rules.}

\item{key}{A single string. Name of the primary key column in both \code{features_df}
and \code{feat_table_name}. This column uniquely identifies each entity/row.}

\item{create_table}{Logical, or character value \code{auto}. If \code{TRUE}, create \code{feat_table_name} when it does not exist.
The created table includes all columns present in \code{features_df} and defines
\verb{PRIMARY KEY(key)}. If \code{FALSE} and \code{feat_table_name} is missing, the function errors.If \code{auto},
create the table it hasn't been already created.}

\item{alter_table}{Logical. If \code{TRUE}, add missing feature columns found in
\code{features_df} but not present in \code{feat_table_name} using \verb{ALTER TABLE ... ADD COLUMN ...}.
Only additive schema changes are performed. If \code{FALSE}, the function errors
when \code{features_df} contains columns not present in \code{feat_table_name}.}

\item{update_table}{Logical. Controls incremental behavior for keys that already exist
in \code{feat_table_name}:
\itemize{
\item If \code{TRUE} (default), existing keys are updated (upsert).
\item If \code{FALSE}, only new keys are inserted; any conflict with existing keys
results in an error (insert-only mode).
}}

\item{chunk_size}{Optional integer batch size. If provided, \code{features_df} is written
in batches of up to \code{chunk_size} rows, each merged via a staging table.
Use this to limit memory/packet sizes for large writes. If \code{NULL}, write in one batch.}

\item{verbose}{Logical. If \code{TRUE}, emit progress messages (e.g., chunk progress,
table creation/alteration actions). Does not affect returned results.}

\item{return_report}{Logical. If \code{TRUE}, return a structured
\code{fd_upsert_report}; if \code{FALSE}, return \code{TRUE} invisibly after successful
side effects.}

\item{dialect}{Optional dialect override. Supported values are \code{"postgres"},
\code{"sqlite"}, and \code{"mysql"}.}
}
\value{
An \code{fd_upsert_report} object (S3) with a structured summary of actions performed:
\itemize{
\item \code{table_created} (logical)
\item \code{columns_added} (character vector)
\item \code{extra_columns} (character vector): target-table columns not present
in the incoming features data, excluding \code{key}
\item \code{counts$would_insert} (integer)
\item \code{counts$would_update} (integer; \code{0} when \code{update_table = FALSE})
\item optional per-chunk breakdown (if chunking is used)
}
}
\description{
Takes a feature \code{data.frame} produced by \code{fd_compute()} (one row per entity id)
and writes it into a database table, using explicit incremental semantics:
new keys are inserted; existing keys are optionally updated.
}
\details{
The function is set-based (staging + merge) and avoids "read all ids into R"
patterns. Writes are performed in chunks (optional) inside a transaction.

Schema evolution is supported in a restricted, safe form: when \code{alter_table = TRUE},
missing feature columns are added to the target table via \verb{ALTER TABLE ... ADD COLUMN ...}.
No column drops/renames/type changes are performed. Columns that exist in
the target table but are not present in \code{features_df} are left untouched and
reported as \code{extra_columns}.

\strong{Incremental semantics}
\itemize{
\item Insert new keys (keys in \code{features_df} not present in \code{feat_table_name}).
\item Update existing keys only when \code{update_table = TRUE}.
\item When \code{update_table = FALSE}, any overlap between staged keys and existing keys
is treated as a conflict and aborts the write.
}

\strong{Counts}
The returned report contains \code{would_insert} / \code{would_update} counts, computed as
existence-based counts \emph{prior to the merge} (within the transaction). These are
not "rows whose values changed", only "rows targeted as insert/update".

\strong{Transactions}
The operation runs in a transaction. If any step fails (schema change, staging write,
merge), the transaction is rolled back and the target table is unchanged.
Existing target tables used with \code{update_table = TRUE} must have a primary
key or unique constraint/index on \code{key}. Tables created by \code{fd_upsert()} get
this primary key automatically.

\strong{Concurrency}
\code{fd_upsert()} does not currently take an explicit table-level lock. Avoid
running concurrent writes to the same feature table. Concurrent writes to
different feature tables are independent.
}
\examples{
if (requireNamespace("RSQLite", quietly = TRUE)) {
con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
on.exit(DBI::dbDisconnect(con), add = TRUE)

feats <- data.frame(
  id = c(1, 2, 3),
  f_age = c(10, 20, 30),
  f_flag = c(TRUE, FALSE, TRUE)
)

# Create table and insert initial rows
r1 <- fd_upsert(
  con = con,
  features_df = feats,
  feat_table_name = "features_tbl",
  key = "id",
  create_table = TRUE,
  alter_table = FALSE,
  update_table = TRUE,
  verbose = FALSE
)

# Upsert: update ids 2-3, insert id 4
feats2 <- data.frame(
  id = c(2, 3, 4),
  f_age = c(21, 31, 40),
  f_flag = c(FALSE, TRUE, FALSE)
)

r2 <- fd_upsert(
  con = con,
  features_df = feats2,
  feat_table_name = "features_tbl",
  key = "id",
  update_table = TRUE,
  verbose = FALSE
)

# Schema evolution: add a new feature column
feats3 <- data.frame(
  id = c(4, 5),
  f_age = c(41, 50),
  f_flag = c(TRUE, TRUE),
  f_new = c("A", "B")
)

r3 <- fd_upsert(
  con = con,
  features_df = feats3,
  feat_table_name = "features_tbl",
  key = "id",
  alter_table = TRUE,
  update_table = TRUE,
  verbose = FALSE
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_define.R
\name{fd_define}
\alias{fd_define}
\title{Define featdelta feature definitions}
\usage{
fd_define(
  ...,
  defs = NULL,
  description = NULL,
  overwrite = FALSE,
  envir = NULL
)
}
\arguments{
\item{...}{Definitions supplied inline.
Ordinary named inputs such as \code{new_feature = hp / cyl} are stored as
single-column definition steps.}

\item{defs}{Optional list of programmatically supplied definitions.
Supported inputs are the same as for \code{...}.}

\item{description}{Optional character scalar describing the definitions.}

\item{overwrite}{Logical. If \code{FALSE}, duplicate step names are rejected.
If \code{TRUE}, later steps replace earlier ones.}

\item{envir}{Optional environment used as the explicit evaluation
environment for normalized definitions. If \code{NULL}, environments are taken
from the captured inputs.}
}
\value{
A list of class \code{"featdelta_defs"} containing:
\itemize{
\item \code{steps}: ordered normalized definition steps
\item \code{description}: optional description
\item \code{created_at}: creation time
\item \code{defs_version}: internal version marker
\item \code{envir_policy}: whether environments were inherited or explicit
\item \code{envir}: the explicit environment, if supplied
}
}
\description{
Creates a \code{featdelta_defs} object that stores ordered definition steps in a
normalized internal representation suitable for later evaluation by
\code{fd_compute()}.
}
\details{
\code{fd_define()} is the definitions-construction step of the package. It does
not compute features and does not access the database. Its role is to:
\itemize{
\item capture user definitions,
\item normalize supported input styles to a canonical internal structure,
\item enforce definition-level guardrails such as name validity and
duplicate handling,
\item store the resulting ordered definition steps for later execution.
}

Definitions are evaluated later by \code{fd_compute()} in the order they are
stored. This means later definitions may depend on columns created by
earlier steps. Reordering or removing earlier steps may therefore break
downstream definitions.

Supported definition inputs include:
\itemize{
\item inline expressions captured from \code{...}
\item quosures
\item \code{expression()} objects of length 1
\item quoted calls or names
\item direct scalar constants
\item \code{fd_block()} objects for multi-column feature generation
}

Ordinary named expressions define one output column per step.

\code{fd_block()} defines a multi-column step. At compute time, a block is
expected to return a \code{data.frame} with one row per input row. The returned
column names become the produced feature names. Optionally, expected output
names may be declared in advance via \code{fd_block(expected_names = ...)}.

Direct atomic constants must be scalar. If vectorized behavior is desired,
provide an expression instead of embedding a fixed-length atomic vector in
the definitions.

Duplicate step handling is part of definitions construction:
\itemize{
\item with \code{overwrite = FALSE}, duplicate step names error;
\item with \code{overwrite = TRUE}, the last definition wins.
}
}
\examples{
# Single-column definitions
defs <- fd_define(
  hp_per_cyl = hp / cyl,
  wt_per_hp = wt / hp
)

# Definitions can also be supplied programmatically
predef <- expression(log(hp))
defs <- fd_define(
  log_hp = predef
)

# Multi-column block using an inline expression
defs <- fd_define(
  engine_ratios = fd_block({
    data.frame(
      hp_per_cyl = hp / cyl,
      disp_per_cyl = disp / cyl
    )
  })
)

# Multi-column block with declared expected names
defs <- fd_define(
  engine_ratios = fd_block(
    {
      data.frame(
        hp_per_cyl = hp / cyl,
        disp_per_cyl = disp / cyl
      )
    },
    expected_names = c("hp_per_cyl", "disp_per_cyl")
  )
)

# Function-based block
make_engine_features <- function(data) {
  data.frame(
    hp_per_cyl = data$hp / data$cyl,
    disp_per_cyl = data$disp / data$cyl
  )
}

defs <- fd_define(
  engine_ratios = fd_block(make_engine_features)
)

# A block can contain a small feature script with temporary variables
defs <- fd_define(
  engine_script = fd_block({
    hp_per_cyl <- hp / cyl
    disp_per_cyl <- disp / cyl

    data.frame(
      hp_per_cyl = hp_per_cyl,
      disp_per_cyl = disp_per_cyl,
      engine_index = hp_per_cyl + disp_per_cyl
    )
  })
)

# A function-based block can create a variable number of columns in a loop
make_scaled_features <- function(data) {
  vars <- c("hp", "disp", "wt")
  out <- list()

  for (var in vars) {
    center <- mean(data[[var]], na.rm = TRUE)
    spread <- stats::sd(data[[var]], na.rm = TRUE)
    out[[paste0(var, "_scaled")]] <- (data[[var]] - center) / spread
  }

  as.data.frame(out)
}

defs <- fd_define(
  scaled_inputs = fd_block(make_scaled_features)
)

}
\seealso{
Other featdelta defs helpers: 
\code{\link{fd_block}()},
\code{\link{print.featdelta_defs}()},
\code{\link{summary.featdelta_defs}()}
}
\concept{featdelta defs helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_define.R
\name{summary.featdelta_defs}
\alias{summary.featdelta_defs}
\title{Summarize a featdelta definitions object}
\usage{
\method{summary}{featdelta_defs}(object, ...)
}
\arguments{
\item{object}{A \code{featdelta_defs} object.}

\item{...}{Unused.}
}
\value{
A data frame with one row per definition step and columns such as
step name, type, mode, declared outputs, and stored expression text.
}
\description{
Returns a compact data-frame summary of the ordered definition steps stored
in a \code{featdelta_defs} object.
}
\seealso{
Other featdelta defs helpers: 
\code{\link{fd_block}()},
\code{\link{fd_define}()},
\code{\link{print.featdelta_defs}()}
}
\concept{featdelta defs helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_run.R
\name{print.fd_run_report}
\alias{print.fd_run_report}
\title{Print a featdelta run report}
\usage{
\method{print}{fd_run_report}(x, ...)
}
\arguments{
\item{x}{An \code{fd_run_report} object.}

\item{...}{Unused.}
}
\value{
The input object invisibly.
}
\description{
Print a featdelta run report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_block.R
\name{fd_block}
\alias{fd_block}
\title{Define a multi-column feature block}
\usage{
fd_block(x, expected_names = NULL, envir = NULL)
}
\arguments{
\item{x}{Block definition input. Must be either a function or a single
expression-like input that later evaluates to a \code{data.frame}.}

\item{expected_names}{Optional character vector of expected output column
names. If supplied, names must be non-empty, non-\code{NA}, and unique.
These names are not required for normal operation, but they can later be
used by \code{fd_compute()} for validation and for optional NA-completion of
expected-but-missing outputs.}

\item{envir}{Optional environment used when normalizing expression-like
inputs. If \code{NULL}, the calling environment is used.}
}
\value{
An object of class \code{"featdelta_block"}.
}
\description{
Creates a block definition object for use inside \code{fd_define()}.
}
\details{
A block represents one ordered definition step that may produce multiple
feature columns when later evaluated by \code{fd_compute()}. Unlike ordinary
named single-column definitions, a block may return a whole \code{data.frame}
whose column names become the produced feature names.

Supported block inputs are:
\itemize{
\item an inline expression, including braced expressions such as
\code{{ data.frame(...) }}
\item a quosure
\item an \code{expression()} object of length 1
\item a quoted call or symbol
\item a function, intended to receive the current working data and return a
\code{data.frame}
}

\code{fd_block()} does not compute anything by itself. It only stores the block
definition in a normalized form so that \code{fd_define()} can include it in the
ordered definitions object.
}
\examples{
# Inline expression block
blk <- fd_block({
  data.frame(
    hp_per_cyl = hp / cyl,
    disp_per_cyl = disp / cyl
  )
})

# Function-based block
make_engine_features <- function(data) {
  data.frame(
    hp_per_cyl = data$hp / data$cyl,
    disp_per_cyl = data$disp / data$cyl
  )
}

blk <- fd_block(make_engine_features)

# Optional expected names
blk <- fd_block(
  {
    data.frame(
      hp_per_cyl = hp / cyl
    )
  },
  expected_names = c("hp_per_cyl", "disp_per_cyl")
)

}
\seealso{
Other featdelta defs helpers: 
\code{\link{fd_define}()},
\code{\link{print.featdelta_defs}()},
\code{\link{summary.featdelta_defs}()}
}
\concept{featdelta defs helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_connect.R
\name{fd_connect}
\alias{fd_connect}
\title{Connect to a database with featdelta defaults}
\usage{
fd_connect(
  driver,
  ...,
  raw_table = NULL,
  raw_table_name = NULL,
  feat_table_name = NULL,
  key = NULL,
  meta_enabled = FALSE,
  meta_schema = NULL
)
}
\arguments{
\item{driver}{A DBI driver object, such as \code{RSQLite::SQLite()},
\code{RPostgres::Postgres()}, or \code{RMariaDB::MariaDB()}.}

\item{...}{Additional arguments passed to \code{DBI::dbConnect()}.}

\item{raw_table}{Optional character scalar naming the raw/source table.}

\item{raw_table_name}{Deprecated alias for \code{raw_table}.}

\item{feat_table_name}{Optional character scalar naming the feature table.}

\item{key}{Optional character scalar naming the primary key column.}

\item{meta_enabled}{Logical. Whether featdelta metadata tracking is enabled.}

\item{meta_schema}{Optional list of metadata schema settings.}
}
\value{
A \code{featdelta_con} object.
}
\description{
Creates a \code{featdelta_con} context object from a DBI driver and connection
arguments. The returned object stores the live DBI connection plus optional
featdelta defaults such as the raw table, feature table, key column, and
metadata settings.
}
\concept{featdelta context helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_run.R
\name{fd_run}
\alias{fd_run}
\title{Run the featdelta incremental feature pipeline}
\usage{
fd_run(
  con,
  sql,
  defs,
  key,
  feat_table_name,
  verbose = FALSE,
  fetch_mode = c("new_only", "all"),
  use_max_key = FALSE,
  fetch_limit = NULL,
  compute_strict = TRUE,
  compute_envir = NULL,
  create_table = "auto",
  alter_table = TRUE,
  update_table = TRUE,
  dialect = NULL,
  chunk_size = NULL,
  fail_fast = TRUE,
  return_data = c("none", "features", "raw", "both"),
  preview_n = 10L,
  ...
)
}
\arguments{
\item{con}{A live \code{DBI} connection.}

\item{sql}{SQL query used to fetch raw rows.}

\item{defs}{A \code{featdelta_defs} object created by \code{fd_define()}.}

\item{key}{Character scalar naming the key column shared by raw data and
feature table.}

\item{feat_table_name}{Character scalar naming the target feature table.}

\item{verbose}{Logical. If \code{TRUE}, print progress messages.}

\item{fetch_mode}{Fetch mode. \code{"new_only"} fetches rows whose key is not yet
present in \code{feat_table_name}. \code{"all"} fetches all rows returned by \code{sql},
allowing explicit refresh/backfill runs.}

\item{use_max_key}{Logical. Passed to \code{fd_fetch()} when \code{fetch_mode = "new_only"} and the feature table already exists.}

\item{fetch_limit}{Optional positive row limit applied after fetching. This
is intended for previews and small dry development runs, not as a SQL
optimizer.}

\item{compute_strict}{Logical strictness flag passed to \code{fd_compute()}.}

\item{compute_envir}{Optional environment override passed to \code{fd_compute()}.}

\item{create_table}{Logical or \code{"auto"}. Passed to \code{fd_upsert()}.}

\item{alter_table}{Logical. Passed to \code{fd_upsert()}.}

\item{update_table}{Logical. Passed to \code{fd_upsert()}.}

\item{dialect}{Optional dialect override. Supported values are \code{"postgres"},
\code{"sqlite"}, and \code{"mysql"}.}

\item{chunk_size}{Optional chunk size. Passed to \code{fd_upsert()}.}

\item{fail_fast}{Logical. If \code{TRUE}, stage errors are raised immediately. If
\code{FALSE}, errors from fetch, compute, or upsert are captured in the returned
\code{fd_run_report}.}

\item{return_data}{Controls whether raw and/or computed data should be
included in the returned run report.}

\item{preview_n}{Non-negative number of rows to include in report previews.}

\item{...}{Reserved for future context options passed to \code{resolve_ctx()}.}
}
\value{
An object of class \code{"fd_run_report"} containing stage summaries,
timings, row counts, the compute report, and the upsert report. Depending
on \code{return_data}, it may also contain raw and/or computed feature data.
}
\description{
Executes the standard featdelta pipeline:
}
\details{
\code{fd_fetch()} -> \code{fd_compute()} -> \code{fd_upsert()}

\code{fd_run()} is the orchestration entry point that connects database fetching,
in-memory computation, and database upsert into a single incremental run.

\code{fetch_mode = "new_only"} is the default incremental mode. If the feature
table already exists, \code{fd_fetch()} returns only rows whose key is missing
from the feature table. If the feature table does not exist yet, all rows
returned by \code{sql} are fetched because no rows have been processed. This mode
is key-based: it does not recompute existing feature-table rows just because
feature definitions changed or new feature definitions were added.

\code{fetch_mode = "all"} is an explicit refresh/backfill mode. It recomputes all
rows returned by \code{sql} and passes them to \code{fd_upsert()}. With the default
\code{update_table = TRUE}, existing keys are updated and new keys are inserted.
Use this mode when existing feature values should be refreshed, for example
after changing a definition or adding a feature that should be backfilled for
already-processed keys.

\code{fd_run()} does not currently coordinate concurrent writers. For the MVP,
avoid running multiple \code{fd_run()}/\code{fd_upsert()} calls against the same feature
table at the same time. Concurrent writes to different feature tables are
independent.

The returned upsert report can include \code{extra_columns}: columns that exist in
the target feature table but are not produced by the current definitions.
These columns are left untouched; the package does not drop, rename, or retire
columns automatically.
}
\examples{
if (requireNamespace("RSQLite", quietly = TRUE)) {
con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
on.exit(DBI::dbDisconnect(con), add = TRUE)

raw_cars <- mtcars
raw_cars$id <- seq_len(nrow(raw_cars))
DBI::dbWriteTable(con, "raw_cars", raw_cars)

defs <- fd_define(
  hp_per_cyl = hp / cyl,
  engine_ratios = fd_block({
    data.frame(
      disp_per_cyl = disp / cyl,
      wt_per_hp = wt / hp
    )
  })
)

res <- fd_run(
  con = con,
  sql = "SELECT * FROM raw_cars",
  defs = defs,
  key = "id",
  feat_table_name = "features",
  verbose = FALSE
)
res$success
}

}
\concept{featdelta pipeline helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_compute.R
\name{fd_compute}
\alias{fd_compute}
\title{Compute features from featdelta definitions on in-memory data}
\usage{
fd_compute(
  data,
  defs,
  key,
  compute_envir = NULL,
  compute_strict = TRUE,
  verbose = FALSE,
  return_report = FALSE
)
}
\arguments{
\item{data}{A data frame/tibble containing the raw input variables used to
compute features.}

\item{defs}{A \code{featdelta_defs} object created by \code{fd_define()}.}

\item{key}{Character scalar naming the primary key column in \code{data}.}

\item{compute_envir}{Optional environment. If supplied, it overrides the
stored evaluation environment for expression-based steps.}

\item{compute_strict}{Logical. If \code{TRUE}, any failed step causes the overall
compute call to error after the per-step report is assembled. If \code{FALSE},
failed column steps return \code{NA} columns, and failed block steps return
\code{NA} columns only when \code{output_names} are known in advance.}

\item{verbose}{Logical. If \code{TRUE}, print progress messages and a short
summary.}

\item{return_report}{Logical. If \code{TRUE}, return a list with elements
\code{data} and \code{report}. If \code{FALSE}, return only the computed data frame and
attach the report as \code{attr(x, "fd_report")}.}
}
\value{
Either:
\itemize{
\item a data frame with class \code{"featdelta_features"} and attached report
attribute, or
\item a list \verb{\{data, report\}} with class
\code{"featdelta_compute_result"} when \code{return_report = TRUE}
}

The output preserves row order and row count relative to \code{data}.
}
\description{
Evaluates a \code{featdelta_defs} object created by \code{fd_define()} against an
in-memory data frame/tibble and returns a feature data frame containing the
key column plus computed feature columns.
}
\details{
\code{fd_compute()} is the package's pure in-memory execution step. It does not
access the database and does not use the featdelta context system.

Definition steps are evaluated sequentially in the order stored in \code{defs}.
This means later steps may depend on columns created by earlier steps in the
same compute call.

Guardrails implemented by \code{fd_compute()} are compute-specific:
\itemize{
\item validate the structure of the supplied defs object
\item validate result type and row alignment for each step
\item preserve 1:1 row alignment with the input data
\item reject output-name collisions with \code{key} and previously produced names
\item attach a per-step computation report
}

Guardrails that belong at definition time, such as malformed construction of
individual definition steps, should be handled by \code{fd_define()}.

Supported definition step types are:
\itemize{
\item ordinary single-column steps
\item \code{fd_block()} multi-column steps
}

For single-column steps, the result must be a supported vector-like output.

For block steps:
\itemize{
\item the result must be a \code{data.frame}
\item the number of rows must match \code{nrow(data)}
\item returned column names must be non-empty and unique
\item returned names must not collide with \code{key}
\item returned names must not collide with feature names already produced
earlier in the same compute run
}

If a block declares \code{expected_names}, then:
\itemize{
\item returned names must be a subset of \code{expected_names}
\item missing expected names are added as \code{NA}
\item final output order follows \code{expected_names}
}

Because definitions are evaluated sequentially, later steps may use:
\itemize{
\item raw input columns from \code{data}
\item columns produced by earlier single-column steps
\item columns produced by earlier block steps
}

Reordering or removing earlier steps may therefore break downstream
definitions. When evaluation fails with "object not found"-style errors,
\code{fd_compute()} augments the message with likely reasons.
}
\examples{
raw_df <- mtcars
raw_df$car_id <- seq_len(nrow(raw_df))

defs <- fd_define(
  hp_per_cyl = hp / cyl,
  engine_ratios = fd_block({
    data.frame(
      disp_per_cyl = disp / cyl,
      wt_per_hp = wt / hp
    )
  }),
  double_ratio = disp_per_cyl * 2
)

out <- fd_compute(
  data = raw_df,
  defs = defs,
  key = "car_id",
  compute_strict = TRUE
)

head(out)

}
\concept{featdelta compute helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fd_connect.R
\name{print.featdelta_con}
\alias{print.featdelta_con}
\title{Print a featdelta connection context}
\usage{
\method{print}{featdelta_con}(x, ...)
}
\arguments{
\item{x}{A \code{featdelta_con} object.}

\item{...}{Unused.}
}
\value{
The input object invisibly.
}
\description{
Print a featdelta connection context
}
